const { app, BrowserWindow, Menu, ipcMain, dialog, shell } = require('electron');
const { execFile } = require('child_process');
const path = require('path');
const fs = require('fs');

// Handle creating/removing shortcuts on Windows when installing/uninstalling.
if (require('electron-squirrel-startup')) { // eslint-disable-line global-require
  app.quit();
}

// Only allow the app to run once
const gotTheLock = app.requestSingleInstanceLock()
if (!gotTheLock) {
  app.quit()
}

ipcMain.on('open-driver-download', (event, arg)=>{
  shell.openExternal("https://motu.com/download");
});

ipcMain.on('open-firmware-updater', (event, arg)=>{
  const child = execFile(path.join(__dirname, '../../../../x64/ULMk5Updater.exe'), [], (error, stdout, stderr) => {
    if (error) {
      throw error;
    }
  });
});

const createWindow = () => {
  // Create the browser window.
  const mainWindow = new BrowserWindow({
    title: "CueMix 5",
    width: 1140,
    height: 820,
    minWidth: 800,
    minHeight: 600,
    icon: path.join(__dirname, 'icon.ico'),
    webPreferences: { nodeIntegration: true }
  });
  mainWindow.on('page-title-updated', (e)=>{e.preventDefault();});
  mainWindow.setMenu(null);

  // and load the index.html of the app.
  mainWindow.loadFile(path.join(__dirname, 'www/index.html'));
  var dgram = require('dgram');
  var s = dgram.createSocket({ 'type':'udp4', 'reuseAddr':true });
  s.on('message', function(msg, rinfo) {
    mainWindow.webContents.executeJavaScript("device.rxUDPAnnounce('" + msg.toString() + "');");
  });
  s.bind(1280);

  mainWindow.on('close', ()=>{ s.close(); });

  // Open the DevTools.
  // mainWindow.webContents.openDevTools();

  ipcMain.on('load-preset', (event, arg)=>{
    dialog.showOpenDialog({ filters: [{ name: 'Preset File', extensions: ['preset'] }], properties: ['openFile'] }).then((result)=>{
      if (!result.canceled && result.filePaths && result.filePaths.length)
        fs.readFile(result.filePaths[0], 'utf-8', (err, data)=>{
          if (err)
            mainWindow.webContents.executeJavaScript("console.log('" + JSON.stringify(err) + "');");
          else
            mainWindow.webContents.executeJavaScript("ds.applyPreset('" + data + "');");
        });
    }).catch((err)=>{
      mainWindow.webContents.executeJavaScript("console.log('" + JSON.stringify(err) + "');");
    });
  });

  ipcMain.on('save-preset', (event, arg)=>{
    dialog.showSaveDialog({ filters: [{ name: 'Preset File', extensions: ['preset'] }] }).then((result)=>{
      if (result.filePath)
        fs.writeFile(result.filePath, arg, (err)=>{ 
          if (err) 
            mainWindow.webContents.executeJavaScript("console.log('" + JSON.stringify(err) + "');");
        });
    }).catch((err)=>{
      mainWindow.webContents.executeJavaScript("console.log('" + JSON.stringify(err) + "');");
    });
  });
};
// This method will be called when Electron has finished
// initialization and is ready to create browser windows.
// Some APIs can only be used after this event occurs.
app.on('ready', createWindow);
// Quit when all windows are closed.
app.on('window-all-closed', () => {
  // On OS X it is common for applications and their menu bar
  // to stay active until the user quits explicitly with Cmd + Q
  if (process.platform !== 'darwin') {
    app.quit();
  }
});
app.on('activate', () => {
  // On OS X it's common to re-create a window in the app when the
  // dock icon is clicked and there are no other windows open.
  if (BrowserWindow.getAllWindows().length === 0) {
    createWindow();
  }
});
// In this file you can include the rest of your app's specific main process
// code. You can also put them in separate files and import them here.