'use strict';

const kDefaultG5Dev = 0;

class Gen5Lib {
  constructor() {
    this.electron = undefined;
    this.lib = undefined;
    this.libCallback = this.libCallback.bind(this);
    this.hasLib = this.hasLib.bind(this);
    this.hasDevice = this.hasDevice.bind(this);
    this.postMessage = this.postMessage.bind(this);
    this.getDiagnostics = this.getDiagnostics.bind(this);
    this.getDeviceInfo = this.getDeviceInfo.bind(this);
    this.getDriverVersion = this.getDriverVersion.bind(this);
    this.getCurrentState = this.getCurrentState.bind(this);
    this.synchDSToCurrentState = this.synchDSToCurrentState.bind(this);
    this.setHostBufferSize = this.setHostBufferSize.bind(this);
    this.setExtraSafetyOffset1X = this.setExtraSafetyOffset1X.bind(this);
    this.setWaveInEnables = this.setWaveInEnables.bind(this);
    this.setWaveOutEnables = this.setWaveOutEnables.bind(this);
    this.getWaveSRCEnabled = this.getWaveSRCEnabled.bind(this);
    this.setWaveSRCEnabled = this.setWaveSRCEnabled.bind(this);

    if (typeof require !== 'undefined') {
      this.electron = require('electron').ipcRenderer;
      this.lib = require(__dirname + '/../../build/Release/mseries');
      this.lib.register(this.libCallback);
      var that = this;
      ds.addListener('kHostBuffer1x', function (ev, vals, setByObject) { 
        if (setByObject === that)
          return;
        var cs = that.lib.getCurrentState(kDefaultG5Dev);
        // Driver buffer size is for current mult, convert from 1x
        if((cs !== undefined) && (vals[0] != cs.hostBufferSizeInSamples))
          that.lib.setHostBufferSize(kDefaultG5Dev, parseInt(vals[0]) * getSampleRateMult(cs.currentSampleRateInHertz)); 
      });
      ds.addListener('kWaveInEnables', function (ev, vals, setByObject) { 
        if (setByObject === that)
          return;
        var cs = that.lib.getCurrentState(kDefaultG5Dev);
        if((cs !== undefined) && (vals[0] != cs.waveInEnables))
          that.lib.setWaveInEnables(kDefaultG5Dev,parseInt(vals[0])); 
      });
      ds.addListener('kWaveOutEnables', function (ev, vals, setByObject) { 
        if (setByObject === that)
          return;
        var cs = that.lib.getCurrentState(kDefaultG5Dev);
        if((cs !== undefined) && (vals[0] != cs.waveOutEnables))
          that.lib.setWaveOutEnables(kDefaultG5Dev,parseInt(vals[0])); 
      });
      ds.addListener('kExtraSafetyOffset1X', function (ev, vals, setByObject) { 
        if (setByObject === that)
          return;
        var cs = that.lib.getCurrentState(kDefaultG5Dev);
        if((cs !== undefined) && (vals[0] != cs.xso1x))
          that.lib.setExtraSafetyOffset1X(kDefaultG5Dev, parseInt(vals[0])); 
      });
      ds.addListener('kWaveSRCEnabled', function(ev, vals, setByObject) {
        if (setByObject === that)
          return;;
        var cs = that.lib.getWaveSRCEnabled(kDefaultG5Dev);
        if ((cs !== undefined) && (vals[0] != cs.waveSRCEnabled))
          that.lib.setWaveSRCEnabled(kDefaultG5Dev, parseInt(vals[0]));
      });
    }
  }

  libCallback(arg1, arg2, arg3) {
    if (arg1 == "dllNotify") {
      var n = arg2;
      var d = arg3;
      if (n == (1 << 2)) {
        // audio engine error - diagnostics
        ev.dispatchSync('kWinAEErr', arg3);
      }
      else {
        // parameter or device pnp
        var num_devices = this.lib.getNumDevices();
        if (num_devices > 0) {
          this.synchDSToCurrentState();
        }
      }
    }
  }
  synchDSToCurrentState() {
    var cs = this.lib.getCurrentState(kDefaultG5Dev);
    if(cs !== undefined) {
      // Driver buffer size is for current mult, convert to 1x
      ds.setV('kHostBuffer1x', 0, cs.hostBufferSizeInSamples / getSampleRateMult(cs.currentSampleRateInHertz), this);
      ds.setV('kWaveInEnables', 0, cs.waveInEnables, this);
      ds.setV('kWaveOutEnables', 0, cs.waveOutEnables, this);
      ds.setV('kExtraSafetyOffset1X', 0, cs.xso1x, this); // 0 or 32
    }

    var ans = this.lib.getWaveSRCEnabled(kDefaultG5Dev);
    if (ans !== undefined) {
      ds.setV('kWaveSRCEnabled', 0, ans.waveSRCEnabled);
    }
  }
  hasLib() {
    return this.lib !== undefined;
  }
  hasDevice() {
    return this.lib.getNumDevices();
  }
  postMessage(msg) {
    // send cmd to dll updater
    if(msg == 'open')
      this.electron && this.electron.send('open-firmware-updater');
    else if (msg == 'driver')
      this.electron && this.electron.send('open-driver-download');
    else if (msg == 'loadPreset')
      this.electron && this.electron.send('load-preset');
    else if(msg.indexOf('savePreset&') == 0)
      this.electron && this.electron.send('save-preset', msg.substring(11));
  }
  getDiagnostics() {
    return this.lib.getDiagnostics(kDefaultG5Dev);
  }
  getDeviceInfo() {
    return this.lib.getDeviceInfo(kDefaultG5Dev);
  }
  getDriverVersion() {
    return this.lib.getDriverVersion(kDefaultG5Dev);
  }
  getCurrentState() {
    return this.lib.getCurrentState(kDefaultG5Dev);
  }
  setHostBufferSize(hbs) {
    this.lib.setHostBufferSize(kDefaultG5Dev, hbs);
  }
  setExtraSafetyOffset1X(xso1x) {
    this.lib.setExtraSafetyOffset1X(kDefaultG5Dev, xso1x);
  }
  setWaveInEnables(wie) {
    this.lib.setWaveInEnables(kDefaultG5Dev, wie);
  }
  setWaveOutEnables(woe) {
    this.lib.setWaveOutEnables(kDefaultG5Dev, woe);
  }
  getWaveSRCEnabled() {
    return this.lib.getWaveSRCEnabled(kDefaultG5Dev);
  }
  setWaveSRCEnabled(enabled) {
    this.lib.setWaveSRCEnabled(kDefaultG5Dev, enabled);
  }
}

class Device {
  constructor() {
    this.wsConnect = this.wsConnect.bind(this);
    this.wsOpen = this.wsOpen.bind(this);
    this.wsError = this.wsError.bind(this);
    this.wsClose = this.wsClose.bind(this);
    this.wsRx = this.wsRx.bind(this);
    //this.rxUDPAnnounce = this.rxUDPAnnounce.bind(this);
    this.isConnected = this.isConnected.bind(this);
    this.isTest = this.isTest.bind(this);
    this.tryToReconnect = this.tryToReconnect.bind(this);
    this.checkConnection = this.checkConnection.bind(this);
    this.hasHostMsgPipe = this.hasHostMsgPipe.bind(this);
    this.receiveMsg = this.receiveMsg.bind(this);
    this.proxyStatusChanged = this.proxyStatusChanged.bind(this);

    evSys.addListener('ProxyStatusChanged', this.proxyStatusChanged);

    this.test = false;
    this.opened = false;
    
    this.lastMsgTime = 0;

    // see if there is a mac or windows host msg handler
    this.jsHostMsgPipe = undefined;
    if((window.webkit != undefined) && (window.webkit.messageHandlers != undefined)){
      this.wsIPAddr = 'localhost:1281'; // Connect to the proxy server
      this.jsHostMsgPipe = window.webkit.messageHandlers.msgJStoObjC;
      setTimeout(this.tryToReconnect, 0);
    }
    else
    {
      //this.wsIPAddr = storage.getItem("LastWSIPAddr");
      this.timerID = setInterval(this.checkConnection, 100);
      if(gen5Lib.hasLib()) {
        this.jsHostMsgPipe = gen5Lib;
      }
    }

    //setTimeout(this.tryToReconnect, 0);
  }
  hasHostMsgPipe() { 
    return this.jsHostMsgPipe != undefined;
  }
  // Sends JSON to updater running in objectiveC/C++ or electron
  sendHostMsgPipeMsg(msg) {
    if(this.jsHostMsgPipe != undefined)
      this.jsHostMsgPipe.postMessage(msg);
  }
  receiveMsg(msg) {
    var jsonMsg = JSON.parse(msg);
    if(jsonMsg.driverVersion) {
      evSys.dispatchASync('driverVersion', jsonMsg.driverVersion);
    }
    else if(jsonMsg.applyPreset != undefined) {
      ds.applyPreset(jsonMsg.applyPreset);
    }
    else if((updater != undefined) && updater.isAvailable()) {
      updater.receiveMsg(jsonMsg);
    }
  } 
  tryToReconnect() {
    this.wsConnect(this.wsIPAddr);
  }
  wsRx(d) {
    this.lastMessageTime = perfSeconds();
    ds.recv(d);
  }
  wsConnect(ip){
    if((ip == "Test") || (ip == "test")) {
      this.test = true;
      this.opened = true;
      this.wsIPAddr = ip;
      evSys.dispatchASync('DeviceStatusChanged');
    }
    else if((ip != undefined) && ((this.ws == undefined) || (ip != this.wsIPAddr))) {
      //console.log('wsConnect');
      this.test = false;
      this.wsIPAddr = ip;
      if (ip.indexOf(':') === -1)
        ip = ip + ':1280';
      this.lastMessageTime = perfSeconds();
      this.ws = new WebSocket('ws://'+ip);
      this.ws.binaryType = 'arraybuffer';
      this.ws.onmessage = this.wsRx;
      this.ws.onopen = this.wsOpen;
      this.ws.onerror = this.wsError;
      this.ws.onclose = this.wsClose;
    }
  }
  wsError(){
    //console.log('wsError');
    //this.wsClose();
  }
  wsOpen(){
    this.opened = true;
    storage.setItem("LastWSIPAddr", this.wsIPAddr);
    evSys.dispatchASync('DeviceStatusChanged');
  }
  wsClose(){
    if(this.ws != undefined) {
      //console.log('wsClose');
      if(this.opened) {
        ds.reset();
        evSys.dispatchASync('DeviceStatusChanged');
      }
      this.opened = false;
      this.ws.close();
      this.ws = undefined;
      setTimeout(this.tryToReconnect, 1000);
    }
  }
  checkConnection() {
    if(!this.test) {
      var tnow = perfSeconds();
      if(((perfSeconds() - this.lastMessageTime) > 2.0) && (this.ws != undefined)) {
        //console.log('checkConnection closing connection');
        this.wsClose();
      }
    }
  }
  disconnect() {
    if(this.opened) {
      ds.reset();
      evSys.dispatchASync('DeviceStatusChanged');
    }
    if (this.ws != undefined) {
      this.ws.close();
      this.ws = undefined;
    }
    else {
      this.opened = false;
    }
  }
  rxUDPAnnounce(msg) {
    var jsonMsg = JSON.parse(msg);
    // We only support these devices
    if(jsonMsg.model != "UltraLite-mk5")
      return;

    if(this.isConnected() && !this.test)
      return;

    this.test = false;
    this.wsConnect(jsonMsg.ip);
  }
  isConnected() {
    return this.opened && (this.ws != undefined) && (this.ws.readyState == 1);
  }
  isTest() {
    return this.test;
  }
  send(d) {
    if(!this.test && this.isConnected()) {
      this.ws.send(d);
    }
  }
  proxyStatusChanged(ev, v) {
    if (v == 0)
      ds.reset();
  }
}
var decoder = new TextDecoder('utf8');
var encoder = new TextEncoder('utf8');

function CreateDeviceMessage(kv, i) {
// todo check endianess of host and always send data network (big) endian
  var x = undefined;
  switch(kv.t) {
    default:
      break;
    case kTString: {
      // Note we don't support writing strings bigger than 32 bytes
      x = new Uint8Array(6 + 32); // initialized to 0
      var str = encoder.encode(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = 0;
      x[5]  = (32 & 0x00FF);
      // always send the device a null terminated 32 byte array
      var len = Math.min(str.length, 31);
      for(var i = 0; i < len; ++i) {
        x[6+i] = str[i];
      }
    } break;
    case kTInt32: {
      var len = 4;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0xFF000000) >> 24;
      x[7]  = (val & 0x00FF0000) >> 16;
      x[8]  = (val & 0x0000FF00) >> 8;
      x[9]  = (val & 0x000000FF);
    } break;
    case kTFloat: {
      var len = 4;
      x = new Uint8Array(6 + len);
      var val = To824(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0xFF000000) >> 24;
      x[7]  = (val & 0x00FF0000) >> 16;
      x[8]  = (val & 0x0000FF00) >> 8;
      x[9]  = (val & 0x000000FF);
    } break;
    case kTInt16: {
      var len = 2;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0x0000FF00) >> 8;
      x[7]  = (val & 0x000000FF);
    } break;
    case kTByte: {
      var len = 1;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0x000000FF);
    } break;
  }
  return x;
}
function addUint8Arrays(a, b) {
  // return new array of b appended to a
  if(a.length === 0) return b;
  var c = new Int8Array(a.length + b.length);
  c.set(a);
  c.set(b, a.length);
  return c;
}

function UnpackDeviceMessage(t, msg) {
// todo check endianess of host and always send data network (big) endian
  switch(t) {
    default:
      break;
    case kTArray: {
      if(msg.length >= 5)
        return Array.from(msg.slice(4));
    } break;
    case kTString: {
      if(msg.length >= 5)
        return decoder.decode(msg.slice(4, msg.indexOf(0,4)));
    } break;
    case kTInt32: {
      if(msg.length >= 8)
        return (msg[4] << 24) | (msg[5] << 16) | (msg[6] << 8) | msg[7];
    } break;
    case kTFloat: {
      if(msg.length >= 8)
        return From824((msg[4] << 24) | (msg[5] << 16) | (msg[6] << 8) | msg[7]);
    } break;
    case kTInt16: {
      if(msg.length >= 6)
        return (msg[4] << 8) | msg[5];
    } break;
    case kTByte: {
      if(msg.length >= 5)
        return msg[4];
    } break;
  }
  return undefined;
}

// all items in this datastore are arrays, single items are arrays of 1
// Terminology:
// k: the key to the datastore object
// kv: the datastore object (id, type, value array)
// kv.id: the id of the value array (numeric versoin of k)
// kv.t: the type of the items in the value array
// kv.v: the value array
class DataStore extends Events{
  constructor() {
    super();
    this.reset = this.reset.bind(this);
    this.getKVArray = this.getKVArray.bind(this);
    this.getKVArrayItem = this.getKVArrayItem.bind(this);
    this.setV = this.setV.bind(this);
    this.sendPendingToDevice = this.sendPendingToDevice.bind(this);
    this.recv = this.recv.bind(this);
    this.loadPreset = this.loadPreset.bind(this);
    this.savePreset = this.savePreset.bind(this);
    this.ds = {};
    this.map = [];
    this.devOutData = new Uint8Array();
    this.reset(); 

  }
  loadPreset() {
    // tell the host dll/app to run the open dialog and load a preset, it will send it to applyPreset
    if(device.hasHostMsgPipe()) {
      device.sendHostMsgPipeMsg("loadPreset");
    }
  }
  applyPreset(preset) {
    var devJS = JSON.parse(preset);
    // verify the device is correct
    if((devJS.kDevName != undefined) && 
      (devJS.kDevName.v[0] == this.ds.kDevName.v[0]) &&
      (devJS.kAPIVersion != undefined)) 
    {
      for (let [k, presetKV] of Object.entries(devJS)) {
        // Does the key exist and is it writable by cuemix?
        if(this.ds[k] != undefined) {
          var kv = this.ds[k];
          // ds keys are writable unless specifed
          if((kv.w == undefined) || kv.w) {

            // ID's -1000 through -1999 are from the windows dll
            if(kv.id < 0) {
              if(kv.id > -1000) {
                // send to local storage
                kv.v = presetKV.v;
                this.dispatchASync(k, kv.v);
                storage.setItem(k, JSON.stringify(kv.v));
              }
              // ID's -1000 through -1999 are from the windows dll and I'm not restoring them as presets
            }
            else {
              // send to device
              var changed = false;
              var n = kv.v.length;
              for(var i = 0; i < n; ++i) {
                if(kv.v[i] != presetKV.v[i]) {
                  changed = true;
                  kv.v[i] = presetKV.v[i];
                  var msg = CreateDeviceMessage(kv, i);
                  if(msg != undefined) {
                    if((this.devOutData.length + msg.length) >= 1452) 
                      this.sendPendingToDevice();
                    this.devOutData = addUint8Arrays(this.devOutData, msg);
                    setTimeout(this.sendPendingToDevice, 0);
                  }
                }
              }
              if(changed)
                this.dispatchASync(k, kv.v);
            }
          }
        }
      }
    }
    else {
      // todo: put up a message
    }
  }
  savePreset(t) {
    if(device.hasHostMsgPipe()) {
      // make a copy of the ds with just key names and values
      var preset = {};
      for (let [k, kv] of Object.entries(this.ds)) {
        preset[k] = { v: kv.v };
      }
      device.sendHostMsgPipeMsg("savePreset&" + JSON.stringify(preset));
    }
  }

  reset() {
    this.ds = {};
    this.map = [];
    // this set's the ds to a copy of kDevJS
    this.ds = JSON.parse(JSON.stringify(kDevJS));
    for (let [k, kv] of Object.entries(this.ds)) {
      if(this.map[kv.id] != undefined)
        throw 'PropID is not unique';
      this.map[kv.id] = k;
      // restore negative ID's from storage
      if(kv.id < 0) {
        if(kv.id > -1000) {
          var t = storage.getItem(k);
          if(t != undefined) 
            this.ds[k].v = JSON.parse(t);
        }
        // ID's -1000 through -1999 are from the windows dll
      }
      this.dispatchASync(k, kv.v);
    }
    if(gen5Lib && gen5Lib.hasLib() && gen5Lib.hasDevice())
      gen5Lib.synchDSToCurrentState();
  }
  getKVArray(k){ // gets the entire key value array
    return this.ds[k].v;
  }
  getKVArrayItem(k,i){ //gets the single value of the key at index
    return this.ds[k].v[i];
  }
  setV(k, i, v, setByObject){
    // update the value, notify listeners and send to the device
    //console.log("DS Set " + k + ", i: " + i + ", v " + v);
    var kv = this.ds[k];

    // Is the key writable by cuemix?
    if(kv.w != undefined && !kv.w)
      return;

    // round integral types
    switch(kv.t) {
      default:
        break;
      case kTInt32: 
      case kTInt16: 
      case kTByte: 
        v = Math.round(v);
      break;
    }
    if(kv.v[i] != v) {
      kv.v[i] = v;
      this.dispatchASync(k, kv.v, setByObject);
      if(kv.id < 0) {
        if(kv.id > -1000) {
          // send to local storage
          storage.setItem(k, JSON.stringify(kv.v));
        }
        // ID's -1000 through -1999 are from the windows dll
      }
      else {
        // send to device
        var msg = CreateDeviceMessage(kv, i);
        if(msg != undefined) {
          if((this.devOutData.length + msg.length) >= 1452) 
            this.sendPendingToDevice();
          this.devOutData = addUint8Arrays(this.devOutData, msg);
          setTimeout(this.sendPendingToDevice, 0);
        }
      }
    }
  }
  sendPendingToDevice() {
    device.send(this.devOutData);    
    this.devOutData = new Uint8Array();
  }
  recv(e) {
    var data = new Uint8Array(e.data);
    var id = (data[0] << 8) | data[1];
    var i = (data[2] << 8) | data[3];
    if (id == kMetersID)
      meters.set(data.slice(4));
    else if (id == kProxyStatusID)
      evSys.dispatchASync('ProxyStatusChanged', data[4]);
    else {
      var k = this.map[id];
      if(k != undefined) {
        var kv = this.ds[k];
        var v = UnpackDeviceMessage(kv.t, data);
        if((v != undefined) && (kv.v[i] != v)) {
          kv.v[i] = v;
          this.dispatchASync(k, kv.v);
        }
      }
    }
  }
}
// Client DS helpers
function SetupDSArrayListener(that, k, i) {
  // add the key as an object to that
  that[k] = i;
  ds.addListener(k, function (ev, vals, setByObject) {
    var v = vals[that[k]];
    that.onDSArray(ev, v);
  });
}
function SetDSArrayIndex(that, k, i) {
    that[k] = i;
}
function GetDSArrayIndex(that, k) {
    var i = that[k];
    return i;
}
function DSArrayGet(that, k) {
  return ds.getKVArrayItem(k, that[k]);
}
function DSArraySet(that, k, v) {
  var i = that[k];
  if(i != undefined) 
    ds.setV(k, i, v);
}
// for objects with only 1 dslistener
function SetupDSArrayListenerBasic(that, k, i) {
  that.dsKey = k;
  SetupDSArrayListener(that, k, i);
}
function SetDSArrayIndexBasic(that, i) {
  SetDSArrayIndex(that, that.dsKey, i);
}
function GetDSArrayIndexBasic(that) {
  return GetDSArrayIndex(that, that.dsKey);
}
function DSArrayGetBasic(that) {
  return DSArrayGet(that, that.dsKey);
}
function DSArraySetBasic(that, v) {
  DSArraySet(that, that.dsKey, v);
}
