// type: type/name of bank
// iGain: index of first channel mix gain at sample rates 
// numCh: number of mix input channels (and bus channels) at sample rates
// name: index of first mix input channel name
// stereo: index of first mix input channel stereo control
// pre: index of ffirst mix input channel pre controls (48v, pad trim)
// eq: index of first mix input channel eq controls 
// gate: index of first mix input channel gate controls 
// comp: index of first mix input channel comp controls 
// iBus: index of first mix bus gain at sample rates
// name: index of first mix bus name
// stereo: index of first mix input channel stereo control
// busEq: index of first mix bus eq controls 


// PatchBase (from dylans patchbase.h)
const kPB_TDM_BASE    = 0;  //x0
const kPB_ZERO_BASE   = 23; //x17
const kPB_ADAT_BASE   = 24; //x18
const kPB_I2S_BASE    = 32; //x20
const kPB_SPDF_BASE   = 44; //x2c
const kPB_DSP_BASE    = 46; //x2e

// LocalInputs  (from dylans patchbase.h)
const kLI_Mic1   = 5;
const kLI_Mic2   = 0;
const kLI_Line3  = 6;
const kLI_Line4  = 7;
const kLI_Line5  = 3;
const kLI_Line6  = 4;
const kLI_Line7  = 1;
const kLI_Line8  = 2;

// LocalOutputs  (from dylans patchbase.h)
// Also BS Mapping for DAC trims, main group and meters, also meters are pre trim
const kLO_Main1  = 0x6;
const kLO_Main2  = 0x7;
const kLO_Line3  = 0x8;
const kLO_Line4  = 0x9;
const kLO_Line5  = 0x0;
const kLO_Line6  = 0x1;
const kLO_Line7  = 0x2;
const kLO_Line8  = 0x3;
const kLO_Line9  = 0x4;
const kLO_Line10  = 0x5;
const kLO_Phones1 = 0xA;
const kLO_Phones2 = 0xB;
const kLO_SPDIF1  = 0xC;
const kLO_SPDIF2  = 0xD;

// InputSource  (from dylans patchbase.h)
const kIS_TDM1 = (kPB_TDM_BASE + 0x00);
const kIS_TDM2 = (kPB_TDM_BASE + 0x01);
const kIS_TDM3 = (kPB_TDM_BASE + 0x02);
const kIS_TDM4 = (kPB_TDM_BASE + 0x03);
const kIS_TDM5 = (kPB_TDM_BASE + 0x04);
const kIS_TDM6 = (kPB_TDM_BASE + 0x05);
const kIS_TDM7 = (kPB_TDM_BASE + 0x06);
const kIS_TDM8 = (kPB_TDM_BASE + 0x07);
const kIS_TDM9 = (kPB_TDM_BASE + 0x08);
const kIS_TDM10 = (kPB_TDM_BASE + 0x09);
const kIS_TDM11 = (kPB_TDM_BASE + 0x0A);
const kIS_TDM12 = (kPB_TDM_BASE + 0x0B);
const kIS_TDM13 = (kPB_TDM_BASE + 0x0C); // USB Out SPDIF 1
const kIS_TDM14 = (kPB_TDM_BASE + 0x0D); // USB Out SPDIF 2
const kIS_TDM15 = (kPB_TDM_BASE + 0x0E); // USB Out Optical 1
const kIS_TDM16 = (kPB_TDM_BASE + 0x0F); // USB Out Optical 2
const kIS_TDM17 = (kPB_TDM_BASE + 0x10); // USB Out Optical 3
const kIS_TDM18 = (kPB_TDM_BASE + 0x11); // USB Out Optical 4
const kIS_TDM19 = (kPB_TDM_BASE + 0x12); // USB Out Optical 5
const kIS_TDM20 = (kPB_TDM_BASE + 0x13); // USB Out Optical 6
const kIS_TDM21 = (kPB_TDM_BASE + 0x14); // USB Out Optical 7
const kIS_TDM22 = (kPB_TDM_BASE + 0x15); // USB Out Optical 8
const kIS_TDM23 = (kPB_TDM_BASE + 0x16);
const kIS_Zeros = kPB_ZERO_BASE;
const kIS_ADAT1 = (kPB_ADAT_BASE + 0);
const kIS_ADAT2 = (kPB_ADAT_BASE + 1);
const kIS_ADAT3 = (kPB_ADAT_BASE + 2);
const kIS_ADAT4 = (kPB_ADAT_BASE + 3);
const kIS_ADAT5 = (kPB_ADAT_BASE + 4);
const kIS_ADAT6 = (kPB_ADAT_BASE + 5);
const kIS_ADAT7 = (kPB_ADAT_BASE + 6);
const kIS_ADAT8 = (kPB_ADAT_BASE + 7);
const kIS_Mic1 = (kPB_I2S_BASE + kLI_Mic1);
const kIS_Mic2 = (kPB_I2S_BASE + kLI_Mic2);
const kIS_Line3  = (kPB_I2S_BASE + kLI_Line3);
const kIS_Line4  = (kPB_I2S_BASE + kLI_Line4);
const kIS_Line5  = (kPB_I2S_BASE + kLI_Line5);
const kIS_Line6  = (kPB_I2S_BASE + kLI_Line6);
const kIS_Line7  = (kPB_I2S_BASE + kLI_Line7);
const kIS_Line8  = (kPB_I2S_BASE + kLI_Line8);
const kIS_SPDIF1 = (kPB_SPDF_BASE + 0);
const kIS_SPDIF2 = (kPB_SPDF_BASE + 1);
const kIS_DSP1   = (kPB_DSP_BASE + 0); // Main 1 Mix
const kIS_DSP2   = (kPB_DSP_BASE + 1); // Main 2 Mix 
const kIS_DSP3   = (kPB_DSP_BASE + 2); // Line 3 Mix
const kIS_DSP4   = (kPB_DSP_BASE + 3); // Line 4 Mix
const kIS_DSP5   = (kPB_DSP_BASE + 4);
const kIS_DSP6   = (kPB_DSP_BASE + 5);
const kIS_DSP7   = (kPB_DSP_BASE + 6);
const kIS_DSP8   = (kPB_DSP_BASE + 7);
const kIS_DSP9   = (kPB_DSP_BASE + 8); // Line 5 Mix
const kIS_DSP10  = (kPB_DSP_BASE + 9); // Line 6 Mix
const kIS_DSP11  = (kPB_DSP_BASE + 10); // Line 7 Mix
const kIS_DSP12  = (kPB_DSP_BASE + 11); // Line 8 Mix
const kIS_DSP13  = (kPB_DSP_BASE + 12);
const kIS_DSP14  = (kPB_DSP_BASE + 13);
const kIS_DSP15  = (kPB_DSP_BASE + 14);
const kIS_DSP16  = (kPB_DSP_BASE + 15);
const kIS_DSP17  = (kPB_DSP_BASE + 16); // Line 9 Mix
const kIS_DSP18  = (kPB_DSP_BASE + 17); // Line 10 Mix
const kIS_DSP19  = (kPB_DSP_BASE + 18);
const kIS_DSP20  = (kPB_DSP_BASE + 19);
const kIS_DSP21  = (kPB_DSP_BASE + 20); // Reverb L Dry
const kIS_DSP22  = (kPB_DSP_BASE + 21); // Reverb R Dry
const kIS_DSP23  = (kPB_DSP_BASE + 22); // Reverb L Wet
const kIS_DSP24  = (kPB_DSP_BASE + 23); // Reverb R Wet
const kIS_DSP25  = (kPB_DSP_BASE + 24);
const kIS_DSP26  = (kPB_DSP_BASE + 25);
const kIS_DSP27  = (kPB_DSP_BASE + 26);
const kIS_DSP28  = (kPB_DSP_BASE + 27);
const kIS_DSP29  = (kPB_DSP_BASE + 28);
const kIS_DSP30  = (kPB_DSP_BASE + 29);
const kIS_DSP31  = (kPB_DSP_BASE + 30);
const kIS_DSP32  = (kPB_DSP_BASE + 31);

// OutputDestination (from dylans patchbase.h)
const kOD_TDM1 = (kPB_TDM_BASE + 0x00);
const kOD_TDM2 = (kPB_TDM_BASE + 0x01);
const kOD_TDM3 = (kPB_TDM_BASE + 0x02);
const kOD_TDM4 = (kPB_TDM_BASE + 0x03);
const kOD_TDM5 = (kPB_TDM_BASE + 0x04);
const kOD_TDM6 = (kPB_TDM_BASE + 0x05);
const kOD_TDM7 = (kPB_TDM_BASE + 0x06);
const kOD_TDM8 = (kPB_TDM_BASE + 0x07);
const kOD_TDM9 = (kPB_TDM_BASE + 0x08);
const kOD_TDM10 = (kPB_TDM_BASE + 0x09);
const kOD_TDM11 = (kPB_TDM_BASE + 0x0A);
const kOD_TDM12 = (kPB_TDM_BASE + 0x0B);
const kOD_TDM13 = (kPB_TDM_BASE + 0x0C);
const kOD_TDM14 = (kPB_TDM_BASE + 0x0D);
const kOD_TDM15 = (kPB_TDM_BASE + 0x0E);
const kOD_TDM16 = (kPB_TDM_BASE + 0x0F);
const kOD_TDM17 = (kPB_TDM_BASE + 0x10);
const kOD_TDM18 = (kPB_TDM_BASE + 0x11);
const kOD_TDM19 = (kPB_TDM_BASE + 0x12);
const kOD_TDM20 = (kPB_TDM_BASE + 0x13);
const kOD_TDM21 = (kPB_TDM_BASE + 0x14);
const kOD_TDM22 = (kPB_TDM_BASE + 0x15);
const kOD_TDM23 = (kPB_TDM_BASE + 0x16);
const kOD_Zeros = kPB_ZERO_BASE;
const kOD_ADAT1 = (kPB_ADAT_BASE + 0);
const kOD_ADAT2 = (kPB_ADAT_BASE + 1);
const kOD_ADAT3 = (kPB_ADAT_BASE + 2);
const kOD_ADAT4 = (kPB_ADAT_BASE + 3);
const kOD_ADAT5 = (kPB_ADAT_BASE + 4);
const kOD_ADAT6 = (kPB_ADAT_BASE + 5);
const kOD_ADAT7 = (kPB_ADAT_BASE + 6);
const kOD_ADAT8 = (kPB_ADAT_BASE + 7);
const kOD_Main1   = (kPB_I2S_BASE + kLO_Main1);
const kOD_Main2   = (kPB_I2S_BASE + kLO_Main2);
const kOD_Line3   = (kPB_I2S_BASE + kLO_Line3);
const kOD_Line4   = (kPB_I2S_BASE + kLO_Line4);
const kOD_Line5   = (kPB_I2S_BASE + kLO_Line5);
const kOD_Line6   = (kPB_I2S_BASE + kLO_Line6);
const kOD_Line7   = (kPB_I2S_BASE + kLO_Line7);
const kOD_Line8   = (kPB_I2S_BASE + kLO_Line8);
const kOD_Line9   = (kPB_I2S_BASE + kLO_Line9);
const kOD_Line10  = (kPB_I2S_BASE + kLO_Line10);
const kOD_Phones1 = (kPB_I2S_BASE + kLO_Phones1);
const kOD_Phones2 = (kPB_I2S_BASE + kLO_Phones2);
const kOD_SPDIF1 = (kPB_SPDF_BASE + 0);
const kOD_SPDIF2 = (kPB_SPDF_BASE + 1);
const kOD_DSP1   = (kPB_DSP_BASE + 0);
const kOD_DSP2   = (kPB_DSP_BASE + 1);
const kOD_DSP3   = (kPB_DSP_BASE + 2);
const kOD_DSP4   = (kPB_DSP_BASE + 3);
const kOD_DSP5   = (kPB_DSP_BASE + 4);
const kOD_DSP6   = (kPB_DSP_BASE + 5);
const kOD_DSP7   = (kPB_DSP_BASE + 6);
const kOD_DSP8   = (kPB_DSP_BASE + 7);
const kOD_DSP9   = (kPB_DSP_BASE + 8);
const kOD_DSP10  = (kPB_DSP_BASE + 9);
const kOD_DSP11  = (kPB_DSP_BASE + 10);
const kOD_DSP12  = (kPB_DSP_BASE + 11);
const kOD_DSP13  = (kPB_DSP_BASE + 12);
const kOD_DSP14  = (kPB_DSP_BASE + 13);
const kOD_DSP15  = (kPB_DSP_BASE + 14);
const kOD_DSP16  = (kPB_DSP_BASE + 15);
const kOD_DSP17  = (kPB_DSP_BASE + 16);
const kOD_DSP18  = (kPB_DSP_BASE + 17);
const kOD_DSP19  = (kPB_DSP_BASE + 18);
const kOD_DSP20  = (kPB_DSP_BASE + 19);
const kOD_DSP21  = (kPB_DSP_BASE + 20);
const kOD_DSP22  = (kPB_DSP_BASE + 21);
const kOD_DSP23  = (kPB_DSP_BASE + 22);
const kOD_DSP24  = (kPB_DSP_BASE + 23);
const kOD_DSP25  = (kPB_DSP_BASE + 24);
const kOD_DSP26  = (kPB_DSP_BASE + 25);
const kOD_DSP27  = (kPB_DSP_BASE + 26);
const kOD_DSP28  = (kPB_DSP_BASE + 27);
const kOD_DSP29  = (kPB_DSP_BASE + 28);
const kOD_DSP30  = (kPB_DSP_BASE + 29);
const kOD_DSP31  = (kPB_DSP_BASE + 30);
const kOD_DSP32  = (kPB_DSP_BASE + 31);

// These are duplicates of what's defined above in a format that's more friendly for us
// These are the meter indexes and fpga routing id's
const miUSBMainOut = kIS_TDM1;
const miUSBAnalOut = kIS_TDM3;
const miUSBPhOut = kIS_TDM11;
const miUSBSPDIFOut = kIS_TDM13;
const miUSBADATOut = kIS_TDM14;
//const miMicADCIn = XXX; todo use these values for the adc meters 
// mapping for meters
const miADCInMic1   = kIS_Mic1;
const miADCInMic2   = kIS_Mic2;
const miADCInLine3  = kIS_Line3;
const miADCInLine4  = kIS_Line4;
const miADCInLine5  = kIS_Line5;
const miADCInLine6  = kIS_Line6;
const miADCInLine7  = kIS_Line7;
const miADCInLine8  = kIS_Line8;

const miADATIn = kIS_ADAT1;
const miSPDIFIn = kIS_SPDIF1;
const miMixMainOut = kIS_DSP1;
const miMixAnaOut = [kIS_DSP3,kIS_DSP4,kIS_DSP9,kIS_DSP10,kIS_DSP11,kIS_DSP12,kIS_DSP17,kIS_DSP18];
const miMixPhOut = kIS_DSP19;
const miMixReverb = kIS_DSP23;
const miPostFXMicIn = kIS_DSP5;
const miPostFXAnaIn = [kIS_DSP7,kIS_DSP8,kIS_DSP13,kIS_DSP14,kIS_DSP15,kIS_DSP16];
// NOTE: meterstore.js is counting on the gate and comp meters being contiguous
// if you insert anything here make sure it doesn't break meterstore.js
// I gave up on converting these to dylans map for now
const miPostEQMicIn = 80;
const miPostEQMicInEnd = miPostEQMicIn + 2;
const miPostEQAnaIn = 82;
const miPostEQAnaInEnd = miPostEQAnaIn + 6;
const miGateGRMicIn = 88;
const miGateGRMicInEnd = miGateGRMicIn + 2;
const miPostGateMicIn = 90;
const miPostGateMicInEnd = miPostGateMicIn + 2;
const miCompGRMicIn = 92;
const miCompGRMicInEnd = miCompGRMicIn + 2;
const miCompGRAnaIn = 94;
const miCompGRAnaInEnd = miCompGRMicIn + 6;


const MixChInfo = [
  {
    type:"Mic",
    numCh:[2,2,2,2,2,2],
    iGain:[0,0,0,0,0,0],
    stereo:0,
    pre:0,
    eq:0,
    gate:0,
    comp:0,
    name:0,
    meters:[miPostEQMicIn,miGateGRMicIn,miPostGateMicIn,miCompGRMicIn,miPostFXMicIn],
  },
  {
    type:"Line",
    numCh:[6,6,6,6,6,6],
    iGain:[2,2,2,2,2,2],
    stereo:2,
    eq:2,
    comp:2,
    name:2,
    meters:[-1,-1,miPostEQAnaIn,miCompGRAnaIn,miPostFXAnaIn],
  },
  {
    type:"S/PDIF",
    numCh:[2,2,2,2,0,0],
    iGain:[8,8,8,8,0,0],
    stereo:8,
    name:8,
    meters:[-1,-1,-1,-1,miSPDIFIn],
  },
  {
    type:"ADAT",
    numCh:[8,8,4,4,0,0],
    iGain:[10,10,10,10,0,0],
    stereo:10,
    name:10,
    meters:[-1,-1,-1,-1,miADATIn],
  },
  {
    type:"TOSlink",
    numCh:[2,2,2,2,0,0],
    iGain:[10,10,10,10,0,0],
    stereo:10,
    name:18,
    meters:[-1,-1,-1,-1,miADATIn],
  },
  {
    type:"Reverb",
    numCh:[2,2,2,2,0,0],
    iGain:[30,30,30,30,30,30],
    name:20,
    meters:[-1,-1,-1,-1,miMixReverb],
  },
];

// Bus index's
// 0: main 1
// 1: main 2
// 2: Line 1
// 3: Line 2
// 4: Line 3
// 5: Line 4
// 6: Line 5
// 7: Line 6
// 8: Line 7
// 9: Line 8
// 10: phones 1
// 11: phones 2
// 12: mix return 1
// 13: mix return 2
// 14: reverb 1
// 15: reverb 2
const MixBusInfo = [
  {
    type:"Main",
    numCh:[2,2,2,2,2,2],
    iBus:[0,0,0,0,0,0],
    iGain:[18,18,18,18,18,18],
    busEq:0,
    name:0,
    istereo:18, // stereo index of host output to mixer
    meters:[miMixMainOut, miUSBMainOut],
  },
  {
    type:"Phones",
    numCh:[2,2,2,2,2,2],
    iBus:[10,10,10,10,10,10],
    iGain:[28,28,28,28,28,28],
    busEq:10,
    name:10,
    istereo:28, // stereo index of host output to mixer
    meters:[miMixPhOut, miUSBPhOut],
  },
  {
    type:"Line",
    numCh:[8,8,8,8,8,8],
    iBus:[2,2,2,2,2,2],
    iGain:[20,20,20,20,20,20],
    busEq:2,
    name:2,
    stereo:2,
    istereo:20, // stereo index of host output to mixer
    meters:[miMixAnaOut, miUSBAnalOut],
  },
  {
    type:"Reverb",
    numCh:[2,2,2,2,2,2],
    iBus:[12,12,12,12,12,12],
    busEq:11,
    meters:[miMixReverb, -1],
  },
];

 const kSampleRates = [ 
  { v:44100, l:"44100" }, 
  { v:48000, l:"48000" }, 
  { v:88200, l:"88200" }, 
  { v:96000, l:"96000" }, 
  { v:176400, l:"176400" }, 
  { v:192000, l:"192000" }
];
const kClockSources = [
  { v:3, l:"Internal" }, 
  { v:0, l:"S/PDIF" }, 
  { v:2, l:"Optical" } 
];
const kHostBuffers1x = [
  { v:16, l:"16" }, 
  { v:32, l:"32" }, 
  { v:64, l:"64" }, 
  { v:128, l:"128" }, 
  { v:256, l:"256" }, 
  { v:512, l:"512" }, 
  { v:1024, l:"1024" },
  { v:2048, l:"2048" },
  { v:4096, l:"4096"} 
];
const kHostBuffers2x = [
  { v:16, l:"16" }, 
  { v:32, l:"32" }, 
  { v:64, l:"64" }, 
  { v:128, l:"128" }, 
  { v:256, l:"256" }, 
  { v:512, l:"512" }, 
  { v:1024, l:"1024" },
  { v:2048, l: "2048"}
];
const kHostBuffers4x = [
  { v:16, l:"16" }, 
  { v:32, l:"32" }, 
  { v:64, l:"64" }, 
  { v:128, l:"128" }, 
  { v:256, l:"256" }, 
  { v:512, l:"512" }, 
  { v:1024, l:"1024" }
];
const kUIMeterFPSs = [
  { v:15, l:"Slow" }, 
  { v:30, l:"Medium" }, 
  { v:60, l:"Fast" } 
];
const kExtraSafetyOffset1X =  [
  { v:0, l:"Lowest Latency" }, 
  { v:1, l:"Default" } 
];
const kMixInputMetersMode =  [
  { v:0, l:"Pre-Fader" }, 
  { v:1, l:"Post-Fader" } 
];
const kHostOutputMixMode =  [
  { v:0, l:"Default" }, 
  { v:1, l:"Additional" } 
];
const kLoopbackOrder =  [
  { v:0, l:"USB In 9-10 (Default)" }, 
  { v:1, l:"USB In 1-2" } 
];

// Property types:
const kTString = "S"; //    String "S": v is an array of string
const kTInt32 = "I";  //    Int32 "I": v is an array of int32 
const kTFloat = "F";  //    Float "F": v is an array of float (8.24 float on device)
const kTInt16 = "I16";//    Int16 "16": v is an array of bytes 
const kTByte = "B";   //    Int8 "B": v is an array of bytes 
const kTArray = "A";  //    Array of Bytes

const kMetersID = 6000;
const kProxyStatusID = 0xFFFF;

// The version of the datastore api I support
const kDataStoreAPIVersion = 8;
const kDataStoreAPIVersionUnknown = -1;

const kDevJS = {
  // Data

  kAPIVersion: {
    id:8,
    w:false,
    t:kTInt16,
    // version of the datastore api the device supports
    v:[kDataStoreAPIVersionUnknown]
  },
  // key value pairs defining what we need to know about the device
  kDeviceMAC: {
    id:0,
    w:false,
    // Serial number is "ULM5" + lower 3 bytes, UID is MAC with FFFE in center
    t:kTArray,
    v:[[0,0,0,0,0,0]],
  },
  kDevName: {
    id:3,
    w:false,
    t:kTString,
    v:["UltraLite-mk5"]
  },
  kDevVersion: {
    id:4,
    w:false,
    t:kTString,
    v:["",""]
  },
  kSampleRate: {
    id:10,
    t:kTInt32,
    // kSampleRates
    v:[48000],
  },
  kClockSource: {
    id:11,
    t:kTByte,
    // kClockSources
    v:[0],
  },
  kOpticalMode: {
    id:5006,
    t:kTByte,
    // kOpticalModes: ADAT = 0, TOSlink = 1 
    v:[
      0, // input
      0  // output
    ],
  },
  kFPGALocked: {
    id:15,
    w:false,
    t:kTByte,
    // Bool 1 = locked, combine with PLL to determine if clock is locked
    v:[0],
  },
  kPLLLocked: {
    id:16,
    w:false,
    t:kTByte,
    // Bool 1 = locked, combine with FPGA to determine if clock is locked
    v:[0],
  },
  kLoopbackFirst: {
    id:20,
    t:kTByte,
    v:[0],
  },
  kModelID: {
    id:21,
    w:false,
    t:kTString,
    v:["ULM5"]
  },

  kJackDetect: {
    id:5005,
    w:false,
    t:kTByte,
    // bool 0 = mic, 1 = line
    v:[0,0],
  },
  k48V: {
    id:5004,
    t:kTByte,
    // bool
    v:[0,0],
  },
  kPad: {
    id:5003,
    t:kTByte,
    // bool
    v:[0,0],
  },
  kiGain: {
    id:5001,
    t:kTByte,
    // in dB
    v:[
      0,0,        // mic pres range 0-74 dB currently only up to 65
      0,0,0,0,0,0 // line in's range 0 - 20 dB
    ],
  },
  kiPhase: {
    id:5002,
    t:kTByte,
    // bool, 0 = normal, 1 = invert
    v:[
      0,0,        // mic pres 
      0,0,0,0,0,0 // line in's
    ],
  },
  kFPGAPatch: {
    id:5010,
    t:kTByte,
    // patch/routing table for all outputs, index defined in OutputDestination kOD_XXX
    v:[
      kIS_Mic1,     // kOD_TDM1    (0x00)
      kIS_Mic2,     // kOD_TDM2    (0x01)
      kIS_Line3,    // kOD_TDM3    (0x02)
      kIS_Line4,    // kOD_TDM4    (0x03)
      kIS_Line5,    // kOD_TDM5    (0x04)
      kIS_Line6,    // kOD_TDM6    (0x05)
      kIS_Line7,    // kOD_TDM7    (0x06)
      kIS_Line8,    // kOD_TDM8    (0x07)
      kIS_DSP19,    // kOD_TDM9    (0x08)
      kIS_DSP20,    // kOD_TDM10   (0x09)
      kIS_SPDIF1,   // kOD_TDM11   (0x0A)
      kIS_SPDIF2,   // kOD_TDM12   (0x0B)
      kIS_ADAT1,    // kOD_TDM13   (0x0C)
      kIS_ADAT2,    // kOD_TDM14   (0x0D)
      kIS_ADAT3,    // kOD_TDM15   (0x0E)
      kIS_ADAT4,    // kOD_TDM16   (0x0F)
      kIS_ADAT5,    // kOD_TDM17   (0x10)
      kIS_ADAT6,    // kOD_TDM18   (0x11)
      kIS_ADAT7,    // kOD_TDM19   (0x12)
      kIS_ADAT8,    // kOD_TDM20   (0x13)
      kIS_Zeros,    // kOD_TDM21   (0x14)
      kIS_Zeros,    // kOD_TDM22   (0x15)
      kIS_Zeros,    // kOD_TDM23   (0x16)
      kIS_Zeros,    // kOD_Zeros   (0x17)
      kIS_TDM15,    // kOD_ADAT1   (0x18)
      kIS_TDM16,    // kOD_ADAT2   (0x19)
      kIS_TDM17,    // kOD_ADAT3   (0x1A)
      kIS_TDM18,    // kOD_ADAT4   (0x1B)
      kIS_TDM19,    // kOD_ADAT5   (0x1C)
      kIS_TDM20,    // kOD_ADAT6   (0x1D)
      kIS_TDM21,    // kOD_ADAT7   (0x1E)
      kIS_TDM22,    // kOD_ADAT8   (0x1F)
      kIS_DSP9,     // kOD_Line3   (0x20)
      kIS_DSP10,    // kOD_Line4   (0x21)
      kIS_DSP11,    // kOD_Line5   (0x22)
      kIS_DSP12,    // kOD_Line6   (0x23)
      kIS_DSP17,    // kOD_Line7   (0x24)
      kIS_DSP18,    // kOD_Line8   (0x25)
      kIS_DSP1,     // kOD_Main1   (0x26)
      kIS_DSP2,     // kOD_Main2   (0x27)
      kIS_DSP3,     // kOD_Line1   (0x28)
      kIS_DSP4,     // kOD_Line2   (0x29)
      kIS_DSP19,    // kOD_Phones1 (0x2A)
      kIS_DSP20,    // kOD_Phones2 (0x2B)
      kIS_TDM13,    // kOD_SPDIF1  (0x2C)
      kIS_TDM14,    // kOD_SPDIF2  (0x2D)
      kIS_Mic1,     // kOD_DSP1    (0x2E)
      kIS_Mic2,     // kOD_DSP2    (0x2F)
      kIS_Line3,    // kOD_DSP3    (0x30)
      kIS_Line4,    // kOD_DSP4    (0x31)
      kIS_SPDIF1,   // kOD_DSP5    (0x32)
      kIS_SPDIF2,   // kOD_DSP6    (0x33)
      kIS_ADAT1,    // kOD_DSP7    (0x34)
      kIS_ADAT2,    // kOD_DSP8    (0x35)
      kIS_Line5,    // kOD_DSP9    (0x36)
      kIS_Line6,    // kOD_DSP10   (0x37)
      kIS_Line7,    // kOD_DSP11   (0x38)
      kIS_Line8,    // kOD_DSP12   (0x39)
      kIS_ADAT3,    // kOD_DSP13   (0x3A)
      kIS_ADAT4,    // kOD_DSP14   (0x3B)
      kIS_ADAT5,    // kOD_DSP15   (0x3C)
      kIS_ADAT6,    // kOD_DSP16   (0x3D)
      kIS_TDM1,     // kOD_DSP17   (0x3E)
      kIS_TDM2,     // kOD_DSP18   (0x3F)
      kIS_TDM3,     // kOD_DSP19   (0x40)
      kIS_TDM4,     // kOD_DSP20   (0x41)
      kIS_TDM5,     // kOD_DSP21   (0x42)
      kIS_ADAT7,    // kOD_DSP22   (0x43)
      kIS_ADAT8,    // kOD_DSP23   (0x44)
      kIS_Zeros,    // kOD_DSP24   (0x45)
      kIS_TDM6,     // kOD_DSP25   (0x46)
      kIS_TDM7,     // kOD_DSP26   (0x47)
      kIS_TDM8,     // kOD_DSP27   (0x48)
      kIS_TDM9,     // kOD_DSP28   (0x49)
      kIS_TDM10,    // kOD_DSP29   (0x4A)
      kIS_TDM11,    // kOD_DSP30   (0x4B)
      kIS_TDM12,    // kOD_DSP31   (0x4C)
      kIS_Zeros,    // kOD_DSP32   (0x4D)
      kIS_Zeros,
      kIS_Zeros
    ],
  },
  kLoopbackSource: {
    id:5014,
    t:kTByte,
    v:[
      kIS_DSP19,
      kIS_DSP20
    ],
  },
  
  kiChannelNames: {
    id:6,
    t:kTString,
    // Names must be 32 bytes or less
    v:[
      "", "", // Mic
      "", "", "", "", "", "", // Line
      "", "", // spdif
      "", "", "", "", "", "", "", "", // adat/toslink
    ],
  },

  koChannelNames: {
    id:7,
    t:kTString,
    // Names must be 32 bytes or less
    v:[
      "", "", // main 
      "", "", "", "", "", "", "", "", // Line
      "", "", // phones
      "", "", // spdif
      "", "", "", "", "", "", "", "", // adat/toslink
    ],
  },
  koTrim: {
    id:5000,
    t:kTByte,
    // range is 0-100  range is 0 to -100 dB. 
    // Use LocalOutputs kLO_XXX for index
    v:[
      0, // kLO_Line5   = 0x0,
      0, // kLO_Line6   = 0x1,
      0, // kLO_Line7   = 0x2,
      0, // kLO_Line8   = 0x3,
      0, // kLO_Line9   = 0x4,
      0, // kLO_Line10  = 0x5,
      0, // kLO_Main1   = 0x6,
      0, // kLO_Main2   = 0x7,
      0, // kLO_Line3   = 0x8,
      0, // kLO_Line4   = 0x9,
      4, // kLO_PhonesL = 0xA,
      4, // kLO_PhonesR = 0xB,
    ]
  },
  kMainTrim: {
    id:5011,
    t:kTByte,
    // range is 0-100 range is 0 to -100 dB. 
    v:[0],
  },
  kMainGroup: {
    id:5012,
    t:kTInt16,
    // Each bit is the enable for it's index
    // bit kLO_Main1
    // bit kLO_Main2
    // bit kLO_Line3
    // bit kLO_Line4
    // bit kLO_Line5
    // bit kLO_Line6
    // bit kLO_Line7
    // bit kLO_Line8
    // bit kLO_Line9
    // bit kLO_Line10
    v:[0xC0],
  },
  
  //////////////// MIX ///////////////////

  kiMixStereo: {
    id:1000,
    t:kTByte,
    // bool
    v:[
      0,0, // mic
      0,0,0,0,0,0, // Line
      1,0, // spdif
      0,0,0,0,0,0,0,0, // adat/toslink
      1,0, // host out main
      1,0, // host out phones
      1,0,1,0,1,0,1,0, // host out Line
      1,0  // dsp reverb 
    ], 
  },
  koMixStereo: {
    id:1001,
    t:kTByte,
    // bool
    v:[
      1,0, // main 
      1,0,1,0,1,0,1,0, // Line
      1,0,1,0 // phones (unused) and reverb (unused)
    ], 
  },

  // input eq 4 bands per channel (packed c1: b1 b2 b3 b4, c2: b1 b2 b3 b4 etc)
  kiEQMode: {
    id:1002,
    t:kTByte,
    // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
    v:[
      1,0,0,2,
      1,0,0,2,
      1,0,0,2,
      1,0,0,2,
      1,0,0,2,
      1,0,0,2,
      1,0,0,2,
      1,0,0,2
    ],
  },
  kiEQBypass: {
    id:1003,
    t:kTByte,
    // bool
    v:[
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0
    ],
  },
  kiEQFreq: {
    id:1004,
    t:kTInt32,
    // float
    v:[
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000,
      60,400,2000,14000
    ],
  },
  kiEQGain: {
    id:1005,
    t:kTFloat,
    // float
    v:[
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0,
      0,0,0,0
    ],
  },
  kiEQBW: {
    id:1006,
    t:kTFloat,
    // float
    v:[
      1,1,1,1,
      1,1,1,1,
      1,1,1,1,
      1,1,1,1,
      1,1,1,1,
      1,1,1,1,
      1,1,1,1,
      1,1,1,1
    ],
  },

  kiGateBypass: {
    id:1007,
    t:kTByte,
    // bool
    v:[1,1],
  },
  kiGateThreshold: {
    id:1008,
    t:kTFloat,
    // float dB 0:-inf 1:0
    v:[-46,-46],
  },
  kiGateAttack: {
    id:1009,
    t:kTInt16,
    // 1-100 ms
    v:[10,10],
  },
  kiGateRelease: {
    id:1010, 
    t:kTInt16,
    // 50-2000 ms
    v:[1000,1000],
  },

  kiCompBypass: {
    id:1011,
    t:kTByte,
    // bool
    v:[1,1,1,1,1,1,1,1],
  },
  kiCompThreshold: {
    id:1012,
    t:kTFloat,
    // float dB 0:-inf 1:0
    v:[-10,-10,-10,-10,-10,-10,-10,-10],
  },
  kiCompAttack: {
    id:1013,
    t:kTInt16,
    // 1-100 ms
    v:[10,10,10,10,10,10,10,10],
  },
  kiCompRelease: {
    id:1014,
    t:kTInt16,
    // 50-2000 ms
    v:[1000,1000,1000,1000,1000,1000,1000,1000],
  },
  kiCompRatio: {
    id:1015,
    t:kTFloat,
    // float 1-10?
    v:[2,2,2,2,2,2,2,2],
  },
  kiCompMakeup: {
    id:1021,
    t:kTFloat,
    // float
    v:[1,1,1,1,1,1,1,1],
  },

  // mic Line spdif atos homain holine hophones reverb 
  // The layout of index's is per bus, so channels 0-31 bus 0, then channels 0-31 bus 1 etc
  kiMixFader: {
    id:1016,
    t:kTFloat,
    // 8.24 -> float 32 in x 14 out
    v:[
      //0m 2Line        8sp  10 atos          18hm 20hLine          28hp 30r 
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 1,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,1,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,1,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,1,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,1,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,1,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,1,0, 0,0, 0,0, // Ana 7, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,1, 0,0, 0,0, // Ana 8, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 1,0, 0,0, // Phones 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,1, 0,0, // Phones 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
    ],
  },
  kiMixPan: {
    id:1017,
    t:kTFloat,
    // 8.24 -> float 32 in x 14 out
    v:[
      //0mic   2Line                  8spdif   10 atos                          18hmain  20hLine                        22hphns  30r 
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Main 1, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Main 2, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 1, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 2, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 3, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 4, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 5, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 6, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 7, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 8, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Phones 1, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Phones 2, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Reverb 1, Ch 0-31
      0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Reverb 2, Ch 0-31
    ],
  },
  kiMixSolo: {
    id:1018,
    t:kTByte,
    // bool 32 in x 14 out
    v:[
      //0m 2Line      8sp  10 atos          18hm 20 hLine       22hp 30r 
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 7, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 8, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
    ],
  },
  kiMixMute: {
    id:1019,
    t:kTByte,
    // bool 32 in x 14 out
    v:[
      //0m 2Line      8sp  10 atos          18hm 20 hLine       22hp 30r 
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 7, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 8, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 2, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
      0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
    ],
  },
  kReverbDecay: { 
    id:1030,
    t:kTByte,
    // decay: 0-100 as a percent of whatever 86 = default
    v:[86],
  },
  kReverbDamping: { 
    id:1031, 
    t:kTByte,
    // damping: 0-100 as a percent of whatever 
    v:[90],
  },
  kReverbPreDelay: {
    id:1032, 
    t:kTByte,
    // PreDelay samples 0-100 ms
    v:[50],
  },
  kReverbWidth: {
    id:1033, 
    t:kTByte,
    // Width 0-100 percent
    v:[50],
  },
  kReverbPreset: {
    id:1034, 
    t:kTByte,
    // Preset the hardware appies to the coefficients
    v:[1],
  },

  // output eq 3 bands per channel
  koEQMode: {
    id:1022,
    t:kTByte,
    // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
    v:[
      1,0,2, // Main L-R
      1,0,2, // unused
      1,0,2, // Line 3
      1,0,2, // Line 4
      1,0,2, // Line 5
      1,0,2, // Line 6
      1,0,2, // Line 7
      1,0,2, // Line 8
      1,0,2, // Line 9
      1,0,2, // Line 10
      1,0,2, // Phones L-R
      1,0,2  // Reverb
    ],
  },
  koEQBypass: {
    id:1023,
    t:kTByte,
    // bool
    v:[
      0,0,0, // Main L-R
      0,0,0, // unused
      0,0,0, // Line 3
      0,0,0, // Line 4
      0,0,0, // Line 5
      0,0,0, // Line 6
      0,0,0, // Line 7
      0,0,0, // Line 8
      0,0,0, // Line 9
      0,0,0, // Line 10
      0,0,0, // Phones L-R
      0,0,0  // Reverb
    ],
  },
  koEQFreq: {
    id:1024,
    t:kTInt32,
    // float
    v:[
      60,800,14000, // Main L-R
      60,800,14000, // unused
      60,800,14000, // Line 3
      60,800,14000, // Line 4
      60,800,14000, // Line 5
      60,800,14000, // Line 6
      60,800,14000, // Line 7
      60,800,14000, // Line 8
      60,800,14000, // Line 9
      60,800,14000, // Line 10
      60,800,14000, // Phones L-R
      60,800,14000, // Reverb
    ],
  },
  koEQGain: {
    id:1025,
    t:kTFloat,
    // float
    v:[
      0,0,0, // Main L-R
      0,0,0, // unused
      0,0,0, // Line 3
      0,0,0, // Line 4
      0,0,0, // Line 5
      0,0,0, // Line 6
      0,0,0, // Line 7
      0,0,0, // Line 8
      0,0,0, // Line 9
      0,0,0, // Line 10
      0,0,0, // Phones L-R
      0,0,0  // Reverb
    ],
  },
  koEQBW: {
    id:1026,
    t:kTFloat,
    // float
    v:[
      1,1,1, // Main L-R
      1,1,1, // unused
      1,1,1, // Line 3
      1,1,1, // Line 4
      1,1,1, // Line 5
      1,1,1, // Line 6
      1,1,1, // Line 7
      1,1,1, // Line 8
      1,1,1, // Line 9
      1,1,1, // Line 10
      1,1,1, // Phones L-R
      1,1,1  // Reverb
    ],
  },

  // output gain 8.24fp
  koBusFader: {
    id:1027,
    t:kTFloat,
    // float: 14 out same order as mix buses
    v:[1,1, 1,1, 1,1, 1,1, 1,1, 1,1, 1,1],
  },
  koBusMute: {
    id:1028,
    t:kTByte,
    // float: 14 out same order as mix buses
    v:[0,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0],
  },

  kHostOutMixMode: {
    id:1020,
    t:kTByte,
    // bool: 0 = basic, 1 = extended
    v:[0],
  },

  // ID's -1 through -999 are local storage and not sent to the device
  kMixInMeterMode: {
    id:-1,
    t:kTByte,
    // bool: 0 = Pre Fader, 1 = Post Fader
    v:[0],
  },
  kUIMeterFPS: {
    id:-2,
    t:kTInt32,
    // kUIMeterFPSs
    v:[30],
  },
  kBusStripExp: {
    id:-4,
    t:kTString,
    // "on" or "off"
    v:["on"],
  },
  kTabExp: {
    id:-6,
    t:kTString,
    // "on" or "off"
    v:["on"],
  },

  // ID's -1000 through -1999 are windows only parameters via dll
  kHostBuffer1x: {
    id:-1000, 
    t:kTInt32,
    v:[128],
  },
  kExtraSafetyOffset1X: {
    id:-1001, 
    t:kTInt32,
    // expect 0 (lowest latency) or 1 (default)
    v:[1],
  },
  kWaveInEnables: {
    id:-1002, 
    t:kTInt32,
    v:[0xfff],
  },
  kWaveOutEnables: {
    id:-1003, 
    t:kTInt32,
    v:[0xfff],
  },
  kWaveSRCEnabled: {
    id:-1004,
    t:kTInt32,
    v:[0],
  },
};

