'use strict';

class DiagnosticsModalView extends View {
  constructor() {
    super(0,0,kFxWidth,kFxHeight);

    this.setTimer = this.setTimer.bind(this);
    this.onTimer = this.onTimer.bind(this);
    this.activate = this.activate.bind(this);
    this.onAEErr = this.onAEErr.bind(this)  
    this.onDS = this.onDS.bind(this)  
    this.wantsMouse = true;
    this.setBG(kBlackBG);

    this.modal = new ModalCanvas(false);
    this.modal.mdiv.style.zIndex = 1400; // lower than updater
    this.modal.deactivate();
    this.active = false;

    this.bufferStatusOk = true;
    this.dpcLatencyMax = 0;
    this.dpcLatencyWorst = 0;
    this.curSampleRate = 48000;
    this.curBufferSize = 256;
    this.aeErrs = {};

    ds.addListener('kSampleRate', this.onDS); 
    ds.addListener('kHostBuffer1x', this.onDS);
    evSys.addListener('kWinAEErr', this.onAEErr);

    this.okRect = new Rect(this.r - 100, this.b - 40, 80, 30);
    this.clearRect = new Rect(this.okRect.l - 120, this.okRect.t, this.okRect.w, this.okRect.h);
  }
  activate(showIt) {
    if(!showIt) 
      this.modal.deactivate();
    else {
      this.modal.activate(this);
      this.requestRender();
    }
    // check the diagnostics once every 5 seconds
    this.active = showIt;
    this.setTimer(5000);
  }
  handleMouseDown(xy) {
    if(this.okRect.contains(xy.x, xy.y))
      this.modal.deactivate();
    else if(this.clearRect.contains(xy.x, xy.y)) {
      this.dpcLatencyMax = 0;
      this.dpcLatencyWorst = 0;
      this.aeErrs = {};
      this.requestRender();
    }
  }
  setTimer(ms) {
    if(this.timerID != undefined)
      clearTimeout(this.timerID);
    this.timerID = setTimeout(this.onTimer, ms);
  }
  onTimer() {
    this.timerID = undefined;
    if(this.active) {
      var d = gen5Lib.getDiagnostics();
      if(!d) {
        this.activate(false);
        return;
      }

      // the max time spent processing a callback from when the device sent the interrupt to when the callback exited
      if (d.minUSBLat != 64) // engine sets to 64 when it has no diagnostics
      {
          this.dpcLatencyMax = Math.max((d.maxUSBLat - 1) * 125, 125);
          var maxTime_Sec= (this.dpcLatencyMax + d.maxCBus) / 1000000.0;
          var bufferTime_Sec = this.curBufferSize / this.curSampleRate;
          this.bufferStatusOk = (1.2 * bufferTime_Sec) > maxTime_Sec;
      }

      if (this.dpcLatencyMax > this.dpcLatencyWorst)
        this.dpcLatencyWorst = this.dpcLatencyMax;

      this.setTimer(5000);
      this.requestRender(0);
    }
  }
  onAEErr(err) {
    aeErrs[err] = new Date();
    this.requestRender();
  }
  onDS(k, v, setByObject) {
    if(k == 'kSampleRate')
      this.curSampleRate = v[0];
    else if(k == 'kHostBuffer1x')
      this.curBufferSize = v[0];
  }
  render(c){
    c.fillStyle = c.strokeStyle = kToggleDefault;
    c.font = "Lighter 18px Helvetica";
    c.textAlign = 'left';

    c.lineWidth = 4;
    c.strokeRect(this.l+2, this.t+2, this.w-4, this.h-4);

    const lineHeight = 22;
    var y = (2 * lineHeight) + fillTextWrapped(c, "The information displayed here can be helpful when troubleshooting USB audio. Ignore this view unless you are experiencing frequent audio issues.", this.l+20, this.t+20, this.w-40, lineHeight);

    if(this.dpcLatencyMax != 0)
      y = (2 * lineHeight) + fillTextWrapped(c, "USB DPC Latency: " + this.dpcLatencyMax + " μs, highest measured " + this.dpcLatencyWorst + " μs" , this.l+20, y, this.w-40, lineHeight); 

    for (let [e, d] of Object.entries(this.aeErrs)) {
      var codeStr = d.toString() + " Code " + e.toString(16);
      var err = e >> 16;
      var more = e & 0xffff;
      y = lineHeight + fillTextWrapped(c, codeStr, this.l+20, y, this.w-40, lineHeight);
      switch (err)
      {
      default:
        y += lineHeight;
        codeStr + "\n";
        break;
      case 1:  // (1 << 16) 
        // USB stream failed to start, memory or USB setup could cause this
        // err = kFailStartUSB = 1,    
        y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB controller is reporting errors and the driver is unable to perform audio. Try another USB port on a different card, or a different USB cable.", this.l+20, y, this.w-40, lineHeight);
        break;
      case 2:
        // The isoch stream failed while operating. 
        // err = kFailStreamErr = 2,     
        // More holds the USBStreamStatus
        // chase underflowed, ie chase was not called often enough (probably interrupts are dropped)
        // more = kUSBStreamChaseUnderrun = 1,
        // isoch input packets were not received (dropped, crc failed, or host didn't poll)
        // more = kUSBStreamInputError = 2 
        if (more == 1) { // (2 << 16) & 1
          // I'm filtering out this error unless it occurs multiple times in a short amount of times
          // because it's possible these errors may happen once in a while and we don't want folks to be alarmed
          //if (aeErrTimes.ContainsKey(kv.Key) && (aeErrTimes[kv.Key] != 0) && (aeErrTimes[kv.Key] < 2000))
          //    errStr += codeStr + "System DPC latency was extremely high and has caused an audio dropout. This was most likely caused by another driver on this system. Please contact MOTU tech support if this problem persists.\n\n";
          y = (2 * lineHeight) + fillTextWrapped(c, "The driver was unable to process audio in time and audio was dropped. This was most likely caused by another driver on this system. Please contact MOTU tech support if this problem persists.", this.l+20, y, this.w-40, lineHeight);
        } 
        else if (more == 2) { // (2 << 16) & 1
          y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB host controller has dropped audio. Try another USB port on a different card, or a different USB cable.", this.l+20, y, this.w-40, lineHeight);
        } 
        else { // (2 << 16) & 1
          y += lineHeight;
        }
        break;
      case 3: // (3 << 16)
        // interrupt message sequence numbers mismatched (interrupt was dropped by controller), audio was resynchronized.
        // the driver tries to resync but if it happens a lot there will be obvious drop outs. May be improved by larger buffers
        // kResyncIntSeqErr = 3,
        y = (2 * lineHeight) + fillTextWrapped(c, "The computer's USB controller cannot perform at the current buffer size. Increase the buffer size to avoid audio dropouts. If the problem persists with larger buffer sizes then there may be a problem with the USB controller on this system.", this.l+20, y, this.w-40, lineHeight);
        break;
      }
    }

    // if (string.IsNullOrEmpty(errStr))
    //     diagnosticsOutput.AppendText("No diagnostic events have been reported");
    // else
    //     diagnosticsOutput.AppendText(errStr);

    drawButton(c, this.okRect, "OK", false, kToggleDefault);
    drawButton(c, this.clearRect, "Clear", false, kToggleDefault);
  } 
}


// name and bit
const kWaveInInfo = [
  {n:"Loopback In 1-2", b:4},
  {n:"Mic/Line/Inst In 1-2", b:0},
  {n:"Line In 3-4", b:1},
  {n:"Line In 5-6", b:2},
  {n:"Line In 7-8", b:3},
  {n:"S/PDIF In 1-2", b:5},
  {n:"Optical In 1-2", b:6},
  {n:"Optical In 3-4", b:7},
  {n:"Optical In 5-6", b:8},
  {n:"Optical In 7-8", b:9},
];
const kWaveOutInfo = [
  {n:"Main Out 1-2", b:0},
  {n:"Phones Out 1-2", b:5},
  {n:"Line Out 3-4", b:1},
  {n:"Line Out 5-6", b:2},
  {n:"Line Out 7-8", b:3},
  {n:"Line Out 9-10", b:4},
  {n:"S/PDIF Out 1-2", b:6},
  {n:"Optical Out 1-2", b:7},
  {n:"Optical Out 3-4", b:8},
  {n:"Optical Out 5-6", b:9},
  {n:"Optical Out 7-8", b:10},
];
class WaveEnableMultiSelect extends MultiSelect {
  constructor(d, l, t, w, h, ctrlID, dsKey, names) {
    super(d, l, t, w, h, ctrlID);

    this.onDSArray = this.onDSArray.bind(this);
    var n = names.length;
    for(var i = 0; i < n; i++) {
      this.addOption(names[i].n, names[i].b);
    }        

    SetupDSArrayListenerBasic(this, dsKey, 0);
    this.onDSArray(dsKey, DSArrayGetBasic(this));
    this.scrollBarV.setAutoHide(false);
  }
  changeCB(id, checked){
    var v = DSArrayGetBasic(this);
    if(checked)
      v |= (1 << id);
    else
      v &= ~(1 << id);
    DSArraySetBasic(this, v);
  }
  onDSArray(k,v) {
    var n = this.cBoxes.length;
    for(var i = 0; i < n; i++) {
      this.cBoxes[i].checked = v & (1 << this.cBoxes[i].v);
    }
  }
}

function getBufferSizeArray() {
  var sr = ds.getKVArrayItem('kSampleRate', 0);
  var srMult = getSampleRateMult(sr);
  if (srMult == 1) {
    return kHostBuffers1x;
  }
  else if (srMult == 2) {
    return kHostBuffers2x;
  }
  return kHostBuffers4x;
}

class BufferSizeMenu extends DSLabelSelectMenu {
  constructor(d, l, t) {
    super(d, l, t, "Host Buffer Size:", 'kHostBuffer1x', getBufferSizeArray(), 0); 

    this.myDiv = d;
    this.srChanged = this.srChanged.bind(this);
    ds.addListener('kSampleRate', this.srChanged);
  }

  srChanged() {
    // Choose the correct values array for the sample rate.
    var sr = ds.getKVArrayItem('kSampleRate', 0);
    var srMult = getSampleRateMult(sr);
    this.setMenuVals(this.myDiv, 'kHostBuffer1x', getBufferSizeArray(), 0); // Replace the menu object

    // change the labels of the options to match the mult
    var options = this.menu.menu.sel.options;
    var len =  len = options.length;
    for (var i = 0; i < len; i++ ) {
      options[i].text = options[i].value * srMult;
    }    
  }
}

class SaveLoadPreset extends View {
  constructor(l, t) {
    super(l, t, 330, 40);

    this.savePresetR = new Rect(190, t + 5, 70, 30);
    this.loadPresetR = new Rect(290, t + 5, 70, 30);
    this.wantsMouse = true;

  }
  handleMouseDown(xy) {
    if(this.savePresetR.contains(xy.x, xy.y))
      ds.savePreset();
    else if(this.loadPresetR.contains(xy.x, xy.y)) 
      ds.loadPreset();
  }
  render(c) {
    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';
    c.fillText("Device Presets:", this.l, this.t + 25);
    drawButton(c, this.savePresetR, "Save...", false, "black");
    drawButton(c, this.loadPresetR, "Load...", false, "black");
  }
}

class DiagnosticsIcon extends View {
  constructor(l, t) {
    super(l, t, 20, 20);
    this.wantsMouse = true;
    this.diagnosticsView = new DiagnosticsModalView();
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y))
      this.diagnosticsView.activate(true);
  }
  render(c) {
    drawDiagnosticsIcon(c, this.cx, this.cy, "white");
  }
}


class LockWaveSRCheckbox extends View {
  constructor(d, l, t) {
    super(l, t, 600, 50);
    this.waveSRCEnableChanged = this.waveSRCEnableChanged.bind(this);
    this.changeCB = this.changeCB.bind(this);

    this.wantsMouse = true;

    var cbId = "waveSRCenable_id";

    this.cbox = document.createElement('input');
    this.cbox.type = 'checkbox';
    this.cbox.checked = true;
    this.cbox.id = cbId;

    var label = document.createElement('label');
    label.setAttribute('for', cbId); 
    label.innerHTML = "Sync Windows sample rate to device";

    var that = this;
    this.cbox.addEventListener('change', e => { that.changeCB(e.srcElement.checked)});

    var div = document.createElement('div');
    div.className = 'user-multi-select'

    div.style.width =  this.w + "px";
    div.style.height = this.h + "px";
    div.style.left = this.l + "px";
    div.style.top = this.t + "px";
    div.style.overflow = "hidden";
    div.style.borderStyle = "none";

    div.appendChild(this.cbox);
    div.appendChild(label);

    ds.addListener('kWaveSRCEnabled', this.waveSRCEnableChanged);

    d.appendChild(div);
  }

  waveSRCEnableChanged() {
    this.cbox.checked = !ds.getKVArrayItem('kWaveSRCEnabled', 0);
  }
  
  changeCB(checked){
    ds.setV('kWaveSRCEnabled', 0, !checked);
  }

}


class DevSetup extends View{
  constructor() {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot();
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    this.updateDriverVersion = this.updateDriverVersion.bind(this);
    evSys.addListener('wResized', this.resized);
    evSys.addListener('driverVersion', this.updateDriverVersion);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "devSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.width = "calc(100% - " + xPos + "px)";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "devSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;
    
    this.tabDiv = this.scrollBarVH.containerDiv;

    this.c.addSub(this.rate = new DSLabelSelectMenu(this.c.div, 30, 80, "Sample Rate:", 'kSampleRate', kSampleRates, 0));
    this.c.addSub(this.src = new DSLabelSelectMenu(this.c.div, 30, 140, "Clock Source:", 'kClockSource', kClockSources, 0));
    this.c.addSub(this.opt = new DSLabelSelectMenu(this.c.div, 30, 200, "Loopback Location:", 'kLoopbackFirst', kLoopbackOrder, 0));
    this.c.addSub(this.opt = new DSLabelSelectMenu(this.c.div, 30, 260, "Mix Input Meters:", 'kMixInMeterMode', kMixInputMetersMode, 0));
    this.c.addSub(this.opt = new DSLabelSelectMenu(this.c.div, 30, 320, "Mix USB Channels:", 'kHostOutMixMode', kHostOutputMixMode, 0));
    
    this.driverVersion = undefined;
    
    if(!gen5Lib.hasLib()) {
      if(device.hasHostMsgPipe()) { // mac is connected
        this.addSub(new SaveLoadPreset(30, 380));
        this.c.addSub(this.upd = new UpdateView(30, 440));
        this.driverVersion = "";
      }
      else {
        // display the update view even though there's no host so it can show serial number
        this.c.addSub(this.upd = new UpdateView(30, 380));
      }
    }
    else { // pc is connected
      this.driverVersion = "";
      this.c.addSub(this.so = new BufferSizeMenu(this.c.div, 30, 380));
      this.c.addSub(this.so = new DSLabelSelectMenu(this.c.div, 30, 440, "Safety Offsets:", 'kExtraSafetyOffset1X', kExtraSafetyOffset1X, 0));
      this.c.addSub(this.wi = new WaveEnableMultiSelect(this.c.div, 30, 530, 200, 150, 'wavei', 'kWaveInEnables', kWaveInInfo));
      this.c.addSub(this.wo = new WaveEnableMultiSelect(this.c.div, 260, 530, 200, 150, 'waveo', 'kWaveOutEnables', kWaveOutInfo));
      this.c.addSub(this.lockSRcb = new LockWaveSRCheckbox(this.c.div, 30, 685));
      this.addSub(new SaveLoadPreset(30, 735));
      this.c.addSub(this.upd = new UpdateView(30, 795));
      this.c.addSub(this.diag = new DiagnosticsIcon(this.l + 30, this.upd.b + 26));
      
    }

    this.setBG(kEmptyBG);
  }
  updateDriverVersion(ev, dv) {
    this.driverVersion = dv;
  }
  resized() {
    var w = 400;
    var h = this.upd.b;
    if(this.driverVersion)
      h += 60;

    var cs = getComputedStyle(this.c.div);
    var w = Math.max(w, parseInt(cs.width, 10)) - 1;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
  render(c) {
    drawDeviceNameHeader(c, this.l + 35, this.t, 630);

    kAlertImg.render(c, 358, 198);
    c.font = "600 11px Arial";
    c.textAlign = 'left';
    c.fillText("Changes to Loopback Location", 405, 210);
    c.fillText("will reset the USB connection.", 405, 210 + 17);

    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';

    if(this.wi) {
      c.fillText("Enable Windows Audio Streams:", this.l + 30 + 2, 521);
    }

    if(this.driverVersion)
      c.fillText("Driver Version:", this.l + (this.diag != undefined ? 60 : 30), this.upd.b + 40);
      c.fillStyle = kToggleDefault;
      c.fillText(this.driverVersion, this.l + 190, this.upd.b + 40);
  }
}
