'use strict';

const kULm5BG = new OffScreenImage("ulm5.png", 0.340);
const kSpeakerImg = new OffScreenImage("speaker.svg");
const kPhonesImg = new OffScreenImage("phones.svg");
const kAlertImg = new OffScreenImage("alert.svg");

class MicPreHome extends View {
  constructor(p, l, t, i) {
    super(l,t,135,200);
    p.addSub(this);
    this.onJack = this.onJack.bind(this);
    this.chIdx = i;
    this.addSub(this.meter = new IOMeter(this.l + 17, this.t + 25, 10, 150, kADCInChInfo[i].m));
    this.addSub(this.b48V = new ToggleBtn(l + 53, t + 27, 52, "48V", 'k48V', i, kToggleRed, kEmptyBG));
    this.addSub(this.pad = new ToggleBtn(l + 53, t + 73, 52, "Pad", 'kPad', i, kToggleBlue, kEmptyBG));
    this.addSub(this.gain = new IOKnob(l + 45, t + 115, kToggleBlue, kGateCtrlDisabled, kPreTrimCtrlInfo[0], i));
    
    ds.addListener('kJackDetect', this.onJack);
  }  
  onJack(k, v) {
    if(k == 'kJackDetect') {
      // bool 0 = mic, 1 = line/inst
      this.b48V.enable(v[this.chIdx] == 0);
      this.pad.enable(v[this.chIdx] == 0);
    }
  }
  render(c) {
    drawLabelGroup(c, getIName(this.chIdx), this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}
class PhonesVolHome extends View {
  constructor(p, l, t) {
    super(l,t,130,200);
    p.addSub(this);
    this.chIdx = 10;
    this.addSub(this.m1 = new IOMeter(this.l + 12, this.t + 25, 10, 150, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[10].osi)));
    this.addSub(this.m2 = new IOMeter(this.m1.l + 12, this.t + 25, 10, 150, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[11].osi)));
    this.addSub(this.trim = new IOKnob(l + 45, t + 115, kToggleBlue, kGateCtrlDisabled, kPhonesTrimCtrlInfo[0], kOutChInfo[10].lo, kOutChInfo[11].lo));
  }  
  render(c) {
    drawLabelGroup(c, "Phones Vol", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}
class MainVolGroupHome extends View {
  constructor(p, l, t) {
    super(l,t,130,200);
    p.addSub(this);
    this.chIdx = 10;
    this.addSub(this.m1 = new IOMeter(this.l + 12, this.t + 25, 10, 150, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[0].osi)));
    this.addSub(this.m2 = new IOMeter(this.m1.l + 12, this.t + 25, 10, 150, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[1].osi)));
    this.addSub(this.vol = new IOKnob(l + 45, t + 115, kToggleBlue, kGateCtrlDisabled, kMainVolCtrlInfo[0], 0));
  }  
  render(c) {
    drawLabelGroup(c, "Main Vol", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}
function getChInputMonitorState(ch, faders, mutes, solos, pans, stereo) {
  // 0 = off, 1 = on, 2 = edited
  var roundedFader = Math.round(100*faders[ch])/100;
  if(faders[ch] == 0)
    return 0;
  if(!stereo && (mutes[ch] == 0) && (pans[ch] == 0.5)) {
    if(roundedFader == 1)
      return 1;
  }
  return 2;
}
function getInputMonitorStates(i, faders, stereos, mutes, solos, pans) {
  // 0 = off, 1 = on, 2 = edited
  // this counts on index's for line in mixer channels (Mic1-line8) being contiguous
  var imStates = { ml: 2, mr:2, pl: 2, pr:2 };
  var stereo = stereos[i & 0xe] != 0;
  imStates.ml = getChInputMonitorState(i, faders, mutes, solos, pans, stereo);
  imStates.mr = getChInputMonitorState(i+1, faders, mutes, solos, pans, stereo);
  imStates.pl = getChInputMonitorState(i + (32 * 10), faders, mutes, solos, pans, stereo);
  imStates.pr = getChInputMonitorState(i + (32 * 10) + 1, faders, mutes, solos, pans, stereo);

  return imStates;
}
function enableInputMonitorState(i, enable) {
  if(!enable)
    ds.setV('kiMixFader', i, 0);
  else {
    ds.setV('kiMixFader', i, 1);
    ds.setV('kiMixStereo', i, 0);
    ds.setV('kiMixMute', i, 0);
    ds.setV('kiMixSolo', i, 0);
    ds.setV('kiMixPan', i, 0.5);
  }
}
function drawIMButton(c, r, label, imState) {
  if(imState == 2)
    drawButton(c, r, label, true, kEQLabel);
  else
    drawButton(c, r, label, imState == 1, kToggleBlue);
}

class InputMonitoring extends View {
  constructor(l, t) {
    super(l, t, 605, 210);

    this.update = this.update.bind(this);

    this.setBG(kEmptyBG);
    this.wantsMouse = true;
    this.items = new List();

    this.infoR = new Rect(this.r-35, this.t+20, 30, 30);
    this.xR = new Rect(this.r-40, this.t+30, 20, 20);
    this.showInfoMessage = false;

    var w1 = 49;
    var t1 = this.t + 64;
    var t2 = this.t + 119;
    var l1 = this.l + 195;

    // ln = 0 based channel num of l channel
    // mlr = rect for left ch main 
    // mrr = rect for right ch main
    // plr = rect for left ch phones 
    // prr = rect for right ch phones
    // mix faders channels
    this.items.add({  lch:0, 
                      mlr:(new Rect(l1, t1, 25, 25)),        
                      mrr:(new Rect(l1 + w1, t1, 25, 25)),   
                      plr:(new Rect(l1, t2, 25, 25)),        
                      prr:(new Rect(l1 + w1, t2, 25, 25)),
                      ims:{ ml: 0, mr:0, pl: 0, pr:0 }
    });
    this.items.add({  lch:2, 
                      mlr:(new Rect(l1 + 2*w1, t1, 25, 25)), 
                      mrr:(new Rect(l1 + 3*w1, t1, 25, 25)), 
                      plr:(new Rect(l1 + 2*w1, t2, 25, 25)), 
                      prr:(new Rect(l1 + 3*w1, t2, 25, 25)),
                      ims:{ ml: 0, mr:0, pl: 0, pr:0 }
    });
    this.items.add({  lch:4, 
                      mlr:(new Rect(l1 + 4*w1, t1, 25, 25)), 
                      mrr:(new Rect(l1 + 5*w1, t1, 25, 25)), 
                      plr:(new Rect(l1 + 4*w1, t2, 25, 25)), 
                      prr:(new Rect(l1 + 5*w1, t2, 25, 25)),
                      ims:{ ml: 0, mr:0, pl: 0, pr:0 }
    });
    this.items.add({  lch:6, 
                      mlr:(new Rect(l1 + 6*w1, t1, 25, 25)), 
                      mrr:(new Rect(l1 + 7*w1, t1, 25, 25)), 
                      plr:(new Rect(l1 + 6*w1, t2, 25, 25)), 
                      prr:(new Rect(l1 + 7*w1, t2, 25, 25)),
                      ims:{ ml: 0, mr:0, pl: 0, pr:0 }
    });

    ds.addListener('kiMixFader', this.update);
    ds.addListener('kiMixStereo', this.update);
    ds.addListener('kiMixMute', this.update);
    ds.addListener('kiMixSolo', this.update);
    ds.addListener('kiMixPan', this.update);
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      var n = this.items.length();
      for(var i = 0; i < n; ++i) {
        var item = this.items.get(i);
        if(!this.showInfoMessage) {
          if(item.mlr.contains(xy.x, xy.y))
            return this;
          else if(item.mrr.contains(xy.x, xy.y))
            return this;
          else if(item.plr.contains(xy.x, xy.y))
            return this;
          else if(item.prr.contains(xy.x, xy.y))
            return this;
        }
      }
      if (!this.showInfoMessage && this.infoR.contains(xy.x, xy.y)) {
        return this;
      }
      else if (this.showInfoMessage && this.xR.contains(xy.x, xy.y)) {
        return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) { 
    var n = this.items.length();
    for(var i = 0; i < n; ++i) {
      var item = this.items.get(i);
      if(!this.showInfoMessage) {
        if(item.mlr.contains(xy.x, xy.y))
          enableInputMonitorState(item.lch, item.ims.ml == 0);
        else if(item.mrr.contains(xy.x, xy.y))
          enableInputMonitorState(item.lch + 1, item.ims.mr == 0);
        else if(item.plr.contains(xy.x, xy.y))
          enableInputMonitorState(item.lch + (32 * 10), item.ims.pl == 0);
        else if(item.prr.contains(xy.x, xy.y))
          enableInputMonitorState(item.lch + (32 * 10) + 1, item.ims.pr == 0);
      }
    }
    if (!this.showInfoMessage && this.infoR.contains(xy.x, xy.y)) {
      this.showInfoMessage = true;
      this.requestRender();
    }
    else if (this.showInfoMessage && this.xR.contains(xy.x, xy.y)) {
      this.showInfoMessage = false;
      this.requestRender();
    }
  }
  update() {
    var faders = ds.getKVArray('kiMixFader');
    var stereos = ds.getKVArray('kiMixStereo');
    var mutes = ds.getKVArray('kiMixMute');
    var solos = ds.getKVArray('kiMixSolo');
    var pans = ds.getKVArray('kiMixPan');

    var n = this.items.length();
    for(var i = 0; i < n; ++i) {
      var item = this.items.get(i);
      item.ims = getInputMonitorStates(item.lch, faders, stereos, mutes, solos, pans);
    }
    this.requestRender();
  }
  render(c) {
    var n = this.items.length();
    for(var i = 0; i < n; ++i) {
      var item = this.items.get(i);
      
      drawIMButton(c, item.mlr, item.lch+1, item.ims.ml);
      drawIMButton(c, item.mrr, item.lch+2, item.ims.mr);
      drawIMButton(c, item.plr, item.lch+1, item.ims.pl);
      drawIMButton(c, item.prr, item.lch+2, item.ims.pr);
    }
    c.font = "300 14px Arial";
    c.textAlign = 'center';
    c.strokeStyle = c.fillStyle = kEQLabel;
    c.fillText("Mic/Line/Inst", this.l + 230, this.t + 50);
    c.fillText("Line In", this.l + 425, this.t + 50);
    c.fillText("Main Out 1-2", this.l + 135, this.t +81);
    c.fillText("Phones", this.l + 135, this.t +137);

    drawLabelGroup(c, "Input Monitoring", this.l+2, this.t+2, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);

    c.lineWidth = 1;
    c.beginPath();
    c.moveTo(this.l + 296, this.t + 44);
    c.lineTo(this.l + 395, this.t + 44);
    c.moveTo(this.l + 456, this.t + 44);
    c.lineTo(this.l + 560, this.t + 44);
    c.moveTo(this.l + 40, this.t + 104);
    c.lineTo(this.l + 560, this.t + 104);
    c.stroke();

    kSpeakerImg.render(c, this.l + 45, this.t + 48);
    kPhonesImg.render(c, this.l + 38, this.t + 112);
    kAlertImg.render(c, this.l + 140, this.t + 153);
    c.restore();

    drawInfoIcon(c, this.infoR.cx, this.infoR.cy, kEQLabel);

    var exp1 = "Click an input to hear it directly on the Main Mix or Phones Mix."
    var exp2 = "When doing so, turn off input monitoring in your host software."

    c.font = "600 11px Arial";
    c.textAlign = 'left';
    c.fillText(exp1, this.l + 192, this.t + 169);
    c.fillText(exp2, this.l + 192, this.t + 186);

    if(this.showInfoMessage) {
      roundRectPath(c, this.l+14, this.t+24, this.w-24, this.h-32, 6);
      c.fillStyle = 'black';
      c.fill()
      c.strokeStyle = kEQLabel;
      c.stroke()

      c.beginPath();
      c.moveTo(this.xR.l, this.xR.t);
      c.lineTo(this.xR.r, this.xR.b);
      c.moveTo(this.xR.r, this.xR.t);
      c.lineTo(this.xR.l, this.xR.b);
      c.stroke()

      var info1l1 = "Click an input to hear it on the Main Outs or Headphone Output.";
      var info1l2 = "For more control over the monitor mix, use the Main Out 1-2 Mix";
      var info1l3 = "and Phones 1-2 Mix tabs."

      var info2l1 = "If you make adjustments to an input in its mix tab channel strip,"
      var info2l2 = "the button here will turn gray.";

      var info3l1 = "When using these input monitoring features in the UltraLite-mk5,";
      var info3l2 = "be sure to disable input monitoring in your host software.";
      var info3l3 = "Otherwise, the input signal will get doubled on the output."; 

      c.font = "12px Arial";
      c.fillStyle = kEQLabel;
      var x = this.l + 152;
      var y = this.t + 30;
      c.fillText(info1l1, x, y += 16);
      c.fillText(info1l2, x, y += 16);
      c.fillText(info1l3, x, y += 16);

      c.fillText(info2l1, x, y += 24);
      c.fillText(info2l2, x, y += 16);

      c.fillText(info3l1, x, y += 24);
      c.fillText(info3l2, x, y += 16);
      c.fillText(info3l3, x, y += 16);

      drawIMButton(c, new Rect(this.l + 110, this.t+40, 25, 25), 1, 1);
      drawIMButton(c, new Rect(this.l + 110, this.t+94, 25, 25), 1, 2);
      kAlertImg.render(c, this.l + 102, this.t + 136);
    }
  }
}


class Home extends View{
  constructor() {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot();
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    evSys.addListener('wResized', this.resized);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "homeContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.width = "calc(100% - " + xPos + "px)";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "home";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;
    
    this.tabDiv = this.scrollBarVH.containerDiv;

    //this.c.addSub(this.dnh = new DevNameHeader());
    this.m1 = new MicPreHome(this, 38, 350, 0);
    this.m2 = new MicPreHome(this, this.m1.r + 25, 350, 1);
    this.mv = new MainVolGroupHome(this, this.m2.r + 25, 350);
    this.pv = new PhonesVolHome(this, this.mv.r + 25, 350);
    this.addSub(this.im = new InputMonitoring(this.m1.l, this.m1.b + 20));
    this.setBG(kEmptyBG);
  }
  resized() {
    var w = kULm5BG.width;
    var h = this.im.b + 20;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
    this.requestRender();
  }
  render(c) {
    drawDeviceNameHeader(c, this.l + 35, this.t, 630);
    kULm5BG.render(c, this.l, this.t + 50);
  }
}
