'use strict';

// n = default name
// ni = name index in kOutputNames
// osi = output source index in kFPGAPatch (index for output meter and routing source)
// lo = index in koTrim and bit in kMainGroup
const kOutChInfo = [
  {n:"Main 1",     ni:0,  osi:kOD_Main1,   lo:kLO_Main1 },
  {n:"Main 2",     ni:1,  osi:kOD_Main2,   lo:kLO_Main2 },
  {n:"Line 3",     ni:2,  osi:kOD_Line3,   lo:kLO_Line3 },
  {n:"Line 4",     ni:3,  osi:kOD_Line4,   lo:kLO_Line4 },
  {n:"Line 5",     ni:4,  osi:kOD_Line5,   lo:kLO_Line5 },
  {n:"Line 6",     ni:5,  osi:kOD_Line6,   lo:kLO_Line6 },
  {n:"Line 7",     ni:6,  osi:kOD_Line7,   lo:kLO_Line7 },
  {n:"Line 8",     ni:7,  osi:kOD_Line8,   lo:kLO_Line8 },
  {n:"Line 9",     ni:8,  osi:kOD_Line9,   lo:kLO_Line9 },
  {n:"Line 10",    ni:9,  osi:kOD_Line10,  lo:kLO_Line10 },
  {n:"Phones 1",   ni:10, osi:kOD_Phones1, lo:kLO_Phones1 },
  {n:"Phones 2",   ni:11, osi:kOD_Phones2, lo:kLO_Phones2 },
  {n:"S/PDIF 1",   ni:12, osi:kOD_SPDIF1,  lo:kLO_SPDIF1 },
  {n:"S/PDIF 2",   ni:13, osi:kOD_SPDIF2,  lo:kLO_SPDIF2 },
  {n:"ADAT 1",     ni:14, osi:kOD_ADAT1,   lo:-1 },
  {n:"ADAT 2",     ni:15, osi:kOD_ADAT2,   lo:-1 },
  {n:"ADAT 3",     ni:16, osi:kOD_ADAT3,   lo:-1 },
  {n:"ADAT 4",     ni:17, osi:kOD_ADAT4,   lo:-1 },
  {n:"ADAT 5",     ni:18, osi:kOD_ADAT5,   lo:-1 },
  {n:"ADAT 6",     ni:19, osi:kOD_ADAT6,   lo:-1 },
  {n:"ADAT 7",     ni:20, osi:kOD_ADAT7,   lo:-1 },
  {n:"ADAT 8",     ni:21, osi:kOD_ADAT8,   lo:-1 },
];     

// n = default name
// ni = name index in kOutputNames
// m = meter index
const kADCInChInfo = [
  {n:"Mic/Line/Inst 1", ni:0,  m:kIS_Mic1 },
  {n:"Mic/Line/Inst 2", ni:1,  m:kIS_Mic2 },
  {n:"Line 3",     ni:2,  m:kIS_Line3 },
  {n:"Line 4",     ni:3,  m:kIS_Line4 },
  {n:"Line 5",     ni:4,  m:kIS_Line5 },
  {n:"Line 6",     ni:5,  m:kIS_Line6 },
  {n:"Line 7",     ni:6,  m:kIS_Line7 },
  {n:"Line 8",     ni:7,  m:kIS_Line8 },
  {n:"S/PDIF 1",   ni:8,  m:kIS_SPDIF1 },
  {n:"S/PDIF 2",   ni:9,  m:kIS_SPDIF2 },
  {n:"ADAT 1",     ni:10, m:kIS_ADAT1 },
  {n:"ADAT 2",     ni:11, m:kIS_ADAT2 },
  {n:"ADAT 3",     ni:12, m:kIS_ADAT3 },
  {n:"ADAT 4",     ni:13, m:kIS_ADAT4 },
  {n:"ADAT 5",     ni:14, m:kIS_ADAT5 },
  {n:"ADAT 6",     ni:15, m:kIS_ADAT6 },
  {n:"ADAT 7",     ni:16, m:kIS_ADAT7 },
  {n:"ADAT 8",     ni:17, m:kIS_ADAT8 }
];

// helper to get input and output names
function getIName(i) {
  var n = "Unknown";
  var left = i & ~1;
  if (ds.getKVArrayItem('kiMixStereo', left)) {
    n = ds.getKVArrayItem('kiChannelNames', left);
    if((n == undefined) || n == "")
      n = kADCInChInfo[i].n;
    else
      n += ((left == i) ? ' (L)' : ' (R)');
  }
  else {
    n = ds.getKVArrayItem('kiChannelNames', i);
    if((n == undefined) || n == "")
      n = kADCInChInfo[i].n;
  }
  return n;
}
function getOName(i) {
  var n = "Unknown";
  var left = i & ~1;
  if (ds.getKVArrayItem('koMixStereo', left)) {
    n = ds.getKVArrayItem('koChannelNames', left);
    if((n == undefined) || n == "")
      n = kOutChInfo[left].n + '-' + String(left + 2);
  }
  else {
    n = ds.getKVArrayItem('koChannelNames', i);
    if((n == undefined) || n == "")
      n = kOutChInfo[i].n;
  }
  return n + ' Mix';
}

class IOMeter extends View {
  constructor(l,t,w,h,index) {
    super(l,t,w,h);
    this.onMeters = this.onMeters.bind(this);
    this.setBG(kBlackBG);
    this.lm = {m:0, ph:0};
    this.index = index;
    meters.addListener(this.onMeters);
  }
  render(c) {
    var l = this.l + 1;
    var w = this.w - 2;
    var sh = Math.floor(this.lm.m * this.h);
    var sy = this.h - sh;
    if(sh != 0) {
      c.fillStyle = kMeterBlue;
      c.fillRect(l, this.t + sy, w, sh)
    }

    var y = this.h - (this.lm.ph * this.h);
    if(y < (this.h - 2)) {
      c.fillStyle = "red";
      c.fillRect(l, this.t + y, w, 2);
    }
  }
  onMeters(v) {
    if((this.index >= 0) && (this.index < v.length)) {
      var m = Calc0dBToCtrl24(v[this.index].m);
      var ph = Calc0dBToCtrl24(v[this.index].ph);
      if((this.lm.m != m) || (this.lm.ph != ph)) {
        this.lm.m = m;
        this.lm.ph = ph;
        this.requestRender();
      }
    }
  }
}

const kPreTrimCtrlInfo = [
  {k: 'kiGain', d:0, u:"dB", min:0, max:74, dft:0 },
];
const kLineInTrimCtrlInfo = [
  {k: 'kiGain', d:0, u:"dB", min:0, max:20, dft:0 },
];
const kOutputTrimCtrlInfo = [
  {k: 'koTrim', d:0, u:"dB", min:100, max:0, dft:0 },
];
const kPhonesTrimCtrlInfo = [
  {k: 'koTrim', d:0, u:"dB", min:100, max:0, dft:24 },
];
const kMainVolCtrlInfo = [
  {k: 'kMainTrim', d:0, u:"dB", min:100, max:0, dft:0 },
];

class IOKnobTextEdit extends EditableTextView {
  constructor(l,t,w,h, ctrlInfo) {
    super(l,t,w,h, 'eq-input-detail', '12px sans-serif');
    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);

    this.ctrlInfo = ctrlInfo;
    this.setBG(kEmptyBG);
    this.c = kEQLabel;

    SetupDSArrayListenerBasic(this, ctrlInfo.k, 0);
  };
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    var v = DSArrayGetBasic(this);
    v = this.ctrlInfo.warp ? GainToDB(v) : v; 
    this.v = trimFloat(v, this.ctrlInfo.d) + " " + this.ctrlInfo.u;
    this.requestRender();
  }
  onDSArray(k, v) {
    v = this.ctrlInfo.warp ? GainToDB(v) : v;
    v = trimFloat(v, this.ctrlInfo.d);
    if((this.ctrlInfo.k == 'kMainTrim') || (this.ctrlInfo.k == 'koTrim')) {
      if (v == this.ctrlInfo.min)
        v = '-∞';
      else if (v != this.ctrlInfo.max)
        v = '-' + v;
    }
    this.v = v + ' ' + this.ctrlInfo.u;
    this.requestRender();
  }
  valueChanged(v) {
    v = v.replace(/[∞]|inf/g, this.ctrlInfo.min);
    v = v.replace(/[^0-9\.\-]/g, '');
    if ((v.length == 0) || isNaN(v))
      return this.v;
    v = validateCtrl(this.ctrlInfo, Number(v));
    DSArraySetBasic(this, this.ctrlInfo.warp ? DBToGain(v) : v);
    if((this.ctrlInfo.k == 'kMainTrim') || (this.ctrlInfo.k == 'koTrim')) {
      if (v == this.ctrlInfo.min)
        v = '-∞';
      else if (v != this.ctrlInfo.max)
        v = '-' + v;
    }
    return v + ' ' + this.ctrlInfo.u;
  } 
}

class IOKnob extends SimpleDSKnob {
  constructor(l, t, cE, cD, ctrlInfo, iL, iR) {
    super(l, t, 68, 75, ctrlInfo.k, iL);

    this.setup = this.setup.bind(this);
    this.enable = this.enable.bind(this);

    this.cr = new Rect(l + 2, t, 64, 70);
    this.cE = cE;
    this.cD = cD;
    this.setBG(kEmptyBG);
    this.ctrlInfo = ctrlInfo;
    this.enabled = true;
    this.iR = iR;

    this.addSub(this.edit = new IOKnobTextEdit(this.l, this.t + (this.h - 18), this.w, 18, ctrlInfo));

    this.setup(iL);
  }
  setup(i) {
    super.setup(i);
    this.edit.setup(i);
  }
  offset(dx, dy) {
    this.cr.offset(dx,dy);
    super.offset(dx, dy);
  }
  enable(enable) {
    if(this.enabled != enable) {
      this.enabled = enable;
      this.requestRender();
    }
  }
  handleDblClick() { 
    this.setVal(this.ctrlInfo.dft);
  }
  dsToV(dsv) { 
    return dynDSToV(dsv, this.ctrlInfo);
  }
  vToDS(v) { 
    return dynVToDS(v, this.ctrlInfo); 
  }
  setVal(v) {
    // overridden so we can change both trims when the knob is stereo moved
    var i = this[this.ctrlInfo.k];
    if(i != undefined) {
      ds.setV(this.ctrlInfo.k, i, v);
      if(this.iR)
        ds.setV(this.ctrlInfo.k, this.iR, v);
    }
  }
  render(c) {
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, 2.25 * Math.PI);
    c.lineWidth = 7;
    c.strokeStyle = kDynCtrlBG;
    c.stroke();

    var v = (0.75 + this.v * 1.5) * Math.PI;
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, v);
    c.lineWidth = 7;
    c.strokeStyle = this.enabled ? this.cE : this.cD;
    c.stroke();
  }
}


class InPrePostFXMenu extends View {
  constructor(p, l, t, w, info) {
    super(l,t,w,40);

    this.change = this.change.bind(this);
    this.fpgaPatchChanged = this.fpgaPatchChanged.bind(this);
    this.loopbackFirstChanged = this.loopbackFirstChanged.bind(this);

    this.loopbackFirst = ds.getKVArrayItem('kLoopbackFirst', 0);
    this.info = info;
    this.setBG(kEmptyBG);

    p.addSub(this);
    ds.addListener('kFPGAPatch', this.fpgaPatchChanged);
    ds.addListener('kLoopbackFirst', this.loopbackFirstChanged);
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l, t+12, this.w, 28, this.change));
    this.menu.addOption(this.info.i, "Pre Fx");
    this.menu.addOption(this.info.ifx, "Post Fx");
    this.menu.sel.value = ds.getKVArrayItem('kFPGAPatch', this.info.o + (this.loopbackFirst * 2));
  }
  fpgaPatchChanged(k, v) {
    var vl = v[this.info.o + (this.loopbackFirst * 2)];
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
      this.requestRender();
    }
  }
  loopbackFirstChanged(k, v) {
    this.loopbackFirst = v[0];
  }
  change() {
    var vl = parseInt(this.menu.sel.value);
    ds.setV('kFPGAPatch', this.info.o + (this.loopbackFirst * 2), vl);
  }
  render(c) {
    c.fillStyle = kEQLabel;
    c.font = "300 14px Arial";
    c.textAlign = 'left';
    c.fillText("To USB host:", this.l+2, this.t+8);
  }
}

const kUSBInRouting = [
  { o:kOD_TDM1, i:kIS_Mic1,  ifx:kIS_DSP5 },
  { o:kOD_TDM2, i:kIS_Mic2,  ifx:kIS_DSP6 },
  { o:kOD_TDM3, i:kIS_Line3, ifx:kIS_DSP7 },
  { o:kOD_TDM4, i:kIS_Line4, ifx:kIS_DSP8 },
  { o:kOD_TDM5, i:kIS_Line5, ifx:kIS_DSP13 },
  { o:kOD_TDM6, i:kIS_Line6, ifx:kIS_DSP14 },
  { o:kOD_TDM7, i:kIS_Line7, ifx:kIS_DSP15 },
  { o:kOD_TDM8, i:kIS_Line8, ifx:kIS_DSP16 }
];

class MicPre extends View {
  constructor(p, l, t, i) {
    super(l,t,180,230);
    p.addSub(this);
    this.onJack = this.onJack.bind(this);
    this.chIdx = i;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 126, kADCInChInfo[i].m));
    this.addSub(this.b48V = new ToggleBtn(l + 35, t + 27, 52, "48V", 'k48V', i, kToggleRed, kEmptyBG));
    this.addSub(this.pad = new ToggleBtn(l + 35, t + 73, 52, "Pad", 'kPad', i, kToggleBlue, kEmptyBG));
    this.addSub(new ToggleBtn(l + 35, t + 119, 52, "\u00D8", 'kiPhase', i, kToggleBlue, kEmptyBG));
    this.addSub(this.gain = new IOKnob(l + 100, t + 27, kToggleBlue, kGateCtrlDisabled, kPreTrimCtrlInfo[0], i));
    this.fxmenu = new InPrePostFXMenu(this, this.l+12, t+166, this.w-24, kUSBInRouting[i]);

    ds.addListener('kJackDetect', this.onJack);
  }  
  onJack(k, v) {
    if(k == 'kJackDetect') {
      // bool 0 = mic, 1 = line/inst
      this.b48V.enable(v[this.chIdx] == 0);
      this.pad.enable(v[this.chIdx] == 0);
    }
  }
  render(c) {
    drawDoubleLabelGroup(c, [getIName(this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}


class LineIn extends View {
  constructor(p, l, t, i) {
    super(l,t,113,230);
    p.addSub(this);
    this.chIdx = i;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 126, kADCInChInfo[i].m));
    this.addSub(this.gain = new IOKnob(l + 33, t + 27, kToggleBlue, kGateCtrlDisabled, kLineInTrimCtrlInfo[0], i));
    this.addSub(new ToggleBtn(l + 42, t + 119, 52, "\u00D8", 'kiPhase', i, kToggleBlue, kEmptyBG));
    this.fxmenu = new InPrePostFXMenu(this, this.l+12, t+166, this.w-24, kUSBInRouting[i]);
  }  
  render(c) {
    drawDoubleLabelGroup(c, [getIName(this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}


class OpticalIn extends View {
  constructor(p, l, t, m) {
    super(l,t,180,120);
    this.opticalChanged = this.opticalChanged.bind(this);

    p.addSub(this);
    this.wantsMouse = true;
    this.addSub(this.meter1 = new IOMeter(l + 74, this.t + 25, 10, 80, m));
    this.addSub(this.meter2 = new IOMeter(this.meter1.r + 2, this.t + 25, 10, 80, m+1));
    this.addSub(this.meter3 = new IOMeter(this.meter2.r + 2, this.t + 25, 10, 80, m+2));
    this.addSub(this.meter4 = new IOMeter(this.meter3.r + 2, this.t + 25, 10, 80, m+3));
    this.addSub(this.meter5 = new IOMeter(this.meter4.r + 2, this.t + 25, 10, 80, m+4));
    this.addSub(this.meter6 = new IOMeter(this.meter5.r + 2, this.t + 25, 10, 80, m+5));
    this.addSub(this.meter7 = new IOMeter(this.meter6.r + 2, this.t + 25, 10, 80, m+6));
    this.addSub(this.meter8 = new IOMeter(this.meter7.r + 2, this.t + 25, 10, 80, m+7));
    ds.addListener('kOpticalMode', this.opticalChanged);
    ds.addListener('kSampleRate', this.opticalChanged);
    this.opticalMode = 0;

    this.adatHitR = new Rect(l + 12, t + 27, 52, 30);
    this.tosHitR = new Rect(l + 12, t + 73, 52, 30);
  }
  opticalChanged() {
    this.opticalMode = ds.getKVArrayItem('kOpticalMode', 0);
    var sr = ds.getKVArrayItem('kSampleRate', 0);
    this.meter3.setVis(this.opticalMode == 0);
    this.meter4.setVis(this.opticalMode == 0);
    this.meter5.setVis((this.opticalMode == 0) && (sr <= 48000));
    this.meter6.setVis((this.opticalMode == 0) && (sr <= 48000));
    this.meter7.setVis((this.opticalMode == 0) && (sr <= 48000));
    this.meter8.setVis((this.opticalMode == 0) && (sr <= 48000));
    this.setW(this.opticalMode == 0 ? 180 : 110);
    this.setVis(sr <= 96000);
    this.requestRender();
  }  
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      if(this.adatHitR.contains(xy.x, xy.y) || this.tosHitR.contains(xy.x, xy.y))
        return this;
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) { 
    if(this.opticalMode != 0) {
      if(this.adatHitR.contains(xy.x, xy.y))
        ds.setV('kOpticalMode', 0, 0);
    } else {
      if(this.tosHitR.contains(xy.x, xy.y))
        ds.setV('kOpticalMode', 0, 1);
    }
  }
  render(c) {
    drawLabelGroup(c, "Optical", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    drawButton(c, this.adatHitR, "ADAT", this.opticalMode == 0, kToggleBlue);
    drawButton(c, this.tosHitR, "TOS", this.opticalMode == 1, kToggleBlue);
  }
}

class SPDIFIn extends View {
  constructor(p, l, t, m) {
    super(l,t,80,120);

    this.srChanged = this.srChanged.bind(this);

    p.addSub(this);
    this.addSub(this.meter1 = new IOMeter(l + 24, this.t + 25, 10, 80, m));
    this.addSub(this.meter2 = new IOMeter(this.meter1.r + 12, this.t + 25, 10, 80, m+1));

    ds.addListener('kSampleRate', this.srChanged);
  }
  srChanged() {
    this.setVis(ds.getKVArrayItem('kSampleRate', 0) <= 96000);
  }  
  render(c) {
    drawLabelGroup(c, "S/PDIF", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}


class Inputs extends View {
  constructor(root, l, t) {
    super(l, t, 635, 655);
    root.addSub(this);

    this.setBG(kEmptyBG);
    this.m1 = new MicPre(this, 38 + l, t + 44, 0);
    this.m2 = new MicPre(this, 38 + l, 10 + this.m1.b, 1);
    this.i1 = new LineIn(this, 20 + this.m1.r, t + 44, 2);
    this.i2 = new LineIn(this, 20 + this.m1.r, this.m1.b + 10, 3);
    this.i3 = new LineIn(this, 20 + this.i1.r, t + 44, 4);
    this.i4 = new LineIn(this, 20 + this.i1.r, this.m1.b + 10, 5);
    this.i5 = new LineIn(this, 20 + this.i3.r, t + 44, 6);
    this.i6 = new LineIn(this, 20 + this.i3.r, this.m1.b + 10, 7);
    this.spd = new SPDIFIn(this, l + 20, this.m2.b + 35, kIS_SPDIF1);
    this.opt = new OpticalIn(this, 20 + this.spd.r, this.m2.b + 35, kIS_ADAT1);
  }  
  render(c)
  {
    drawLabelGroup(c, "Input Gain", this.m1.l-18, this.m1.t-30, this.w-20, (this.i6.b - this.m1.t) + 50, 6, kEQLabel, kEQLabel);
  }
}

// Handle mouse clicks. 
class MouseView extends View {
  constructor(mousee, l, t, w, h) {
    super(l,t,w,h);
    this.mousee = mousee;
    this.wantsMouse = true;
  }
  handleMouseDown(xy) { 
    this.mousee.handleMouseDown(xy);
  }
}
class OpticalOut extends View {
  constructor(p, l, t) {
    super(l,t,421,120);

    this.change = this.change.bind(this);
    this.onDS = this.onDS.bind(this);
    this.rebuildUI = this.rebuildUI.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.setPatches = this.setPatches.bind(this);

    this.opticalMode = 0;
    var m = kOutChInfo[14].osi;

    p.addSub(this);
    this.meters = [];

    this.meters.push(new IOMeter(l + 84, this.t + 25, 10, 80, m));
    this.meters.push(new IOMeter(this.meters[0].r + 2, this.t + 25, 10, 80, m+1));
    this.meters.push(new IOMeter(this.meters[1].r + 6, this.t + 25, 10, 80, m+2));
    this.meters.push(new IOMeter(this.meters[2].r + 2, this.t + 25, 10, 80, m+3));
    this.meters.push(new IOMeter(this.meters[3].r + 6, this.t + 25, 10, 80, m+4));
    this.meters.push(new IOMeter(this.meters[4].r + 2, this.t + 25, 10, 80, m+5));
    this.meters.push(new IOMeter(this.meters[5].r + 6, this.t + 25, 10, 80, m+6));
    this.meters.push(new IOMeter(this.meters[6].r + 2, this.t + 25, 10, 80, m+7));
    for(var i = 0; i < 8; ++i)
      this.addSub(this.meters[i]);

    // The optical menu won't get mouse clicks unless I do this
    this.addSub(this.mouser = new MouseView(this, l + 17, t + 27, 52, 73));
    this.adatHitR = new Rect(l + 17, t + 27, 52, 30);
    this.tosHitR = new Rect(l + 17, t + 73, 52, 30);

    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.meters[7].r + 20 , this.t + 60, 190, 30, this.change));

    ds.addListener('kOpticalMode', this.onDS);
    ds.addListener('kFPGAPatch', this.onDS);
    ds.addListener('kSampleRate', this.onDS);
    ds.addListener('koChannelNames', this.labelsChanged);
    ds.addListener('koMixStereo', this.labelsChanged);

    this.rebuildUI();
  }
  setPatches(v) {
    // set the datastore patches given the first patch
    if(this.opticalMode == 0) {
      // adat routing can be from mic-line8 or host optical 1-8
      if(v != kIS_Mic1) 
        v = kIS_TDM15;

      if(v == kIS_Mic1) {
        // inputs are not contiguous
        for(var i = 0; i < 8; ++i)
          ds.setV('kFPGAPatch', kOD_ADAT1 + i, kADCInChInfo[i].m);  
      }
      else {
        for(var i = 0; i < 8; ++i)
          ds.setV('kFPGAPatch', kOD_ADAT1 + i, v + i);  
      }
    }
    else {
      ds.setV('kFPGAPatch', kOD_ADAT1, v);
      ds.setV('kFPGAPatch', kOD_ADAT2, v+1);
    }
  }
  rebuildUI() {
    // setup the meters and menus
    this.menu.clear();
    var v = ds.getKVArrayItem('kFPGAPatch', kOD_ADAT1);
    var n = this.opticalMode == 0 ? 8 : 2;

    if(this.opticalMode == 0) {
      // ADAT
      this.menu.addOption(kIS_TDM15, "USB Out Optical (From Host)"); // first item is adat from host kIS_TDM15
      this.menu.addOption(kIS_Mic1, "Line Inputs (Optical Expansion)"); // first item is Mic 1 through Line 8 kIS_Mic1 - kIS_Line8
      if(v == kIS_Mic1) {
        for(var i = 0; i < 8; ++i) 
          this.meters[i].index = kADCInChInfo[i].m;
      } else {
        v = kIS_TDM15;
        for(var i = 0; i < 8; ++i) 
          this.meters[i].index = kIS_TDM15 + i;
      }
      var is1x = ds.getKVArrayItem('kSampleRate', 0) <= 48000;
      this.meters[2].setVis(true);
      this.meters[3].setVis(true);
      this.meters[4].setVis(is1x);
      this.meters[5].setVis(is1x);
      this.meters[6].setVis(is1x);
      this.meters[7].setVis(is1x);
    }
    else {
      // TOSLink
      this.updateLabels();
      this.menu.addOption(kIS_TDM15, this.labels[0]);
      this.menu.addOption(kIS_DSP1, this.labels[1]);
      this.menu.addOption(kIS_DSP19, this.labels[2]);
      this.menu.addOption(kIS_DSP3, this.labels[3]);
      this.menu.addOption(kIS_DSP9, this.labels[4]);
      this.menu.addOption(kIS_DSP11, this.labels[5]);
      this.menu.addOption(kIS_DSP17, this.labels[6]);
      this.menu.addOption(kIS_DSP21, this.labels[7]);
      if((v != kIS_TDM15) && (v != kIS_DSP1) && (v != kIS_DSP19) && (v != kIS_DSP3) &&
        (v != kIS_DSP9) && (v != kIS_DSP11) && (v != kIS_DSP17) && (v != kIS_DSP21)) {
        v = kIS_TDM15;
      }
      this.meters[0].index = v;  
      this.meters[1].index = v+1;  
      for(var i = 2; i < 8; ++i) 
        this.meters[i].setVis(false);
    }
    this.menu.sel.value = v;
    this.requestRender();
  }
  updateLabels() {
    this.labels = [
      "USB Out Optical 1-2",
      getOutputPairChName(0),
      getOutputPairChName(10),
      getOutputPairChName(2),
      getOutputPairChName(4),
      getOutputPairChName(6),
      getOutputPairChName(8),
      "Reverb Dry Mix (L-R)"
    ];
  }
  labelsChanged(k, v) {
    if (this.opticalMode == 1) {
      this.updateLabels();
      this.menu.renameOptions(this.labels);
    }
  }
  onDS() {
    // datastore patches changed or
    // datastore optical mode changed, rebuild the menu and meters
    var isVis = ds.getKVArrayItem('kSampleRate', 0) <= 96000;
    this.setVis(isVis);
    this.menu.sel.style.display = isVis ? 'block' : 'none';
    this.opticalMode = ds.getKVArrayItem('kOpticalMode', 1);
    this.setPatches(ds.getKVArrayItem('kFPGAPatch', kOD_ADAT1));
    this.rebuildUI();
  }  
  change() {
    // menu changed
    this.setPatches(parseInt(this.menu.sel.value));
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      if(this.adatHitR.contains(xy.x, xy.y) || this.tosHitR.contains(xy.x, xy.y))
        return this;
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) { 
    if(this.opticalMode != 0) {
      if(this.adatHitR.contains(xy.x, xy.y))
        ds.setV('kOpticalMode', 1, 0);
    } else {
      if(this.tosHitR.contains(xy.x, xy.y))
        ds.setV('kOpticalMode', 1, 1);
    }
  }
  render(c) {
    drawLabelGroup(c, "Optical Out", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
    drawButton(c, this.adatHitR, "ADAT", this.opticalMode == 0, kToggleBlue);
    drawButton(c, this.tosHitR, "TOS", this.opticalMode == 1, kToggleBlue);
  }
}

class SPDIFOut extends View {
  constructor(p, l, t) {
    super(l,t,245,120);

    this.change = this.change.bind(this);
    this.fpgaPatchChanged = this.fpgaPatchChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.srChanged = this.srChanged.bind(this);

    var m = kOutChInfo[12].osi;

    p.addSub(this);
    this.addSub(this.meter1 = new IOMeter(l + 17, this.t + 25, 10, 80, m));
    this.addSub(this.meter2 = new IOMeter(this.meter1.r + 2, this.t + 25, 10, 80, m+1));
    
    this.updateLabels();

    ds.addListener('kFPGAPatch', this.fpgaPatchChanged);
    ds.addListener('koChannelNames', this.labelsChanged);
    ds.addListener('koMixStereo', this.labelsChanged);

    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l + 60, this.t + 60, 160, 30, this.change));
    this.menu.addOption(kIS_TDM13, this.labels[0]);
    this.menu.addOption(kIS_DSP1, this.labels[1]);
    this.menu.addOption(kIS_DSP19, this.labels[2]);
    this.menu.addOption(kIS_DSP3, this.labels[3]);
    this.menu.addOption(kIS_DSP9, this.labels[4]);
    this.menu.addOption(kIS_DSP11, this.labels[5]);
    this.menu.addOption(kIS_DSP17, this.labels[6]);
    this.menu.addOption(kIS_DSP21, this.labels[7]);
    this.menu.sel.value = ds.getKVArrayItem('kFPGAPatch', kOD_SPDIF1);
    ds.addListener('kSampleRate', this.srChanged);
  }
  updateLabels() {
    this.labels = [
      "USB Out SPDIF 1-2",
      getOutputPairChName(0),
      getOutputPairChName(10),
      getOutputPairChName(2),
      getOutputPairChName(4),
      getOutputPairChName(6),
      getOutputPairChName(8),
      "Reverb Dry Mix (L-R)"
    ];
  }
  fpgaPatchChanged(k, v) {
    var vl = v[kOD_SPDIF1];
    var vr = v[kOD_SPDIF2];
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
      this.meter1.index = vl;
      this.meter2.index = vr;
      this.requestRender();
    }
  }
  labelsChanged(k, v) {
    this.updateLabels();
    this.menu.renameOptions(this.labels);
  }
  srChanged() {
    var isVis = ds.getKVArrayItem('kSampleRate', 0) <= 96000;
    this.setVis(isVis);
    this.menu.sel.style.display = isVis ? 'block' : 'none';
  }  
  change() {
    var vl = parseInt(this.menu.sel.value);
    var vr =  vl + 1;
    ds.setV('kFPGAPatch', kOD_SPDIF1, vl);
    ds.setV('kFPGAPatch', kOD_SPDIF2, vr);
    this.meter1.index = vl;
    this.meter2.index = vr;
  }
  render(c) {
    drawLabelGroup(c, "S/PDIF Out", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
  }
}

class LineOut extends View {
  constructor(p, l, t, i) {
    super(l,t,113,120);
    p.addSub(this);
    this.chIdx = i;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 80, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[i].osi)));
    this.addSub(this.trim = new IOKnob(l + 35, t + 27, kToggleBlue, kGateCtrlDisabled, kOutputTrimCtrlInfo[0], kOutChInfo[i].lo));
  }  
  render(c) {
    drawDoubleLabelGroup(c, [getOName(this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}

class PhonesVol extends View {
  constructor(p, l, t) {
    super(l,t,376,120);
    this.change = this.change.bind(this);
    this.phonesSourceChanged = this.phonesSourceChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    p.addSub(this);
    this.chIdx = 10;
    this.updateLabels();
    this.addSub(this.m1 = new IOMeter(this.l + 17, this.t + 25, 10, 80, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[10].osi)));
    this.addSub(this.m2 = new IOMeter(this.m1.r + 2, this.t + 25, 10, 80, ds.getKVArrayItem('kFPGAPatch', kOutChInfo[11].osi)));
    this.addSub(this.trim = new IOKnob(l + 53, t + 27, kToggleBlue, kGateCtrlDisabled, kPhonesTrimCtrlInfo[0], kOutChInfo[10].lo, kOutChInfo[11].lo));
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l + 145, this.t + 60, 210, 30, this.change));
    this.menu.addOption(kIS_TDM1, this.labels[0], this.disabled[0]);
    this.menu.addOption(kIS_TDM11, this.labels[1], this.disabled[1]);
    this.menu.addOption(kIS_TDM3, this.labels[2], this.disabled[2]);
    this.menu.addOption(kIS_TDM5, this.labels[3], this.disabled[3]);
    this.menu.addOption(kIS_TDM7, this.labels[4], this.disabled[4]);
    this.menu.addOption(kIS_TDM9, this.labels[5], this.disabled[5]);
    this.menu.addOption(kIS_TDM13, this.labels[6], this.disabled[6]);
    this.menu.addOption(kIS_TDM15, this.labels[7], this.disabled[7]);
    this.menu.addOption(kIS_TDM17, this.labels[8], this.disabled[8]);
    this.menu.addOption(kIS_TDM19, this.labels[9], this.disabled[9]);
    this.menu.addOption(kIS_TDM21, this.labels[10], this.disabled[10]);
    this.menu.addOption(kIS_DSP1, this.labels[11], this.disabled[11]);
    this.menu.addOption(kIS_DSP19, this.labels[12], this.disabled[12]);
    this.menu.addOption(kIS_DSP3, this.labels[13], this.disabled[13]);
    this.menu.addOption(kIS_DSP9, this.labels[14], this.disabled[14]);
    this.menu.addOption(kIS_DSP11, this.labels[15], this.disabled[15]);
    this.menu.addOption(kIS_DSP17, this.labels[16], this.disabled[16]);
    this.menu.addOption(kIS_DSP21, this.labels[17], this.disabled[17]);
    this.menu.addOption(kIS_DSP23, this.labels[18], this.disabled[18]);
    this.menu.sel.value = ds.getKVArrayItem('kFPGAPatch', kOD_Phones1);
    ds.addListener('kFPGAPatch', this.phonesSourceChanged);
    ds.addListener('koChannelNames', this.labelsChanged);
    ds.addListener('koMixStereo', this.labelsChanged);
    ds.addListener('kSampleRate', this.labelsChanged);
  }  
  render(c) {
    drawLabelGroup(c, "Phones", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
  }
  updateLabels() {
    var sr = ds.getKVArrayItem('kSampleRate', 0);
    var is4x = (sr > 96000);
    var is2x4x = (sr > 48000);
    var unavailable4x = is4x ? " (Unavailable)" : "";
    var unavailable2x4x = is2x4x ? " (Unavailable)" : "";
    this.labels = [
      "USB Out Main 1-2",
      "USB Out Phones 1-2" + unavailable4x,
      "USB Out Line 3-4",
      "USB Out Line 5-6",
      "USB Out Line 7-8",
      "USB Out Line 9-10",
      "USB Out SPDIF 1-2" + unavailable4x,
      "USB Out Optical 1-2" + unavailable4x,
      "USB Out Optical 3-4" + unavailable4x,
      "USB Out Optical 5-6" + unavailable2x4x,
      "USB Out Optical 7-8" + unavailable2x4x,
      getOutputPairChName(0),
      getOutputPairChName(10) + " (Default)",
      getOutputPairChName(2),
      getOutputPairChName(4),
      getOutputPairChName(6),
      getOutputPairChName(8),
      "Reverb Dry Mix (L-R)" + unavailable4x,
      "Reverb Wet Mix (L-R)" + unavailable4x
    ];
    this.disabled = [
      false,
      is4x,
      false,
      false,
      false,
      false,
      is4x,
      is4x,
      is4x,
      is2x4x,
      is2x4x,
      false,
      false,
      false,
      false,
      false,
      false,
      is4x,
      is4x
    ];
  }
  phonesSourceChanged(k, v) {
    var vl = v[kOD_Phones1];
    var vr = v[kOD_Phones2];
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
      this.requestRender();
    }
    this.m1.index = vl;
    this.m2.index = vr;
  }
  labelsChanged(k, v) {
    this.updateLabels();
    this.menu.renameOptions(this.labels, this.disabled);
  }
  change() {
    var vl = parseInt(this.menu.sel.value);
    var vr =  vl + 1;
    ds.setV('kFPGAPatch', kOD_Phones1, vl);
    ds.setV('kFPGAPatch', kOD_Phones2, vr);
    this.m1.index = vl;
    this.m2.index = vr;
  }
}

class Loopback extends View {
  constructor(p, l, t) {
    super(l,t,290,120);
    this.change = this.change.bind(this);
    this.loopbackSourceChanged = this.loopbackSourceChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    p.addSub(this);
    this.updateLabels();
    this.addSub(this.m1 = new IOMeter(this.l + 17, this.t + 25, 10, 80, ds.getKVArrayItem('kLoopbackSource', 0)));
    this.addSub(this.m2 = new IOMeter(this.m1.r + 2, this.t + 25, 10, 80, ds.getKVArrayItem('kLoopbackSource', 1)));
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l + 60, this.t + 60, 210, 30, this.change));
    this.menu.addOption(kIS_TDM1, this.labels[0], this.disabled[0]);
    this.menu.addOption(kIS_TDM11, this.labels[1], this.disabled[1]);
    this.menu.addOption(kIS_TDM3, this.labels[2], this.disabled[2]);
    this.menu.addOption(kIS_TDM5, this.labels[3], this.disabled[3]);
    this.menu.addOption(kIS_TDM7, this.labels[4], this.disabled[4]);
    this.menu.addOption(kIS_TDM9, this.labels[5], this.disabled[5]);
    this.menu.addOption(kIS_TDM13, this.labels[6], this.disabled[6]);
    this.menu.addOption(kIS_TDM15, this.labels[7], this.disabled[7]);
    this.menu.addOption(kIS_TDM17, this.labels[8], this.disabled[8]);
    this.menu.addOption(kIS_TDM19, this.labels[9], this.disabled[9]);
    this.menu.addOption(kIS_TDM21, this.labels[10], this.disabled[10]);
    this.menu.addOption(kIS_DSP1, this.labels[11], this.disabled[11]);
    this.menu.addOption(kIS_DSP19, this.labels[12], this.disabled[12]);
    this.menu.addOption(kIS_DSP3, this.labels[13], this.disabled[13]);
    this.menu.addOption(kIS_DSP9, this.labels[14], this.disabled[14]);
    this.menu.addOption(kIS_DSP11, this.labels[15], this.disabled[15]);
    this.menu.addOption(kIS_DSP17, this.labels[16], this.disabled[16]);
    this.menu.addOption(kIS_DSP21, this.labels[17], this.disabled[17]);
    this.menu.addOption(kIS_DSP23, this.labels[18], this.disabled[18]);
    this.menu.sel.value = ds.getKVArrayItem('kLoopbackSource', 0);
    ds.addListener('kLoopbackSource', this.loopbackSourceChanged);
    ds.addListener('koChannelNames', this.labelsChanged);
    ds.addListener('koMixStereo', this.labelsChanged);
    ds.addListener('kSampleRate', this.labelsChanged);
  }  
  render(c) {
    drawLabelGroup(c, "Loopback", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
  }
  updateLabels() {
    var sr = ds.getKVArrayItem('kSampleRate', 0);
    var is4x = (sr > 96000);
    var is2x4x = (sr > 48000);
    var unavailable4x = is4x ? " (Unavailable)" : "";
    var unavailable2x4x = is2x4x ? " (Unavailable)" : "";
    this.labels = [
      "USB Out Main 1-2",
      "USB Out Phones 1-2" + unavailable4x,
      "USB Out Line 3-4",
      "USB Out Line 5-6",
      "USB Out Line 7-8",
      "USB Out Line 9-10",
      "USB Out SPDIF 1-2" + unavailable4x,
      "USB Out Optical 1-2" + unavailable4x,
      "USB Out Optical 3-4" + unavailable4x,
      "USB Out Optical 5-6" + unavailable2x4x,
      "USB Out Optical 7-8" + unavailable2x4x,
      getOutputPairChName(0),
      getOutputPairChName(10) + " (Default)",
      getOutputPairChName(2),
      getOutputPairChName(4),
      getOutputPairChName(6),
      getOutputPairChName(8),
      "Reverb Dry Mix (L-R)" + unavailable4x,
      "Reverb Wet Mix (L-R)" + unavailable4x
    ];
    this.disabled = [
      false,
      is4x,
      false,
      false,
      false,
      false,
      is4x,
      is4x,
      is4x,
      is2x4x,
      is2x4x,
      false,
      false,
      false,
      false,
      false,
      false,
      is4x,
      is4x
    ];
  }
  loopbackSourceChanged(k, v) {
    var vl = v[0];
    var vr = v[1];
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
      this.requestRender();
    }
    this.m1.index = vl;
    this.m2.index = vr;
  }
  labelsChanged(k, v) {
    this.updateLabels();
    this.menu.renameOptions(this.labels, this.disabled);
  }
  change() {
    var vl = parseInt(this.menu.sel.value);
    var vr =  vl + 1;
    ds.setV('kLoopbackSource', 0, vl);
    ds.setV('kLoopbackSource', 1, vr);
    this.m1.index = vl;
    this.m2.index = vr;
  }
}

class MainVolGroupEnableMultiSelect extends SimpleDSView {
  constructor(l, t) {
    super(l, t, 685, 120, 'kMainGroup', 0);

    this.setBG(kEmptyBG);
    this.wantsMouse = true;
    this.items = new List();

    this.addSub(this.vol = new IOKnob(l + 53, t + 27, kToggleBlue, kGateCtrlDisabled, kMainVolCtrlInfo[0], 0));

    var w1 = 33;
    var w2 = w1 + 50;
    var t1 = this.t + 70;
    var l1 = this.l + 185;
    var l2 = l1 + w2;
    var l3 = l2 + w2;
    var l4 = l3 + w2;
    var l5 = l4 + w2;
    //this.items.add({ r:(new Rect(l1, t1, 25, 25)), i:0 });
    //this.items.add({ r:(new Rect(l1 + w1, t1, 25, 25)), i:1 });
    this.items.add({ r:(new Rect(l2, t1, 25, 25)), i:2 });
    this.items.add({ r:(new Rect(l2 + w1, t1, 25, 25)), i:3 });
    this.items.add({ r:(new Rect(l3, t1, 25, 25)), i:4 });
    this.items.add({ r:(new Rect(l3 + w1, t1, 25, 25)), i:5 });
    this.items.add({ r:(new Rect(l4, t1, 25, 25)), i:6 });
    this.items.add({ r:(new Rect(l4 + w1, t1, 25, 25)), i:7 });
    this.items.add({ r:(new Rect(l5, t1, 25, 25)), i:8 });
    this.items.add({ r:(new Rect(l5 + w1, t1, 25, 25)), i:9 });
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      var n = this.items.length();
      for(var i = 0; i < n; ++i) {
        var item = this.items.get(i);
        if(item.r.contains(xy.x, xy.y))
          return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) { 
    var bits = DSArrayGetBasic(this);
    var n = this.items.length();
    for(var i = 0; i < n; ++i) {
      var item = this.items.get(i);
      if(item.r.contains(xy.x, xy.y))
      {
        var bit = (1 << kOutChInfo[i+2].lo);
        bits = (bits & bit) ? (bits & ~bit) : (bits | bit);
        DSArraySetBasic(this, bits);
        this.requestRender();
        return;
      }
    }
  }
  render(c) {
    var bits = DSArrayGetBasic(this);
    var n = this.items.length();
    for(var i = 0; i < n; ++i) {
      var item = this.items.get(i);
      var bit = (1 << kOutChInfo[i+2].lo);
      var isOn = (bits & bit) ? 1 : 0;
      drawButton(c, item.r, i+3, isOn, kToggleBlue);
    }
    c.font = "300 14px Arial";
    c.textAlign = 'center';
    c.fillStyle = c.strokeStyle = kEQLabel;
    c.fillText("Main Out", this.l + 132 + 60, this.t + 56);
    c.fillText("Line Out", this.l + 340 + 80, this.t + 56);
    c.font = "300 16px Arial";
    c.fillText("1 - 2", this.l + 132 + 60, this.t + 88);

    c.beginPath();
    c.moveTo(this.l + 188 + 80, this.t + 52);
    c.lineTo(this.l + 300 + 80, this.t + 52);
    c.moveTo(this.l + 380 + 80, this.t + 52);
    c.lineTo(this.l + 495 + 80, this.t + 52);
    c.lineWidth = 1;
    c.stroke();

    drawLabelGroup(c, "Main Volume Group", this.l+2, this.t+2, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}

//todo listen to kFPGAPatch and uptate meters

class Outputs extends View {
  constructor(root, l, t) {
    super(l, t, 705, 700);
    root.addSub(this);
    this.setBG(kEmptyBG);

    this.addSub(this.mvs = new MainVolGroupEnableMultiSelect(l + 20, t + 12));
    this.pv = new PhonesVol(this, 20 + l, this.mvs.b + 20);
    this.lb = new Loopback(this, 20 + this.pv.r, this.pv.t);
    this.o1 = new LineOut(this, 38 + l, this.pv.b + 50, 0);
    this.o2 = new LineOut(this, 38 + l, this.o1.b + 10, 1);
    this.o3 = new LineOut(this, 20 + this.o1.r, this.pv.b + 50, 2);
    this.o4 = new LineOut(this, 20 + this.o1.r, this.o1.b + 10, 3);
    this.o5 = new LineOut(this, 20 + this.o3.r, this.pv.b + 50, 4);
    this.o6 = new LineOut(this, 20 + this.o3.r, this.o1.b + 10, 5);
    this.o7 = new LineOut(this, 20 + this.o5.r, this.pv.b + 50, 6);
    this.o8 = new LineOut(this, 20 + this.o5.r, this.o1.b + 10, 7);
    this.o9 = new LineOut(this, 20 + this.o7.r, this.pv.b + 50, 8);
    this.o10 = new LineOut(this, 20 + this.o7.r, this.o1.b + 10, 9);
    this.spd = new SPDIFOut(this, l + 20, this.o2.b + 35);
    this.opt = new OpticalOut(this, 20 + this.spd.r, this.o2.b + 35);
  }  
  render(c)
  {
    drawLabelGroup(c, "Output Trim", this.o1.l-18, this.o1.t-30, this.w-20, (this.o10.b - this.o1.t) + 50, 6, kEQLabel, kEQLabel);
  }
}

// phase, trim, 48v amd pad controls for mic pres
// phones and loopback assignment menus
// Separate volume controls for phones and main outs?

class ISetup extends View{
  constructor() {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot();
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    evSys.addListener('wResized', this.resized);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "inSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "inSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;

    this.ins = new Inputs(this, 30, 60);

    this.setBG(kEmptyBG);
  }
  render(c){
    drawDeviceNameHeader(c, this.l + 50, this.t, 630);
  }
  resized() {
    var w = this.ins.r + 30;
    var h = this.ins.b + 30;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
}

class OSetup extends View{
  constructor() {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot();
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    evSys.addListener('wResized', this.resized);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "outSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "outSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;
    this.outs = new Outputs(this, 30, 60);

    this.setBG(kEmptyBG);
  }
  render(c){
    drawDeviceNameHeader(c, this.l + 50, this.t, 690);
  }
  resized() {
    var w = this.outs.r + 30;
    var h = this.outs.b + 30;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
}
