'use strict';


// preload canvas images
const kGearImage = new OffScreenImage("settings.png");
const kLogoImage = new OffScreenImage("logo.svg");

class MixerScrollableCanvas extends CanvasRoot {
    constructor(w, scrollH) {
    super(true);

    this.scrolled = this.scrolled.bind(this);
    this.resized = this.resized.bind(this);
    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    
    this.div.id = "mixInCh";
    this.div.style.width = "100%";
    this.div.style.height = "calc(100% - " + scrollH + "px)";
    this.div.style.overflow = "scroll";
    this.div.style.backgroundColor = kMixBG;
    this.div.style.WebkitOverflowScrolling = "touch";

    this.div.onscroll = this.scrolled;

    this.scroller = document.createElement('div');
    this.scroller.id = "scroller";
    this.scroller.style.width = w.toFixed() + "px";
    this.scroller.style.height = "100%";
    this.scroller.style.backgroundColor = kMixBG;

    // Attach the scroller to the canvas and the canvas to the scroller
    this.rootDiv = this.scroller;    
    this.div.appendChild(this.scroller);
    this.rootDiv.appendChild(this.canvas);

    evSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  mixCfgChanged() {
    var w = mixMgr.getNumCh() * kChStripW;
    this.scroller.style.width = w.toFixed() + "px";
    evSys.dispatchASync('wResized');
  }
  resized() {
    super.resized();
    evSys.dispatchSync('updateMixScroll', this);
  }
  scrolled() {
    this.setVisibleRect(this.div.scrollLeft, this.div.scrollTop, this.r.w, this.r.h);
    evSys.dispatchSync('updateMixScroll', this);
  }
}


class MixerBusStripCanvas extends CanvasRoot {
    constructor(scrollH) {
    super();
    
    this.div.id = "mix out";
    this.div.style.width = "100%";
    this.div.style.height = "100%";
    this.div.style.backgroundColor = kBusStripBG;
    this.div.style.float = "left";
    this.div.style.boxShadow = "-3px 0 6px 0 rgba(0,0,0,0.5)";
    this.resized = this.resized.bind(this);
  }

  resized() {
    super.resized();
   }
}


class StripMeter extends View {
  constructor(l,t,w,h) {
    super(l,t,w,h);
    this.onMeters = this.onMeters.bind(this);
    this.setup = this.setup.bind(this);
    this.setAttenuation = this.setAttenuation.bind(this);
    this.updateZeroPosition = this.updateZeroPosition.bind(this);
    
    this.updateZeroPosition();

    this.lm = {m:0, ph:0};
    this.rm = {m:0, ph:0};
    this.index = 0;
    this.attDB = 0;
    this.hasClip = 0;
    this.c = kMeterBlue;
    meters.addListener(this.onMeters);
  }
  updateZeroPosition() {
    this.zeroH = Math.floor(Calc12dBToCtrl24(0) * this.h);
    this.zeroT = this.t + (this.h - this.zeroH);
  }
  setup(info) {
    this.isStereo = info.isStereo;
    this.index = info.meter ? info.meter : info.meters[kMIPostFx];
    this.c = mixMgr.getActiveBusInfo().color;
    this.hasClip = (info.stripType == "input") && ((info.type == "Mic") || (info.type == "Line"));
  }
  setAttenuation(attDB){
    // attenuation in dB set by fader in post fader mode
    this.attDB = attDB;
  }
  renderMeter(c, m, l, w) {
    var sh = Math.floor(Math.min(Calc12dBToCtrl24(m.m + this.attDB) * this.h, this.zeroH));
    var sy = this.h - sh;
    if(sh && (m.m <= 0)) {
      c.fillStyle = this.c;
      c.fillRect(l, this.t + sy, w, sh)
    }

    var y = this.h - Math.floor(Math.min(Calc12dBToCtrl24(m.ph + this.attDB) * this.h, this.zeroH));
    if(y < (this.h - 2)) {
      c.fillStyle = "red";
      c.fillRect(l, this.t + y, w, 2);
      if(this.hasClip && (Math.floor(m.ph + this.attDB) >= 0)) {
        c.fillRect(l, this.zeroT - 19 - 1, w, 9);
      }
    }
  }
  render(c) {
    c.fillStyle = kBlackBG;
    c.fillRect(this.l, this.zeroT, this.w, this.zeroH);
    if(this.hasClip)
      c.fillRect(this.l, this.zeroT - 19 - 6, this.w, 19);
    if(this.isStereo){
      var w = (this.w/2)-6;
      this.renderMeter(c, this.lm, this.l + 5, w);
      this.renderMeter(c, this.rm, this.l + 7 + w, w);
    }
    else {
      this.renderMeter(c, this.lm, this.l + 5, this.w - 10);
    }

  }
  setMeter(v, i, dest) {
    var m = v[i].m;
    var ph = v[i].ph;
    if((dest.m != m) || (dest.ph != ph)) {
      dest.m = m;
      dest.ph = ph;
      return true;
    }
    return false;
  }
  onMeters(v) {
    if((this.index >= 0) && (this.index < v.length)) {
      if(this.setMeter(v, this.index, this.lm))
        this.requestRender();
      if(this.isStereo) {
        this.setMeter(v, this.index + 1, this.rm);
      }
    }
  }
}


class ChName extends View {
  constructor(l,t,w,h, dsKey) {
    super(l,t,w,h);

    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);

    this.dsKey = dsKey;
    this.info = 0;
    this.setBG(kMixBG);

    if((this.dsKey == 'kiChannelNames') || (this.dsKey == 'koChannelNames'))
      this.wantsMouse = true;

    ds.addListener(dsKey, this.refresh);
  };
  handleMouseDown(xy) {
    if(this.dsKey == 'kiChannelNames') {
      mixMgr.setActiveCh(this.info.chNum);
      chFxEdit.activate();
    }
    else if(this.dsKey == 'koChannelNames') {
      busFxEdit.activate();
    }
  }
  refresh() {
    this.v = getChannelName(this.dsKey, this.info);
    this.requestRender();
  }
  setup(info) {
    this.info = info;
    this.refresh();
  }
  render(c) {
    c.font = '16px Helvetica';
    c.textAlign = 'center';

    // fill it black then stroke if it's not an input with the bus color
    roundRectPath(c, this.l, this.t+2, this.w, this.h, 2);
    c.fillStyle = "black";
    c.fill();

    if(this.info.stripType !== "input") {
      switch (this.info.type) {
        default:
          c.strokeStyle = mixMgr.getActiveBusInfo().color;
          break;
        case "Main":
          c.strokeStyle = kBusColors[0];
          break;
        case "Line":
          c.strokeStyle = kBusColors[2 + this.info.bankChNum];
          break;
        case "Phones":
          c.strokeStyle = kBusColors[10];
          break;
        case "Reverb":
          c.strokeStyle = kBusColors[12];
          break;
      }
      c.lineWidth = 1.1;
      c.stroke();
    }
    c.fillStyle = "white";

    var t = fitText(c, this.v, this.w);
    c.fillText(t, this.l + (this.r - this.l)/2, 7 + this.cy);
    
    if(this.dsKey == 'kiChannelNames') {
      var w = c.measureText(t).width;
      c.fillRect(this.cx - w/2, 9 + this.cy, w, 1);
    }
  }
}


class RudeSoloListener {
  constructor(soloChangedCB) {
    this.soloChanged = this.soloChanged.bind(this);
    this.soloChangedCB = soloChangedCB
    this.busNum = -1;
    this.rs = false;
    evSys.addListener('soloChanged', this.soloChanged);
  }
  setBusNum(bn) { // -1 means disable
    this.busNum = bn;
    this.soloChanged();
  }
  soloChanged() {
    var rs = this.rs;

    if(this.busNum < 0)
      rs = false;
    else
      rs = mixMgr.getBusRudeSolo(mixMgr.getBusInfo(this.busNum).gainOCh);

    if(this.rs != rs) {
      this.rs = rs;
      this.soloChangedCB();
    }
  }
}


class XBox extends View {
  constructor(l,t,w,h) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y))
      modal.deactivate();
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.l + t, this.t + t);
    c.lineTo(this.r -t, this.b -t);
    c.moveTo(this.r -t, this.t + t);
    c.lineTo(this.l + t, this.b -t);
    c.stroke();
  }
}


class RBox extends View {
  constructor(l,t,w,h) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      mixMgr.incActiveCh(false);
    }
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.l + t, this.t + t);
    c.lineTo(this.r - t*2, this.cy);
    c.lineTo(this.l + t, this.b - t);
    c.stroke();
  }
}


class LBox extends View {
  constructor(l,t,w,h) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      mixMgr.incActiveCh(true);
    }
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.r - t, this.t + t);
    c.lineTo(this.l + t*2, this.cy);
    c.lineTo(this.r - t, this.b - t);
    c.stroke();
  }
}


class ChNameHeader extends EditableTextView {
  constructor(l,t,w,h,dsKey) {
    super(l,t,w,h, 'user-input-input', 'Bold 30px Helvetica');

    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);

    this.fx = "";
    this.n = "";
    this.setBG(kFxEditBG);

    this.info = 0;
    this.dsKey = dsKey;
    ds.addListener(this.dsKey, this.refresh);
  }
  handleMouseDown() {
    if ((this.info.type != "Reverb") && (this.info.stripType != "host"))
      super.handleMouseDown();
  }
  valueChanged(v) {
    ds.setV(this.dsKey, this.info.name, v);
    return v; 
  } 
  refresh() {
    this.v = getChannelName(this.dsKey, this.info);
    this.requestRender();
  }
  setup(info) {
    this.info = info;
    this.refresh();
  }
}



class ChFxEditView extends View {
  constructor() {
    super(0,0,kFxWidth,kFxHeight);
  
    const ki = 4;
    const ki2 = 2 * ki;
    const t = ki + 50;
    const h = kFxHeight - (ki2 + 50);

    this.activate = this.activate.bind(this)  
    this.activeChChanged = this.activeChChanged.bind(this)  
    this.addSub(this.lBox = new LBox(kFxChStripW + 25, this.t + 10, 40, 40));
    this.addSub(this.rBox = new RBox(kFxChStripW + 95, this.t + 10, 40, 40));
    this.addSub(this.xBox = new XBox(this.r - 75, this.t + 10, 40, 40));
    this.addSub(this.chNav = new ChNameHeader(this.rBox.r, this.t + 10, this.w - (this.rBox.r + 135), 40, 'kiChannelNames'));

    this.addSub(this.cs = new ChannelStrip(ki, ki, kFxChStripW, kFxHeight - ki2, 0, true));
    this.cs.resize(kFxHeight - ki2, kFxHeight - 80, kFxHeight - 400);
    this.eq = new EQFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, true); 
    this.dyn = new DynamicsView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h); 

    this.fx  = { [kEQFxType]:this.eq, [kDynamicsFxType]:this.dyn };
    this.setBG(kMixBG);
    this.lastFXType = kEQFxType;

    evSys.addListener('activeChChanged', this.activeChChanged);
  }
  activeChChanged() {
    var ch = mixMgr.getActiveCh();
    var info = mixMgr.getChInfo(ch);
    this.cs.setChNum(ch);
    this.chNav.setup(info); 
    this.eq.setup(info); 
    this.dyn.setup(info); 
  }
  activate(fxType) {
    if(fxType == undefined)
      fxType = this.lastFXType;

    this.lastFXType = fxType;
    this.removeSub(this.eq);
    this.removeSub(this.dyn);

    var f = this.fx[fxType];
    if(f != undefined) {

      this.addSub(f);
      this.bg = f.bg;

      modal.activate(this);
      this.requestRender();
    }
  }
  render(c){
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);
  }
}


class ChannelStrip extends View {
  constructor(l,t,w,h, chNum, isFxChStrip) {
    super(l,t,w,h);

    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.resize = this.resize.bind(this);
    this.setChNum = this.setChNum.bind(this);
    this.isFxChStrip = isFxChStrip;

    const itemW = (isFxChStrip ?  kFxChStripW : kChStripW) - 14;
    const itemL = l + 7;
    var fxSlotT = 10;
    if(isFxChStrip) {
      this.addSub(this.stereo = new ToggleBtn(itemL, t + 10, itemW, "Stereo", 'kiMixStereo', 0, kToggleBlue));
      fxSlotT = this.stereo.b + 5;
    }

    this.eq = new EQThumb(itemL, fxSlotT, true, isFxChStrip);
    this.fxSlotH = this.eq.h + 5;
    this.dyn = new DynamicsThumb(itemL, fxSlotT + this.fxSlotH, isFxChStrip);
    if(isFxChStrip) 
    {
      this.addSub(this.pan = new PanFader(itemL, this.dyn.b + 5, itemW, 30));
      this.pan.setH(60);
    }
    else {
      this.addSub(this.name = new ChName(l+7, this.dyn.b + 5, itemW, 35, 'kiChannelNames'));
      this.addSub(this.pan = new PanFader(itemL, this.name.b + 5, itemW, 30));
    }

    this.addSub(this.mute = new ToggleBtn(itemL, this.pan.b + 10, itemW, "MUTE", 'kiMixMute', 0, kToggleRed));
    this.addSub(this.meter = new StripMeter(l+26, this.mute.b + 20, isFxChStrip ? 25 : 19, 357));
    this.addSub(this.fader = new StripFader(l+(isFxChStrip ? kFaderL + 28 : kFaderL), this.mute.b + 20, kFaderL, 367, 'kiMixFader', this.meter));
    this.addSub(this.solo = new ToggleBtn(itemL, this.fader.b + 10, itemW, "SOLO", 'kiMixSolo', 0, kToggleYellow));
    this.setBG(kMixBG);

    this.chNum = chNum;
    this.stripType = "input";

    evSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  setChNum(chNum) {
    this.chNum = chNum;
    this.mixCfgChanged();  
  }
  mixCfgChanged() {

    this.setVis(mixMgr.getNumCh() > this.chNum);
    if(!this.isVis)
      return;

    var info = mixMgr.getChInfo(this.chNum);
    var busInfo = mixMgr.getActiveBusInfo();
    var gainIndex = mixMgr.getChGainIndex(this.chNum);

    if(this.stereo != undefined) {
      this.stereo.setup(info.stereoIndex);
      this.stereo.setVis(info.stereoIndex != undefined);
    }

    this.stripType = info.stripType;
    this.eq.setup(info);
    this.dyn.setup(info);
    if(this.name)
      this.name.setup(info);
    this.pan.setup(gainIndex, info.isStereo);
    this.pan.setVis(busInfo.isStereo);
    this.meter.setup(info);
    this.fader.setup(gainIndex, info.isStereo, this.meter.index);
    this.solo.setup(gainIndex);
    this.mute.setup(gainIndex);
    this.resize(this.h);
  }
  resize(chH) {
    this.setH(chH);

    if(this.stereo != undefined)
      this.removeSub(this.stereo);
    this.removeSub(this.eq);
    this.removeSub(this.dyn);

    var t = this.t + 10;
    var minH = 340;
    var seq = 0;
    var sd = 0;
    
    if(chH > (minH + this.fxSlotH)) 
      seq = 1;
    if(chH > (minH +  2 * this.fxSlotH)) 
      sd = 1;

    if((this.stereo != undefined) && (chH > (minH + this.stereo.h + 5 + 2 * this.fxSlotH))){
      this.addSub(this.stereo);
      t += this.stereo.h + 5;
    }  

    if(getSampleRateMult(ds.getKVArrayItem('kSampleRate', 0)) == 4) {
      // leave an empty slot because the bus strips will have eq at 4x
      t += this.fxSlotH;
    }
    else {
      if(seq) {
        if(this.eq.isValid) {
          this.eq.move(this.eq.l, t);
          this.addSub(this.eq);
        }
        t += this.fxSlotH;
      }
      if(sd) {
        if(this.dyn.isValid) {
          this.dyn.move(this.dyn.l, t);
          this.addSub(this.dyn);
        }
        t += this.fxSlotH;
      }
    }

    if(this.name) {
      this.name.setT(t);
      t = this.name.b + 5;
    }
    if(this.pan.isVis) {
      this.pan.setT(t);
      t = this.pan.b + 10;
    }
    this.mute.setT(t);
    this.solo.setT(chH - (this.solo.h + 10));
    var faderH = (this.solo.t - this.mute.b) - 20;
    var meterT = this.mute.b + 12 + kFaderCapRadius;
    var meterH = faderH - (2 * (kFaderCapRadius + 2));
    this.meter.setT(meterT);
    this.meter.setH(meterH + 3);
    this.meter.updateZeroPosition();
    this.fader.setMeterTH(meterT, meterH);
    this.fader.setT(this.mute.b + 10);
    this.fader.setH(faderH);

    this.requestRender();
  }
  render(c) {

    if(this.name && (this.stripType == "host")) {
      c.fillStyle = kMeterTickColor;
      c.textAlign = 'center';
      c.font = "16px Helvetica";
      c.fillText("Computer", this.cx, this.name.t - 35);
      c.fillText("USB", this.cx, this.name.t - 13);
    }

    draw12dBTicks(c, kMeterTickColor, this.fader.meterT, this.l, this.fader.meterH, this.fader.l);

    c.lineWidth = 1.5;
    if (this.chNum && !this.isFxChStrip) {
      c.strokeStyle = "rgb(67,67,67)";
      c.beginPath();
      c.moveTo(this.l, this.t);
      c.lineTo(this.l, this.b-1);
      c.stroke();
    }
    c.strokeStyle = "rgb(60,60,60)";
    c.beginPath();
    c.moveTo(this.r, this.t);
    c.lineTo(this.r, this.b-1);
    c.stroke();
    if (!this.isFxChStrip) {
      c.strokeStyle = "black";
      c.beginPath();
      c.moveTo(this.l, this.b);
      c.lineTo(this.r, this.b);
      c.stroke(); 
    }
  }
}


class BusFxEditView extends View {
  constructor() {
    super(0,0,kFxWidth,kFxHeight);
  
    const ki = 4;
    const ki2 = 2 * ki;
    const t = ki + 50;
    const h = kFxHeight - (ki2 + 50);

    this.activate = this.activate.bind(this)  
    this.mixCfgChanged = this.mixCfgChanged.bind(this)  

    this.addSub(this.xBox = new XBox(this.r - 75, this.t + 10, 40, 40));
    this.addSub(this.chNav = new ChNameHeader(kFxChStripW + 95, this.t + 10, this.xBox.l - (kFxChStripW + 95), 40, 'koChannelNames'));
    this.addSub(this.bs = new BusStrip(ki, ki, kFxChStripW, kFxHeight - ki2, true));
    this.bs.resize(kFxHeight - ki2, 235, kFxHeight - 255);
    this.eq = new EQFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, false); 
    this.rvb = new ReverbFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h); 
    this.fx = { [kBusEQFxType]:this.eq, [kReverbFxType]:this.rvb };
    this.setBG(this.eq.bg);

    evSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  mixCfgChanged() {
    var info = mixMgr.getActiveBusInfo();
    this.chNav.setup(info); 
    this.eq.setup(info); 
    this.rvb.setup(info);
  }
  activate(fxType) {
    if(fxType == undefined)
      fxType = kBusEQFxType;
    
    this.removeSub(this.eq);
    this.removeSub(this.rvb);

    var f = this.fx[fxType];
    if (f != undefined) {
      this.addSub(f);
      this.bg = f.bg;

      modal.activate(this);
      this.requestRender();
    }
  }
  render(c){
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);
  }
}


class CollapseBtn extends View {
  constructor(l, t) {
    super(l, t, 30, 30)
    this.wantsMouse = true;
    this.setBG(kBusStripBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      ds.setV('kBusStripExp', 0, ds.getKVArrayItem('kBusStripExp', 0) == "on" ? "off" : "on");
      this.requestRender();
    }
  }
  render(c) {
    var isExp = ds.getKVArrayItem('kBusStripExp', 0) != "off";
    if(isExp) {
      drawRArrow(c, 1, this.r-21, this.b-9, kToggleDefault);
      drawLArrow(c, 1, this.r-9, this.b-21, kToggleDefault);
    }
    else {
      drawLArrow(c, 1, this.r-21, this.b-9, kToggleDefault);
      drawRArrow(c, 1, this.r-9, this.b-21, kToggleDefault);
    }
  }
}


class ClearSoloBtn extends View {
  constructor(l, t) {
    super(l, t, 30, 30)
    this.soloChanged = this.soloChanged.bind(this);
    this.wantsMouse = true;
    this.rs = new RudeSoloListener(this.soloChanged);
    this.setBG(kBusStripBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      mixMgr.clearBusSolos(mixMgr.getActiveBusInfo().gainOCh);
    }
  }
  soloChanged() {
    this.requestRender();
  }
  render(c) {
    if(this.rs.rs)
      drawButton(c, this, "CLEAR SOLO", false, mixMgr.getActiveBusInfo().color);
  }
}


class BusStrip extends View {
  constructor(l,t,w,h, isFxBusStrip) {
    super(l,t,w,h);

    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.resize = this.resize.bind(this);

    this.isFxBusStrip = isFxBusStrip;

    const itemW = (isFxBusStrip ?  kFxChStripW : kBusStripW) - 14;
    const itemL = l + 7;

    var fxSlotT = t + 10;
    if(isFxBusStrip) {
      this.addSub(this.stereo = new ToggleBtn(itemL, t + 10, itemW, "Stereo", 'koMixStereo', 0, kToggleBlue, kBusStripBG));
      fxSlotT = this.stereo.b + 5;
    }
    var tt = t + 0;

    // note that the eq thumb is the same size as the chstrip because I wan't it's hieght to be the same
    this.addSub(this.eq = new EQThumb(itemL + (isFxBusStrip? 0 : 13), fxSlotT, false, isFxBusStrip));
    this.fxSlotH = this.eq.h + 5;
    this.addSub(this.rvb = new ReverbThumb(itemL + (isFxBusStrip? 0 : 13), fxSlotT + this.fxSlotH, isFxBusStrip));

    this.addSub(this.mute = new ToggleBtn(itemL, tt, itemW, "MUTE", 'koBusMute', 0, kToggleRed));
    this.mute.setBG(kBusStripBG);
    
    this.addSub(this.meter = new StripMeter(l+26, tt, 30, 357));
    this.addSub(this.fader = new StripFader(l+kFaderL+17, tt, kFaderL, 357, 'koBusFader'));

    this.setBG(kBusStripBG);
    this.fader.setBG(kBusStripBG);

    if(!isFxBusStrip) {
      this.addSub(this.name = new ChName(l+12, this.mute.t - 80, kBusStripW - 24, 35, 'koChannelNames'));
      this.name.setBG(kBusStripBG);

      this.addSub(this.collapseBtn = new CollapseBtn(l+26, this.b-35));
      this.addSub(this.csb = new ClearSoloBtn(30,30));
    }

    this.isStereo = true;

    evSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  updateNamePosition() {
    if (this.name)
      this.name.setT(this.mute.t - (40 + (40 * this.isStereo)));
  }
  mixCfgChanged() {
    var info = mixMgr.getActiveBusInfo();
    if(this.stereo != undefined) {
      this.stereo.setup(info.stereoIndex);
      this.stereo.setVis(info.stereoIndex != undefined);
    }

    var fxSlotT = this.t + 10;
    if(this.isFxBusStrip && (info.type != "Reverb")) {
      fxSlotT = this.stereo.b + 5;
    }

    if(this.eq.t != fxSlotT) {
      // EQ and Reverb views need to be constructed at their postion
      // they do not gracefully move around by changing their view rect
      var l = this.eq.l;
      this.removeSub(this.eq);
      this.removeSub(this.rvb);
      this.addSub(this.eq = new EQThumb(l, fxSlotT, false, this.isFxBusStrip));
      this.addSub(this.rvb = new ReverbThumb(l, fxSlotT + this.fxSlotH, this.isFxBusStrip));
    }

    if(this.name)
      this.name.setup(info);
    this.meter.setup(info);
    this.fader.setup(info.gainOCh, info.isStereo);

    this.mute.setup(info.gainOCh);
    this.eq.setup(info);
    this.rvb.setup(info);

    if(this.csb)
      this.csb.rs.setBusNum(mixMgr.getActiveBus());

    if(this.isStereo != info.isStereo) {
      this.isStereo = info.isStereo;
      this.updateNamePosition();
    }
    this.requestRender();
  }
  resize(chH, faderT, faderH) {
    var expBus = (ds.getKVArrayItem('kBusStripExp', 0) != "off") || this.isFxBusStrip;
    this.setWH(expBus ? (this.isFxBusStrip ?  kFxChStripW : kBusStripW) : kBusStripWSmall, chH);

    this.removeSub(this.eq);
    this.removeSub(this.rvb);
    if((chH > 434) && expBus)
    {
      this.addSub(this.eq);
      if (chH > (434 + this.fxSlotH))
        this.addSub(this.rvb);
    }
    this.mute.setVis(expBus);
    this.fader.setVis(expBus);

    this.mute.setT(faderT - 40);
    if(this.isFxBusStrip) {
      this.meter.setL(30);
      this.fader.setL(this.l+kFaderL+32);
    }
    else {
      this.meter.setL(expBus ? 26 : 11);
      this.fader.setL(this.l+kFaderL+17);
    }
    var meterT = faderT + 2 + kFaderCapRadius;
    var meterH = faderH - (2 * (kFaderCapRadius + 2));
    this.meter.setT(meterT);
    this.meter.setH(meterH + 3);
    this.meter.updateZeroPosition();
    this.fader.setT(faderT);
    this.fader.setH(faderH);
    this.fader.setMeterTH(meterT, meterH);
    if(this.csb) {
      this.csb.setVis(expBus);
      this.csb.setLTWH(this.l+14,  this.fader.b + 10, this.w - 28, 30);
    }

    if(this.collapseBtn) {
      this.collapseBtn.setT(this.b - 35);
      this.collapseBtn.setL(this.cx - this.collapseBtn.w/2);
    }

    if(this.name) {
      this.updateNamePosition();
      this.name.setVis(expBus);
    }
  }
  render(c){
    if(this.fader.isVis){
      draw12dBTicks(c, kMeterTickColor, this.fader.meterT, this.l, this.fader.meterH, this.fader.l);
    }
  }
}


class Mixer {
  constructor() {
    var xPos = kTabSelW;
    var mixOutW = kBusStripW;

    // Container for all mixer
    this.div = document.createElement('div');
    this.div.id = "mixer";
    this.div.style.position = "relative";
    this.div.style.width = "calc(100% - " + xPos + "px)";
    this.div.style.height = "100%";
    this.div.style.backgroundColor = kMixBG;
    this.div.style.float = "left";

    // Container for mixer ins and scroller
    this.mixIn = document.createElement('div');
    this.mixIn.id = "mixIn";
    this.mixIn.style.width = "calc(100% - " + mixOutW + "px)";
    this.mixIn.style.height = "100%";
    this.mixIn.style.backgroundColor = kMixBG;
    this.mixIn.style.float = "left";

    // Container for mixer outs
    this.mixOut = document.createElement('div');
    this.mixOut.id = "mixOut";
    this.mixOut.style.width = mixOutW + "px";
    this.mixOut.style.height = "100%";
    this.mixOut.style.backgroundColor = kBusStripBG;
    this.mixOut.style.float = "left";
    this.mixOut.style.zIndex = 100;
    this.mixOut.style.position = "absolute";
    this.mixOut.style.right = 0;
    this.mixOut.style.top = 0;

    this.setVis = this.setVis.bind(this);

    // The mix canvas that scrolls
    this.mixInCanvas = new MixerScrollableCanvas(kChStripW * kNumChStrips, kScrollBarSize);

    this.chStrips = [];
    for (var i = 0; i < kNumChStrips; i++) {
      var cs = new ChannelStrip(i * kChStripW, 0, kChStripW, 868, i, false);
      this.mixInCanvas.addSub(cs);
      this.chStrips.push(cs);
    }

    // The bus fader
    this.mixOutCanvas = new MixerBusStripCanvas(kScrollBarSize);
    this.mixOutCanvas.addSub(this.busStrip = new BusStrip(0, 0, mixOutW, 868, false));
    this.mixOutCanvas.subViews.add(this.busStrip);
    // The scroll bar
    this.scrollBarView = new ScrollBar(this.mixInCanvas, 'updateMixScroll');

    this.mixIn.appendChild(this.mixInCanvas.div);
    this.mixIn.appendChild(this.scrollBarView.c.div);
    this.mixOut.appendChild(this.mixOutCanvas.div);
    this.resized = this.resized.bind(this);
    evSys.addListener('wResized', this.resized);

    this.div.appendChild(this.mixIn);
    this.div.appendChild(this.mixOut);
    this.tabDiv = this.div;

    this.onDS = this.onDS.bind(this);
    ds.addListener('kBusStripExp', this.onDS);
    this.onDS('kBusStripExp', ds.getKVArray('kBusStripExp'));
  }
  setVis(isVis) {
    this.mixInCanvas.setVis(isVis);
    this.mixOutCanvas.setVis(isVis);
    this.scrollBarView.setVis(isVis);
  }
  onDS(k, v) {
    if(k == 'kBusStripExp') {
      var expBus = v[0]; 

      var mixOutW = (expBus == "on") ? kBusStripW : kBusStripWSmall; 

      this.mixIn.style.width = "calc(100% - " + mixOutW + "px)";
      this.mixOut.style.width = mixOutW + "px";
      evSys.dispatchASync('wResized');
    }
  }  
  resized() {
    // tell the ui views the window resized
    this.mixInCanvas.resized();
    this.scrollBarView.resized();
    this.mixOutCanvas.resized();

    var h = this.mixInCanvas.r.h;
    var c = this.chStrips[0];


    var n = this.chStrips.length;
    for (var i = 0; i < n; i++) {
      this.chStrips[i].resize(h);
    }
    this.busStrip.resize(h+kScrollBarSize, c.fader.t, c.fader.h);
  }
}

const kTabItemH = 50;
class TabItem extends View {
  constructor(label, v, t, p) {
    super(0, t, kTabSelW-2, kTabItemH);
    p.addSub(this);

    this.onEv = this.onEv.bind(this);
    evSys.addListener(kTabEv, this.onEv);

    this.setBG(kTabBG);

    this.label = label;
    this.v = v;
    this.on = false;
  }

  onEv(k,v) {
    var isItMe = (v == this.v);
    if(isItMe != this.on) {
      this.on = isItMe;
      this.setBG(this.on ? kTabSel : kTabBG);
      this.requestRender();
    }
    else if (isItMe)
      this.requestRender();
  } 
  render(c) {
    if(this.on) {
      c.fillStyle = kTabTxtSel;
    } else {
      c.fillStyle = kTabTxt;
    }
    c.textAlign = "left";
    c.font = "16px Helvetica";
    c.fillText(this.label, this.l + 50, 7 + this.cy);
  }
}

class HomeTabItem extends TabItem {
  constructor(label, v, t, p) {
    super(label, v, t, p);
  }
  render(c) {
    super.render(c);

    var l = this.l+ 7;
    var t = this.t + 14;
    c.save();
    c.strokeStyle = 'RGB(192,194,195)';
    c.lineWidth = 2;
    c.lineJoin = "miter";
    c.miterLimit="0.6";
    c.beginPath();
    c.moveTo(l+30, t+12);
    c.lineTo(l+24, t+12);
    c.lineTo(l+24, t+23);
    c.lineTo(l+7,  t+23);
    c.lineTo(l+7,  t+12);
    c.lineTo(l+0,  t+12);
    c.lineTo(l+15, t+0);
    c.lineTo(l+20, t+4);
    c.lineTo(l+20, t+1);
    c.lineTo(l+24, t+1);
    c.lineTo(l+24, t+7);
    c.closePath();
    c.stroke();
    c.restore();
  }
}
class DevTabItem extends TabItem {
  constructor(label, v, t, p) {
    super(label, v, t, p);
  }
  render(c) {
    super.render(c);
    kGearImage.render(c, this.l + 9, this.t + 14);
  }
}
class IOTabItem extends TabItem {
  constructor(l1, l2, v, t, p) {
    super(l2, v, t, p);
    this.l1 = l1;
  }
  render(c) {
    super.render(c);

    var x = this.l + 10;
    var y = this.t + 42;
    
    c.fillStyle = 'RGB(192,194,195)';
    c.textAlign = "center";
    c.font = "16px Helvetica";
    c.fillText(this.l1, this.l + 22, 7 + this.cy);
  }
}


class MixTabItem extends TabItem {
  constructor(busNum, p) {
    super("", busNum,0, p);
    this.soloChanged = this.soloChanged.bind(this);
    this.c = "black";
    this.setVis(false);
    this.rs = new RudeSoloListener(this.soloChanged);
  }
  update(t) {
    var n = mixMgr.getNumBuses();
    this.setVis(this.v < n);
    if(this.isVis) {
      var bi = mixMgr.getBusInfo(this.v);
      this.label = getChannelName('koChannelNames', bi).toUpperCase();
      if(bi.type != "Reverb")
         this.label += " MIX";
      this.c = bi.color;
      this.setT(t);
      t += kTabItemH;
    }

    this.rs.setBusNum(this.isVis ? this.v : -1);
    return t;
  }
  soloChanged() {
    this.requestRender();
  }
  render(c) {
    super.render(c);
    c.beginPath();
    c.strokeStyle = this.c;
    c.lineWidth = this.on ? 8 : 1.5;
    c.moveTo(this.l+10,this.cy);
    c.lineTo(this.l+35,this.cy);
    c.stroke();
    if(this.rs.rs) {
      c.font = "12px Arial";
      c.fillStyle = this.c;
      c.textAlign = "center";
      c.fillText("solo", this.l + 22, this.t+18);
    }
  }
}


class TabView extends View {
  constructor() {
    super(0, 0, 0, 0);
    
    this.resized = this.resized.bind(this);
    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.getTab = this.getTab.bind(this);
    this.wantsMouse = true;
    this.mouseDown = false;

    evSys.addListener('wResized', this.resized);
    evSys.addListener('mixCfgChanged', this.mixCfgChanged);
    this.c = new CanvasRoot();
    this.c.addSub(this);

    this.scrollBar = new VScrollBar(this.c.div);
    this.div = this.scrollBar.containerDiv;

    this.div.id = "tabViewContainer";
    this.div.style.position = "relative";
    this.div.style.height = "100vh";
    this.div.style.float = "left";
    this.div.style.width = kTabSelW + "px";
    this.div.style.boxShadow = "rgba(0,0,0,0.5) 3px 0px 6px 0px";
    this.div.style.zIndex = "100";

    this.c.div.id = "tabView";
    this.c.div.style.overflowX = "hidden";
    this.c.div.style.overflowY = "auto";
    this.c.div.style.position = "relative";
    this.c.canvas.style.height = "1200px";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kTabBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.setBG(kTabBG);

    this.tabs = [];
    this.homeTab = new HomeTabItem("HOME", kHome, kTabItemH, this);
    this.devTab = new DevTabItem("DEVICE", kDev, this.homeTab.b, this);
    this.iTab = new IOTabItem("I", "INPUT", kISetup, this.devTab.b, this);
    this.oTab = new IOTabItem("O", "OUTPUT", kOSetup, this.iTab.b, this);
    this.addSub(this.homeTab); this.tabs.push(this.homeTab);
    this.addSub(this.devTab); this.tabs.push(this.devTab);
    this.addSub(this.iTab); this.tabs.push(this.iTab);
    this.addSub(this.oTab); this.tabs.push(this.oTab);

    this.mixTabs = [];
    for(var i = 0; i < 16; ++i) {
      var tab = new MixTabItem(i, this);
      this.mixTabs.push(tab); this.tabs.push(tab);
    }
    this.downTab = undefined;

    ds.addListener('koChannelNames', this.mixCfgChanged);
  }
  mixCfgChanged() {
    var t = this.oTab.b;
    for(var i = 0; i < 16; ++i)
      t = this.mixTabs[i].update(t);
    if((!this.homeTab.on) && (!this.devTab.on) && (!this.iTab.on) && (!this.oTab.on))
      evSys.dispatchASync(kTabEv, mixMgr.getActiveBus());
  }
  getTab(xy) {
    var numTabs = this.tabs.length;
    for(var i = 0; i < numTabs; ++i) {
      var tab = this.tabs[i];
      if(tab.isVis && (tab.l <= xy.x) && ((tab.r -20) > xy.x) && (tab.t <= xy.y) && (tab.b > xy.y))
        return tab;
    }
    return undefined;
  }
  handleMouseDown(xy, e) {
    this.mouseDown = true;
    if(xy.y < kTabItemH) {
      ds.setV('kTabExp', 0, ds.getKVArrayItem('kTabExp', 0) == "on" ? "off" : "on");
      this.requestRender();
    }
    else {
      var tab = this.getTab(xy);
      if(tab != undefined) {
        this.lastY = e.touches ? e.touches[0].clientY : e.y;
        this.downTab = tab;
      }
    }
  }
  handleMouseMoved(xy, delta, e) {
    var tab = this.getTab(xy);
    if((this.downTab != undefined) && 
      ((delta.y < -5) || (delta.y > 5) || (tab != this.downTab)))
    {
      this.downTab = undefined;
    }
    else {
      var y = e.touches ? e.touches[0].clientY : e.y;
      var st = this.scrollBar.contentDiv.scrollTop - (y - this.lastY);
      this.lastY = y;
      this.scrollBar.setScrollTop(st);
    }
  }
  handleMouseUp() {
    this.mouseDown = false;
    if(this.downTab != undefined) {
      evSys.dispatchASync(kTabEv, this.downTab.v);
      this.downTab = undefined;
    }
  }
  render(c)
  {
    kLogoImage.render(c, 60, 15);
    if(ds.getKVArrayItem('kTabExp', 0) == "off") {
      drawRArrow(c, 1, 28, 19, "white");
      drawLArrow(c, 1, 16, 31, "white");
    }
    else {
      drawLArrow(c, 1, 28, 19, "white");
      drawRArrow(c, 1, 16, 31, "white");
    }
  }
  resized() {
    if(this.mouseDown)
      return;

    var w = kTabSelW;
    var h = this.mixTabs[mixMgr.getNumBuses() - 1].b;

    this.c.setVisibleRect(0, 0, w, h);
    this.setR(this.c.r);

    this.scrollBar.updateScroll(h);
    this.scrollBar.setEnable(ds.getKVArrayItem('kTabExp', 0) == "on");
  }
}


class FxModalCanvas extends ModalCanvas {
  constructor() {
    super(true);
  }
  filterKey(keycode) {
    if(this.mv != null) {
      if(keycode == 'escape') {
        this.deactivate();
      }
      else if(keycode == 'larrow') {
        mixMgr.incActiveCh(true);
      }
      else if(keycode == 'rarrow') {
        mixMgr.incActiveCh(false);
      }
      else { 
        return false;
      }
      return true;
    }
    return false;
  }
}

function MMAddListener(mqString, cb) {
  // safari didn't like addeventlistener so I changed to this but I'm not sure if it's correct
  var mm = matchMedia(mqString);
  if(mm.addEventListener != undefined)
    mm.addEventListener("change", cb);
  else
    mm.addListener(cb);
}

class App {
  constructor() {

    this.resized = this.resized.bind(this);
    this.onTab = this.onTab.bind(this);
    this.onDS = this.onDS.bind(this);

    document.body.onresize = this.resized;

    this.tab = new TabView();
    evSys.addListener(kTabEv, this.onTab);
    
    // create the mixer ui in a scrollable canvas
    this.home = new Home();
    this.setup = new DevSetup();
    this.isetup = new ISetup();
    this.osetup = new OSetup();
    this.mix = new Mixer();
    
    document.body.appendChild(this.tab.div);
    this.tabChild = null;
        
    let pixelRatioBox = document.querySelector(".pixel-ratio");
    let mqString = `(resolution: ${window.devicePixelRatio}dppx)`;
    MMAddListener(mqString, this.resized);

    ds.addListener('kTabExp', this.onDS);
    this.onDS('kTabExp', ds.getKVArray('kTabExp'));

    var t = storage.getItem("tab");
    evSys.dispatchSync(kTabEv, t ? t : kHome);
  }
  onDS(k, v) {
    if(k == 'kTabExp') {
      var expTab = v[0];
      var tabW = (expTab == "on") ? kTabSelW : 42; 

      this.mix.div.style.width = "calc(100% - " + tabW + "px)";
      this.home.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.setup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.isetup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.osetup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.tab.div.style.width = tabW + "px";
      evSys.dispatchASync('wResized');
    }
  }
  onTab(k,v) {
    var tc = this.home;
    switch(v) {
      case kHome: {
        tc = this.home;
      } break;
      case kDev: {
        tc = this.setup;
      } break;
      case kISetup: {
        tc = this.isetup;
      } break;
      case kOSetup: {
        tc = this.osetup;
      } break;
      default:{
        mixMgr.setActiveBus(v);
        tc = this.mix;
      } break;
    }
    storage.setItem("tab", v);

    if((tc != null) && (tc != this.tabChild)) {
      if(this.tabChild != null) {
        document.body.removeChild(this.tabChild.tabDiv);
        this.tabChild.setVis(false);
      }
      this.tabChild = tc;
      this.tabChild.setVis(true);
      document.body.appendChild(tc.tabDiv);
      tc.resized();
    }
  }
  resized() {
    kCanvasScale = window.devicePixelRatio || 2;
    evSys.dispatchSync('wResized');
  }

}


window.onload = function()
{
  isIOS = (/iPad|iPhone|iPod/.test(navigator.platform) || 
    (navigator.platform === 'MacIntel' && navigator.maxTouchPoints > 1)) &&
    !window.MSStream;

  // this disables zooming/scaling by the broswer, it stops mobile browsers from scaling the page
  // it's not needed because I set it in the html meta. I may want this later ...
  // document.querySelector("meta[name=viewport]").setAttribute('content', 'width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0');

  isMobile = isIOS; // I consider iOS mobile
  // device detection, i consider it mobile if the user agent says so or if the screen is small
  if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|ipad|iris|kindle|Android|Silk|lge |maemo|midp|mmp|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows (ce|phone)|xda|xiino/i.test(navigator.userAgent) 
    || /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(navigator.userAgent.substr(0,4))) { 
    isMobile = true;
  }

  // init globals here
  storage = new Storage;
  ds = new DataStore;
  gen5Lib = new Gen5Lib;
  evSys = new Events;
  device = new Device;
  afd = new AnimationFrameDispatcher;
  meters = new MeterStore;
  mixMgr = new MixMgr;
  modal = new FxModalCanvas;
  chFxEdit = new ChFxEditView;
  busFxEdit = new BusFxEditView;
  keyFilter = new KeyDownFilter;
  updater = new Updater;
  welcome = new WelcomeView;
  mixMgr.update();

  new App();

  // enable the updater last
  updater.start();

  if(gen5Lib.hasLib()) {
    gen5Lib.synchDSToCurrentState();
    updater.updateDriverVersion(gen5Lib.getDriverVersion());
  }
}
