'use strict';


function drawRArrow(c, lw, cx, cy, color) {
    // draw arrow pointing up and right centered on cx, cy
    const k1 = 0.8;
    const k2 = 4;
    const k3 = 8;
    const k4 = 3;

    var sx = cx - k3/2; 
    var sy = cy + k3/2; 
    c.lineWidth = lw;
    c.strokeStyle = c.fillStyle =  color; 
    c.beginPath();
    c.moveTo(sx, sy);
    c.lineTo(sx+k1, sy+k1);
    c.lineTo(sx+k2+k1, (sy-k2)+k1);
    c.lineTo(sx+k2+k4, (sy-k2)+k4);
    c.lineTo(sx+k3, sy-k3);
    c.lineTo((sx+k2)-k4, (sy-k2)-k4);
    c.lineTo((sx+k2)-k1, (sy-k2)-k1);
    c.lineTo(sx-k1, sy-k1);
    c.closePath();
    c.lineJoin = "miter";
    c.fill();
}

function drawLArrow(c, lw, cx, cy, color) {
    // draw arrow pointing down and left centered on cx, cy
    const k1 = 0.8;
    const k2 = 4;
    const k3 = 8;
    const k4 = 3;

    var sx = cx + k3/2; 
    var sy = cy - k3/2; 

    c.strokeStyle = c.fillStyle =  color; 
    c.lineWidth = lw;
    c.beginPath();
    c.moveTo(sx, sy);
    c.lineTo(sx-k1, sy-k1);
    c.lineTo((sx-k2)-k1, (sy+k2)-k1);
    c.lineTo((sx-k2)-k4, (sy+k2)-k4);
    c.lineTo(sx-k3, sy+k3);
    c.lineTo((sx-k2)+k4, (sy+k2)+k4);
    c.lineTo((sx-k2)+k1, (sy+k2)+k1);
    c.lineTo(sx+k1, sy+k1);
    c.closePath();
    c.lineJoin = "miter"; 
    c.fill();
}


function drawPwrIcon(c, lw, cx, cy, r, color) {

    c.strokeStyle =  color; 
    c.beginPath();
    c.arc(cx, cy, r, 1.75 * Math.PI, 1.25 * Math.PI);
    c.lineWidth = lw;
    c.lineCap = "round";
    c.moveTo(cx, cy);
    c.lineTo(cx, cy - r);
    c.stroke();
    c.lineCap = "butt"; 
}

function drawInfoIcon(c, cx, cy, color) {

    c.fillStyle = c.strokeStyle =  color; 
    c.beginPath();
    c.arc(cx, cy, 10, 0, 2 * Math.PI);
    c.lineWidth = 2.2;
    c.lineCap = "round";
    c.stroke();
    c.lineCap = "butt"; 
    c.font = "bolder 15px Arial";
    c.textAlign = 'center';
    c.fillText("?", cx, cy+5);
}

function drawDiagnosticsIcon(c, cx, cy, color) {

    c.strokeStyle =  color; 
    c.lineWidth = 2.2;
    c.lineCap = "round";
    c.beginPath();
    c.moveTo(cx-8, cy);
    c.lineTo(cx-4, cy);
    c.lineTo(cx-2, cy-6);
    c.lineTo(cx+2, cy+6);
    c.lineTo(cx+4, cy);
    c.lineTo(cx+8, cy);
    c.stroke();
    c.lineCap = "butt"; 
}

function roundRectPath(c, x, y, w, h, rad) {
  c.beginPath();
  c.moveTo(x + rad, y);
  c.lineTo(x + w - rad, y);
  c.quadraticCurveTo(x + w, y, x + w, y + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + rad);
  c.quadraticCurveTo(x, y, x + rad, y);
  c.closePath();
}

function drawLabelGroup(c, label, x, y, w, h, rad, rClr, tClr) {

  var to = 12;
  c.font = "300 16px Arial";
  c.textAlign = 'left';

  var fitLabel = fitText(c, label, w - (2*rad + 6));
  var tw = c.measureText(fitLabel).width;

  c.beginPath();
  c.moveTo(x + rad + tw + 5, y + to);
  c.lineTo(x + w - rad, y + to);
  c.quadraticCurveTo(x + w, y + to, x + w, y + to + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + to + rad);
  c.quadraticCurveTo(x, y + to, x + rad, y + to);

  c.lineWidth = 2;
  c.strokeStyle = rClr; 
  c.stroke();

  c.fillStyle = tClr;
  c.fillText(fitLabel, x + 8, y + 15);
}

function drawDoubleLabelGroup(c, labels, x, y, w, h, rad, rClr, tClr) {

  var to = 10;
  c.font = "300 14px Arial";
  c.textAlign = 'left';

  var fitLabel = fitText(c, labels[0], w - (2*rad));
  var tw = c.measureText(fitLabel).width;

  c.beginPath();
  c.moveTo(x + rad + tw + 5, y + to);
  c.lineTo(x + w - rad, y + to);
  c.quadraticCurveTo(x + w, y + to, x + w, y + to + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + to + rad);
  c.quadraticCurveTo(x, y + to, x + rad, y + to);
  
  c.lineWidth = 1;
  c.strokeStyle = rClr;
  c.stroke();

  var w2 = 16;
  var h2 = 16;
  var x2 = x + w - w2;
  var y2 = y + h - h2;
  c.beginPath();
  c.moveTo(x2 + rad, y2);
  c.lineTo(x2 + w2, y2);
  c.lineTo(x2 + w2, y2 + h2 - rad);
  c.quadraticCurveTo(x2 + w2, y2 + h2, x2 + w2 - rad, y2 + h2);
  c.lineTo(x2, y2 + h2);
  c.lineTo(x2, y2 + rad);
  c.quadraticCurveTo(x2, y2, x2 + rad, y2);
  c.stroke();
  c.fillStyle = rClr;
  c.fill();

  c.fillStyle = tClr;
  c.fillText(fitLabel, x + 8, y + 15);

  c.font = "11px Arial";
  c.textAlign = 'center';
  c.fillStyle = kEmptyBG;
  c.fillText(labels[1], x2 + w2/2, y + h - 4, w2);
}


function drawButton(c, r, label, on, onColor, enabled = true) {
  roundRectPath(c, r.l+2, r.t+2, r.w-4, r.h-4, 6);
  if(!enabled) {
    c.strokeStyle = c.fillStyle = "#303030";
    c.lineWidth = 1;
    c.stroke();
  }
  else if(on) {
    c.fillStyle = onColor;
    c.fill();
    c.fillStyle = kBlackBG;
    // c.strokeStyle = onColor;
    // c.fillStyle = onColor;//kToggleDefault;
    // c.lineWidth = 2;
    //c.stroke();
  } else {
    c.strokeStyle = c.fillStyle = kToggleDefault;
    c.lineWidth = 1;
    c.stroke();
  }

  c.font = "14px Arial";
  c.textAlign = 'center';
  c.fillText(label, r.cx, 5 + r.cy);
}

function fillCircle(c, x, y, radius, fillColor) {
  c.beginPath();
  c.arc(x,y,radius,0,2*Math.PI);
  c.fillStyle = fillColor; 
  c.fill();
};

function fillTextWrapped(c, t, x, y, maxWidth, lineHeight) {
  var words = t.split(' ');
  var line = '';

  for(var n = 0; n < words.length; n++) {
    var testLine = line + words[n] + ' ';
    var metrics = c.measureText(testLine);
    var testWidth = metrics.width;
    if (testWidth > maxWidth && n > 0) {
      c.fillText(line, x, y);
      line = words[n] + ' ';
      y += lineHeight;
    }
    else {
      line = testLine;
    }
  }
  c.fillText(line, x, y);
  return y;
}

const kULm5LogoBG = new OffScreenImage("Utralitemk5Logo.svg");
function drawDeviceNameHeader(c, l, t, w) {
  c.fillStyle = kTabTxt;
  c.fillRect(l, t + 40, w - kULm5LogoBG.width, 2);
  kULm5LogoBG.render(c, l + (w - kULm5LogoBG.width), t);
}

function fitText(c, str, maxWidth) {
    var width = c.measureText(str).width;
    var ellipsis = '...';
    var ellipsisWidth = c.measureText(ellipsis).width;
    if (width<=maxWidth || width<=ellipsisWidth) {
        return str;
    } else {
        var len = str.length;
        while (width>=maxWidth-ellipsisWidth && len-->0) {
            str = str.substring(0, len);
            width = c.measureText(str).width;
        }
        return str+ellipsis;
    }
}


class EditableTextView extends View {
  constructor(l,t,w,h, cssName, font) {
    super(l,t,w,h);
    this.activate = this.activate.bind(this);
    this.change = this.change.bind(this);
    this.remove = this.remove.bind(this);
    this.filterKey = this.filterKey.bind(this);
    this.valueChanged = this.valueChanged.bind(this);
    this.enabled = true;

    this.v = "";
    this.wantsMouse = true;
    this.cssName = cssName;
    this.font = font;
    this.setBG(kBlackBG);
    this.prefix = "";
    this.c = "gray";
  }
  setPrefix(p) {
    this.prefix = p;
  }
  render(c) {
    c.fillStyle = this.enabled ? this.c : kEQLabel;
    c.font = this.font;
    c.textAlign = 'center';
    c.fillText(fitText(c, this.prefix + this.v, this.w), this.l + (this.r - this.l)/2, 4 + this.cy);
  }
  handleMouseDown(){
    if(this.enabled)
      this.activate();
  }
  enable(e) {
    this.remove();
    this.enabled = e;
    this.requestRender();
  }
  setRoot(root) {
    this.remove();
    super.setRoot(root);
  }
  activate() { 
    if(this.enabled && (this.tb == undefined)) {
      this.wantsMouse = false;
      this.tb = document.createElement('input');
      this.tb.setAttribute('type', 'text');
      this.tb.value = this.prefix + this.v;
      
      this.tb.className = this.cssName;
      var s = this.tb.style;
      s.font = this.font;
      s.left = this.l + "px";
      s.top = this.t + "px";
      s.width = this.w + "px";
      s.height = this.h + "px";
      s.lineHeight = s.height;

      this.tb.addEventListener('change', this.change);
      this.tb.addEventListener('blur', this.remove);
      this.root.getRootDiv().appendChild(this.tb);
      keyFilter.add(this);
    }
    this.tb.select();
    this.tb.focus();
  }
  valueChanged(v) { return v; } // extend this to filter and apply the change
  change() {
    if(this.tb != undefined) {
      this.v = this.valueChanged(this.tb.value);
      this.tb.blur();
    }
  }
  remove() {
    if(this.tb != undefined) {
      keyFilter.remove();
      var tb = this.tb;
      this.tb = undefined;
      this.root.getRootDiv().removeChild(tb);
      this.wantsMouse = true;
      this.requestRender();
    }
  }
  filterKey(keycode) {
    if((keycode == 'enter') && (this.tb != undefined)) {
      this.v = this.valueChanged(this.tb.value);
      this.tb.blur();
      return true;
    }
    if((keycode == 'escape') && (this.tb != undefined)) {
      this.tb.blur();
      return true;
    }
    return false;
  }
}


class MultiSelect extends View {
  constructor(d, l, t, w, h, ctrlID) {
    super(l, t, w, h);
    this.addOption = this.addOption.bind(this);
    this.changeCB = this.changeCB.bind(this);

    this.ctrlID = ctrlID;
    this.resized = this.resized.bind(this);
    evSys.addListener('wResized', this.resized);

    this.div = document.createElement('div');
    this.div.style.overflowY = 'scroll';
    this.scrollBarV = new VScrollBar(this.div);
    
    this.scrollBarV.containerDiv.className = 'user-multi-select';

    this.cBoxes = [];

    var s = this.scrollBarV.containerDiv.style;
    s.left = this.l + "px";
    s.top = (1 + this.t) + "px";
    this.div.style.width = s.width = this.w + "px";
    this.div.style.height = s.height = this.h + "px";

    this.div.onscroll = this.resized;

    d.appendChild(this.scrollBarV.containerDiv);
    this.setBG(kEmptyBG);

    this.scrollBarV.updateScroll(h);
  }
  resized() {
    this.scrollBarV.updateScroll(this.div.scrollHeight);
  }
  render(c) {}
  addOption(t, id) {
    var label = document.createElement('label');
    label.setAttribute('for', this.ctrlID + id); 
    label.innerHTML = t;

    var cbox = document.createElement('input');
    cbox.type = 'checkbox';
    cbox.id = this.ctrlID + id;
    cbox.v = id;
    cbox.checked = false;
    var that = this;
    cbox.addEventListener('change', e => { that.changeCB(e.srcElement.v, e.srcElement.checked) });

    var div = document.createElement('div');
    div.appendChild(cbox);
    div.appendChild(label);
    this.div.appendChild(div);
    this.cBoxes.push(cbox);
    this.scrollBarV.updateScroll(this.div.scrollHeight);
  }
  changeCB(id, checked){}
}


class SelectMenu extends View {
  constructor(d, l, t, w, h, changeCB) {
    super(l, t, w, h);
    this.addOption = this.addOption.bind(this);

    this.sel = document.createElement('select');
    this.sel.className = 'user-input-select';
    
    this.sel.style.left = this.l + "px";
    this.sel.style.top = (1 + this.t) + "px";
    this.sel.style.width = this.w + "px";

    this.sel.addEventListener('change', changeCB);
    d.appendChild(this.sel);
    this.setBG(kEmptyBG);
  }
  visChanged(){
    super.visChanged();
    this.sel.style.display = (this.isVis && this.isPVis) ? "block" : "none";
  }
  uRBCXCY() {
    // this is called when the rect of the view changes
    // hook it so we can update the select
    super.uRBCXCY();
    this.sel.style.left = this.l + "px";
    this.sel.style.top = (1 + this.t) + "px";
    this.sel.style.width = this.w + "px";
  }
  addOption(v, t, d) {
    var option = document.createElement('option');
    option.value = v;
    option.text = t;
    option.style.backgroundColor = kEmptyBG;
    if (d != undefined)
      option.disabled = d;
    this.sel.appendChild(option);
  }
  renameOptions(t,d) {
    var opts = this.sel.children;
    for (var i = 0; i < Math.min(opts.length, t.length); ++i) {
      opts[i].text = t[i];
    }
    if (d != undefined) {
      for (var i = 0; i < Math.min(opts.length, d.length); ++i) {
        opts[i].disabled = d[i];
      }
    }
  }
  clear() {
    this.sel.innerHTML = "";
  }
}


class DSSelectMenu extends View {
  constructor(d, l, t, w, h, dsKey, vals, i) {
    super(l, t, w, h);
    
    this.change = this.change.bind(this);
    this.onDS = this.onDS.bind(this);

    this.addSub(this.menu = new SelectMenu(d, this.l, this.t, this.w, this.h, this.change));

    this.dsI = i; // index of the datastore array to set
    this.dsKey = dsKey;

    ds.addListener(dsKey, this.onDS);

    var n = vals.length;
    for(var i = 0; i < n; i++)
      this.menu.addOption(vals[i].v, vals[i].l);

    this.menu.sel.value = ds.getKVArrayItem(dsKey, this.dsI);
  }
  onDS(k, v) {
    var vi = v[this.dsI];
    if(this.menu.sel.value != vi) {
      this.menu.sel.value = vi;
      this.requestRender();
    }
  }
  change() {
    ds.setV(this.dsKey, this.dsI, this.menu.sel.value);
  }
}


class DSLabelSelectMenu extends View {
  constructor(d, l, t, label, dsKey, vals, i) {
    super(l, t, 320, 28);
    
    this.addSub(this.menu = new DSSelectMenu(d, this.cx, t, this.w/2, this.h, dsKey, vals, i));
    this.label = label;
  }
  render(c) {

    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';
    c.fillText(this.label, this.l + 2, 7 + this.cy);
  }

  setMenuVals(d, dsKey, vals, i) {
    this.removeSub(this.menu);
    this.addSub(this.menu = new DSSelectMenu(d, this.cx, this.t, this.w/2, this.h, dsKey, vals, i));
  }
}


class FxThumb  extends View {
  constructor(l, t, fxType, isFxChStrip) {
    super(l, t, (isFxChStrip ? kFxChStripW : kChStripW) - 14, (isFxChStrip ? kFxThumbH : kThumbH) - 6);
    this.wantsMouse = true;
    this.fxType = fxType;
    this.isValid = false;
    this.setBG(kMixBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y) && this.isValid) {
      mixMgr.setActiveCh(this.info.chNum);
      chFxEdit.activate(this.fxType);
    }
  }
  setup(info) {
    this.info = info;
    this.requestRender();
  }
  setValid(iv){
    this.isValid = iv;
    this.setBG(iv ? kBlackBG : kMixBG);
    this.setVis(iv);
  }
}


// base class for simple ds views
class SimpleDSView extends View {
  constructor(l,t,w,h, dsKey, i) {
    super(l,t,w,h);

    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);
    this.dsToV = this.dsToV.bind(this);
    this.vToDS = this.vToDS.bind(this);

    SetupDSArrayListenerBasic(this, dsKey, i);
    this.v = 0;//this.dsToV(DSArrayGetBasic(this));
  }
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    this.v = this.dsToV(DSArrayGetBasic(this));
    this.requestRender();
  }
  onDSArray(k, v) {
    this.v = this.dsToV(DSArrayGetBasic(this));
    this.requestRender();
  }
  // override these to have a scaled UI representation of the value
  dsToV(dsv) { return dsv; }
  vToDS(v) { return v; }
}


class SimpleDSBtn extends SimpleDSView {
  constructor(l,t,w,h, dsKey, i)  {
    super(l,t,w,h, dsKey, i);

    this.wantsMouse = true;
    this.dragging = false;
    this.mIn = true;
    this.onV = 1;
    this.offV = 0;
    this.v = this.offV;
  }
  handleMouseDown(xy) { 
    if(this.contains(xy.x, xy.y))
    {
      this.v = this.v == this.offV ? this.onV : this.offV;
      this.dragging = true;
      this.mIn = true;
      this.requestRender();
    }
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging) { 
      var mIn = this.contains(xy.x, xy.y);
      if(mIn != this.mIn) {
        this.mIn = mIn;
        this.v = this.v == 0 ? this.onV : this.offV;
        this.requestRender();
      }
    }
  }
  handleMouseUp() {
    this.dragging = false;
    if(this.mIn) {
      DSArraySetBasic(this, this.vToDS(this.v));
    }
    this.mIn = false;
  }
}


class ToggleBtn extends SimpleDSBtn {
  constructor(l, t, w, label, dsKey, i, c, bg = undefined)  {
    super(l, t, w, 30, dsKey, i);
    this.label = label;
    this.setBG(bg == undefined ? kMixBG : bg);
    this.c = c;
    this.enabled = true;
  }
  enable(e) {
    this.enabled = e;
    this.requestRender();
  }
  render(c) {
    drawButton(c, this, this.label, this.v == this.onV, this.c, this.enabled);
  }
}


class SimpleDSKnob extends SimpleDSView {
  constructor(l,t,w,h, dsKey, i) {
    super(l,t,w,h, dsKey, i);

    this.setVal = this.setVal.bind(this);
    this.wantsMouse = true;
    this.dragging = false;
  }
  handleMouseDown(xy) { 
    if(this.contains(xy.x, xy.y))
    {
      this.dragging = true;
      this.mouse = xy;
    }
  }
  setVal(v) {
    // derived classes can override this to hook when the control is changed by the user
    DSArraySetBasic(this, v);
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging) { 
      var delta = (xy.x - this.mouse.x) - (xy.y - this.mouse.y);
      this.mouse = xy;
      if(delta != 0) {      
        var v  = (delta * 0.002) + this.v;
        v = v > 1 ? 1 : v;
        v = v < 0 ? 0 : v;
        this.v = v;
        //console.log("k " + v);
        this.setVal(this.vToDS(this.v));
        this.requestRender();
      }
    }
  }
  handleMouseUp() {
    this.dragging = false;
  }
  render(c) { }
}
