'use strict';


class View extends Rect{
  constructor(l,t,w,h) {

    super(l, t, w, h);

    this.setRoot = this.setRoot.bind(this);
    this.addSub = this.addSub.bind(this);
    this.removeSub = this.removeSub.bind(this);
    this.setVis = this.setVis.bind(this);
    this.visChanged = this.visChanged.bind(this);
    this.offset = this.offset.bind(this);
    this.render = this.render.bind(this);
    this.renderSelfAndSubs = this.renderSelfAndSubs.bind(this);
    this.requestRender = this.requestRender.bind(this);
    this.getMouseHandler = this.getMouseHandler.bind(this);
    this.handleDblClick = this.handleDblClick.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);
    this.handleMouseMoved = this.handleMouseMoved.bind(this);
    this.handleMouseUp = this.handleMouseUp.bind(this);
    this.findTop = this.findTop.bind(this);
    this.setBG = this.setBG.bind(this);
    this.dbgStrokeRect = this.dbgStrokeRect.bind(this);
    
    // note  that coordinates are always relative to the root view 
    this.root = undefined;
    this.subViews = new List;
    this.wantsMouse = false;
    this.isVis = true;
    this.isPVis = true;
  }
  setRoot(root) {
    this.root = root;
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).setRoot(root);
  }
  addSub(v) {
    this.subViews.add(v);
    v.setRoot(this.root);
    v.setParentVis(this.isVis && this.isPVis);
  }
  removeSub(v) {
    this.subViews.remove(v);
    v.setRoot(undefined);
    v.setParentVis(false);
    this.requestRender();
  }
  visChanged(){
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).setParentVis(this.isVis && this.isPVis);

    if(this.isVis && this.isPVis)
      this.requestRender();
  }
  setParentVis(isPVis) {
    if(this.isPVis != isPVis) {
      this.isPVis = isPVis;
      this.visChanged();
    }
  }
  setVis(isVis) {
    if(this.isVis != isVis) {
      this.isVis = isVis;
      this.visChanged();
    }
  }
  findTop(r) {
    // find the top view that contains the rect
    if(this.containsR(r)) {
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i).findTop(r);
        if(v != undefined)
          return v;
      }

      return this;
    }
    return undefined;
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    var n = this.subViews.length();
    for(var i = 0; i < n; i++)
      this.subViews.get(i).offset(dx, dy);
  }
  move(l, t) {
    this.offset(l - this.l, t - this.t);
  }

  requestRender() {
    if(this.root != undefined && this.isVis && this.isPVis)
      this.root.requestRender(this);
  }
  
  renderSelfAndSubs(c) {
    if(this.isVis && this.isPVis) {
      if(this.bg != undefined) {
        c.fillStyle = this.bg;
        c.fillRect(this.l, this.t, this.w, this.h);
      }
      this.render(c);

      // draw all the visible children
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i);
        v.renderSelfAndSubs(c);
      }
      
      //this.dbgStrokeRect(c);
    }
  }

  setBG(bg) { this.bg = bg; }

  getMouseHandler(xy) {
    if(this.isVis && this.isPVis && this.contains(xy.x, xy.y)) {
      // find the top view that contains the coordinate
      var n = this.subViews.length();
      for(var i = 0; i < n; i++) {
        var v = this.subViews.get(i).getMouseHandler(xy);
        if(v != undefined)
          return v;
      }

      if(this.wantsMouse)
        return this;
    }
    return undefined;
  }

  dbgStrokeRect(c) { 
    c.lineWidth = 1;
    c.strokeStyle = "yellow";
    c.strokeRect(this.l, this.t, this.w, this.h);
  }
      
  // extendend views implement render to draw their contents
  render(c) {}
  // extend these to handle the mouse
  handleDblClick() {}
  handleMouseDown(xy) {}
  handleMouseMoved(xy, delta) {}
  handleMouseUp() {}
}


