'use strict';


class IPAddrEdit extends EditableTextView {
  constructor(l,t,w,h) {
    super(l,t,w,h, 'user-input-input', 'Bold 30px Helvetica');

    this.setup = this.setup.bind(this);

    // this is the default used during development
    this.v = '169.254.1.255';
    this.setBG(kFxEditBG);
  }
  valueChanged(v) {
    device.wsConnect(v);
    return v; 
  } 
  setup(info) {
    this.info = info;
    this.refresh();
  }
}

const kWelcomeBG = new OffScreenImage("welcomeBG.png");
    
class WelcomeView extends View {
  constructor() {
    super(0,0,kFxWidth,kFxHeight);
  
    this.modal = new ModalCanvas(false);
    this.activate = this.activate.bind(this);
    this.deviceStatusChanged = this.deviceStatusChanged.bind(this);
    this.proxyStatusChanged = this.proxyStatusChanged.bind(this);
    this.updateConnectingProgress = this.updateConnectingProgress.bind(this);
    this.wantsMouse = true;
    this.setBG(kBlackBG);

    // mac or pc is connected, maybe show updater
    if(gen5Lib.hasLib() || device.hasHostMsgPipe()) {
        this.addSub(this.upd = new UpdateView(620, 355, 7));
        this.upd.setVis(false);
        //this.upd.setBG('rgba(0,0,0,1)')
    }

    this.addSub(this.ipEdit = new IPAddrEdit(this.cx - 100, this.b - 60, 200, 45));
    this.ipEdit.setVis(false);

    evSys.addListener('ProxyStatusChanged', this.proxyStatusChanged);
    evSys.addListener('DeviceStatusChanged', this.deviceStatusChanged);
    ds.addListener('kAPIVersion', this.deviceStatusChanged);

    this.connected = false;
    this.apiVersion = kDataStoreAPIVersionUnknown;
    this.proxyStatus = 2;

    this.activate();
  }
  proxyStatusChanged(ev, v) {
    this.proxyStatus = v;
    switch (v) {
      case 0:
        this.progress = 0;
        this.activate();
        break;
      case 1:
        this.startConnecting();
        break;
      case 2:
        break;
    }
  }
  deviceStatusChanged() {
    this.connected = device.isConnected() || device.isTest();
    this.apiVersion = ds.getKVArrayItem('kAPIVersion', 0);
    if(this.upd)
      this.upd.setVis(this.connected && (this.apiVersion != kDataStoreAPIVersion) && (this.proxyStatus == 2));
    if(this.connected && (device.isTest() || (this.apiVersion == kDataStoreAPIVersion)))
      this.deactivate();
    else {
      this.activate();
    }
  }
  deactivate() {
    if (this.progressTimer != undefined) {
      clearInterval(this.progressTimer);
      this.progressTimer = undefined;
    }
    this.modal.deactivate();
  }
  activate() {
    this.modal.activate(this);
    this.requestRender();
  }
  handleMouseDown(xy) {
    if(!this.ipEdit.isVis) {
      this.ipEdit.setVis(true);
      this.ipEdit.activate();
      this.requestRender();
    }
  }
  startConnecting() {
    this.progress = 0;
    this.requestRender();
    if (this.progressTimer)
      clearInterval(this.progressTimer);
    this.progressTimer = setInterval(this.updateConnectingProgress, 50);
  }
  updateConnectingProgress() {
    var next = this.progress + 1/80;
    if (next < 1){
      this.progress = next;
    }
    else {
      clearInterval(this.progressTimer);
      this.progressTimer = undefined;
      this.progress = 1;
    }
    this.requestRender();
  }
  render(c) {
    c.drawImage(kWelcomeBG.canvas, this.l, this.t, this.w, this.h);
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);
    if(this.ipEdit.isVis)
      c.strokeRect(this.ipEdit.l, this.ipEdit.t, this.ipEdit.w, this.ipEdit.h);

    c.fillStyle = "rgb(200,202,204)";
    c.textAlign = 'left';
    if((this.apiVersion != kDataStoreAPIVersion) && device.isConnected() && (this.proxyStatus == 2)) {
      c.font = "Lighter 20px Helvetica";
      var t = "";
      if((this.apiVersion == 0) || (kDataStoreAPIVersion > this.apiVersion)) {
        t = "CueMix requires a newer firmware version. Please update the firmware on your device.";
      }
      else if(kDataStoreAPIVersion < this.apiVersion) {
        t = "The Device requires a newer version of CueMix. Please update CueMix.";
      }
      fillTextWrapped(c, t, this.cx + 60, this.cy-30, 460, 24);
    }
    else if(this.proxyStatus != 1) {
      c.font = "Lighter 30px Helvetica";
      c.fillText("Waiting for device to connect.", this.cx + 60, this.cy-15);
      c.fillText("Check power and cables...", this.cx + 60, this.cy+25);
    }

    c.font = "Lighter 18px Helvetica";
    c.textAlign = 'center';
    if(this.proxyStatus == 1) {
      c.strokeStyle = "rgb(200,202,204)";
      c.fillStyle = "rgb(200,202,204)";
      var status = (this.progress < 1) ? "Connecting..." : "Any second now...";
      c.fillText(status, this.cx, this.t + 550);
      c.strokeRect(this.l + 60, this.t + 560, 960, 25);
      c.fillRect(this.l + 60, this.t + 560, 960*this.progress, 25);
    }
    else {
      c.fillText("IP Connect", this.cx, this.b - 40);
      var w = c.measureText("IP Connect").width;
      c.fillRect(this.cx - w/2, this.b - 35, w, 1);
    }
  }
}

